(*

HiggsMassSquared.m by D.Ochoa on 18 Oct 2017 

3-loop order Higgs boson squared pole mass of O(alpha_t alpha_3^2) x
O((v^2/M_S^2)^0) expressed in terms of MSSM DR-bar parameters in the
MSSM.

*** List of symbols ***

k = 1/(4 Pi)^2      : Loop factor 
M                   : Renormalization scale
MZ                  : Z boson pole mass
z2                  : Zeta[2]
z3                  : Zeta[3]

MSSM DR parameters

mt = yt vu                      : Top quark mass, 

a3 = k^(1/2) g3^2               : Strong gauge coupling "fine structure constant"
at = k^(1/2) yt^2 Sin[beta]^2   : Top Yukawa coupling "fine structure constant"

beta   : ArcTan of the ratio of the vacuum expectation values of the Higgs fields

m3     : Soft-SUSY breaking gluino mass parameter
mQ3    : Soft-SUSY breaking doublet stop mass parameter
mU3    : Soft-SUSY breaking singlet stop mass parameter
msq    : Soft-SUSY breaking squark mass parameter
Xt     : Soft-SUSY breaking stop mixing

MSSM MS bar threshold corrections

dlatas2        : Threshold correction of the SM self-interaction coupling of O(at a3^2)

*)

SMConstatas2 = (-3776/9
                -2048/3 PolyLog[4,1/2]
                +320 Zeta[3]
                +(2816 Zeta[2]^2)/15
                +256/9 Log[2]^2 (-Log[2]^2 + 6 Zeta[2]));

Mh2=( 
(-4*a3^2*at*k^(3/2)*mt^2*(128*m3^16 - 550*m3^14*mQ3^2 + 506*m3^12*mQ3^4 - 
    85*m3^10*mQ3^6 - 88*m3^8*mQ3^8 + 4*m3^6*mQ3^10 + 38*m3^4*mQ3^12 - 
    9*m3^2*mQ3^14 + 60*m3^12*mQ3^2*msq^2 + 120*m3^10*mQ3^4*msq^2 - 
    180*m3^8*mQ3^6*msq^2 - 90*m3^10*mQ3^2*msq^4 + 60*m3^8*mQ3^4*msq^4 + 
    30*m3^6*mQ3^6*msq^4 - 346*m3^14*mU3^2 + 1902*m3^12*mQ3^2*mU3^2 - 
    1873*m3^10*mQ3^4*mU3^2 + 253*m3^8*mQ3^6*mU3^2 + 612*m3^6*mQ3^8*mU3^2 - 
    368*m3^4*mQ3^10*mU3^2 + 47*m3^2*mQ3^12*mU3^2 - 3*mQ3^14*mU3^2 - 
    60*m3^12*msq^2*mU3^2 - 300*m3^10*mQ3^2*msq^2*mU3^2 - 
    180*m3^8*mQ3^4*msq^2*mU3^2 + 540*m3^6*mQ3^6*msq^2*mU3^2 + 
    90*m3^10*msq^4*mU3^2 + 210*m3^8*mQ3^2*msq^4*mU3^2 - 
    210*m3^6*mQ3^4*msq^4*mU3^2 - 90*m3^4*mQ3^6*msq^4*mU3^2 + 
    280*m3^12*mU3^4 - 2509*m3^10*mQ3^2*mU3^4 + 3011*m3^8*mQ3^4*mU3^4 - 
    1194*m3^6*mQ3^6*mU3^4 - 98*m3^4*mQ3^8*mU3^4 + 175*m3^2*mQ3^10*mU3^4 - 
    mQ3^12*mU3^4 + 180*m3^10*msq^2*mU3^4 + 540*m3^8*mQ3^2*msq^2*mU3^4 - 
    180*m3^6*mQ3^4*msq^2*mU3^4 - 540*m3^4*mQ3^6*msq^2*mU3^4 - 
    270*m3^8*msq^4*mU3^4 - 90*m3^6*mQ3^2*msq^4*mU3^4 + 
    270*m3^4*mQ3^4*msq^4*mU3^4 + 90*m3^2*mQ3^6*msq^4*mU3^4 - 13*m3^10*mU3^6 + 
    1341*m3^8*mQ3^2*mU3^6 - 1842*m3^6*mQ3^4*mU3^6 + 874*m3^4*mQ3^6*mU3^6 - 
    73*m3^2*mQ3^8*mU3^6 - 63*mQ3^10*mU3^6 - 180*m3^8*msq^2*mU3^6 - 
    420*m3^6*mQ3^2*msq^2*mU3^6 + 420*m3^4*mQ3^4*msq^2*mU3^6 + 
    180*m3^2*mQ3^6*msq^2*mU3^6 + 270*m3^6*msq^4*mU3^6 - 
    90*m3^4*mQ3^2*msq^4*mU3^6 - 150*m3^2*mQ3^4*msq^4*mU3^6 - 
    30*mQ3^6*msq^4*mU3^6 - 37*m3^8*mU3^8 - 268*m3^6*mQ3^2*mU3^8 + 
    450*m3^4*mQ3^4*mU3^8 - 268*m3^2*mQ3^6*mU3^8 + 67*mQ3^8*mU3^8 + 
    60*m3^6*msq^2*mU3^8 + 120*m3^4*mQ3^2*msq^2*mU3^8 - 
    180*m3^2*mQ3^4*msq^2*mU3^8 - 90*m3^4*msq^4*mU3^8 + 
    60*m3^2*mQ3^2*msq^4*mU3^8 + 30*mQ3^4*msq^4*mU3^8)*Log[mQ3^2/MR^2]^3)/
  (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) - 
 (4*a3^2*at*k^(3/2)*mt^2*(128*m3^10 + 38*m3^8*mQ3^2 - 2*m3^6*mQ3^4 + 
    9*m3^4*mQ3^6 - 60*m3^6*mQ3^2*msq^2 + 90*m3^4*mQ3^2*msq^4 - 
    550*m3^8*mU3^2 + 276*m3^6*mQ3^2*mU3^2 - 33*m3^4*mQ3^4*mU3^2 - 
    6*m3^2*mQ3^6*mU3^2 + 60*m3^6*msq^2*mU3^2 - 120*m3^4*mQ3^2*msq^2*mU3^2 - 
    90*m3^4*msq^4*mU3^2 - 60*m3^2*mQ3^2*msq^4*mU3^2 + 494*m3^6*mU3^4 - 
    429*m3^4*mQ3^2*mU3^4 + 36*m3^2*mQ3^4*mU3^4 - 3*mQ3^6*mU3^4 + 
    120*m3^4*msq^2*mU3^4 + 180*m3^2*mQ3^2*msq^2*mU3^4 + 60*m3^2*msq^4*mU3^4 - 
    30*mQ3^2*msq^4*mU3^4 - 59*m3^4*mU3^6 + 234*m3^2*mQ3^2*mU3^6 - 
    mQ3^4*mU3^6 - 180*m3^2*msq^2*mU3^6 + 30*msq^4*mU3^6 - 136*m3^2*mU3^8 - 
    63*mQ3^2*mU3^8 + 67*mU3^10)*Log[mU3^2/MR^2]^3)/
  (3*(m3^2 - mU3^2)^4*(-mQ3^2 + mU3^2)) + 
 Xt^6*((-2560*a3^2*at*k^(3/2)*m3^6*mt^2*Log[m3^2/MR^2]^2)/
    (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2) + 
   (5120*a3^2*at*k^(3/2)*m3^4*mQ3^2*mt^2*Log[m3^2/MR^2]*Log[mQ3^2/MR^2])/
    (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^3) + 
   (1280*a3^2*at*k^(3/2)*m3^6*mt^2*(mQ3^2 + mU3^2)*Log[m3^2/MR^2]^2*
     Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*
     (mQ3^2 - mU3^2)^3) - (2560*a3^2*at*k^(3/2)*m3^2*mQ3^4*mt^2*
     Log[mQ3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^4) - 
   (2560*a3^2*at*k^(3/2)*m3^4*mQ3^2*mt^2*(mQ3^2 + mU3^2)*Log[m3^2/MR^2]*
     Log[mQ3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*
     (mQ3^2 - mU3^2)^4) + (1280*a3^2*at*k^(3/2)*m3^2*mQ3^4*mt^2*
     (mQ3^2 + mU3^2)*Log[mQ3^2/MR^2]^3)/(3*(m3^2 - mQ3^2)^2*
     (mQ3^2 - mU3^2)^5) + 
   ((-5120*a3^2*at*k^(3/2)*m3^4*mt^2*mU3^2*Log[m3^2/MR^2])/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) + 
     (5120*a3^2*at*k^(3/2)*m3^2*mQ3^2*mt^2*mU3^2*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^4))*Log[mU3^2/MR^2] + 
   ((-1280*a3^2*at*k^(3/2)*m3^6*mt^2*(mQ3^2 + mU3^2)*Log[m3^2/MR^2]^2)/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) + 
     (2560*a3^2*at*k^(3/2)*m3^4*mt^2*(mQ3^2 + mU3^2)*
       (m3^2*mQ3^2 + m3^2*mU3^2 - 2*mQ3^2*mU3^2)*Log[m3^2/MR^2]*
       Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*
       (mQ3^2 - mU3^2)^4) - (1280*a3^2*at*k^(3/2)*m3^2*mt^2*(mQ3^2 + mU3^2)*
       (m3^2*mQ3^4 + 2*m3^2*mQ3^2*mU3^2 - 3*mQ3^4*mU3^2)*Log[mQ3^2/MR^2]^2)/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^5))*
    Log[mU3^2/MR^2] - (2560*a3^2*at*k^(3/2)*m3^2*mt^2*mU3^4*
     Log[mU3^2/MR^2]^2)/(3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^4) + 
   ((-2560*a3^2*at*k^(3/2)*m3^4*mt^2*mU3^2*(mQ3^2 + mU3^2)*Log[m3^2/MR^2])/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^4) + 
     (1280*a3^2*at*k^(3/2)*m3^2*mt^2*(mQ3^2 + mU3^2)*(2*m3^2*mQ3^2*mU3^2 + 
        m3^2*mU3^4 - 3*mQ3^2*mU3^4)*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^5))*
    Log[mU3^2/MR^2]^2 + (1280*a3^2*at*k^(3/2)*m3^2*mt^2*mU3^4*(mQ3^2 + mU3^2)*
     Log[mU3^2/MR^2]^3)/(3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^5)) + 
 Log[mQ3^2/MR^2]^2*((-16*a3*at*k*(3*m3^4 - 2*m3^2*mQ3^2 + mQ3^4)*mt^2)/
    (m3^2 - mQ3^2)^2 + (16*a3^2*at*k^(3/2)*mt^2*(64*m3^16 - 245*m3^14*mQ3^2 + 
      240*m3^12*mQ3^4 - 170*m3^10*mQ3^6 + 110*m3^8*mQ3^8 - 27*m3^6*mQ3^10 - 
      139*m3^14*mU3^2 + 633*m3^12*mQ3^2*mU3^2 - 432*m3^10*mQ3^4*mU3^2 + 
      162*m3^8*mQ3^6*mU3^2 - 157*m3^6*mQ3^8*mU3^2 + 45*m3^4*mQ3^10*mU3^2 + 
      87*m3^12*mU3^4 - 700*m3^10*mQ3^2*mU3^4 + 400*m3^8*mQ3^4*mU3^4 - 
      20*m3^6*mQ3^6*mU3^4 + 101*m3^4*mQ3^8*mU3^4 - 36*m3^2*mQ3^10*mU3^4 + 
      22*m3^10*mU3^6 + 316*m3^8*mQ3^2*mU3^6 - 146*m3^6*mQ3^4*mU3^6 - 
      80*m3^4*mQ3^6*mU3^6 - 12*m3^2*mQ3^8*mU3^6 + 12*mQ3^10*mU3^6 - 
      28*m3^8*mU3^8 - 34*m3^6*mQ3^2*mU3^8 - 2*m3^4*mQ3^4*mU3^8 + 
      48*m3^2*mQ3^6*mU3^8 - 12*mQ3^8*mU3^8)*Log[m3^2/MR^2])/
    (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) + 
   a3^2*at*k^(3/2)*mt^2*((8*(2*m3^2*mQ3^2 - mQ3^4))/(3*(m3^2 - mQ3^2)^2) - 
     (48*(2*m3^2*mQ3^2 - mQ3^4)^2)/(m3^2 - mQ3^2)^4 - 
     (16*(3*m3^4 - 4*m3^2*mQ3^2 + 2*mQ3^4)*(-11*m3^4 + 3*m3^2*mQ3^2 + 
        3*m3^2*mU3^2 + 5*mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)) + 
     (8*(-309*m3^10*mQ3^2 + 419*m3^8*mQ3^4 - 21*m3^6*mQ3^6 - 100*m3^4*mQ3^8 - 
        9*m3^2*mQ3^10 + 90*m3^6*mQ3^4*msq^2 + 30*m3^4*mQ3^6*msq^2 + 
        53*m3^10*mU3^2 + 711*m3^8*mQ3^2*mU3^2 - 1011*m3^6*mQ3^4*mU3^2 + 
        34*m3^4*mQ3^6*mU3^2 + 288*m3^2*mQ3^8*mU3^2 - 3*mQ3^10*mU3^2 - 
        90*m3^6*mQ3^2*msq^2*mU3^2 - 210*m3^4*mQ3^4*msq^2*mU3^2 - 
        60*m3^2*mQ3^6*msq^2*mU3^2 - 106*m3^8*mU3^4 - 557*m3^6*mQ3^2*mU3^4 + 
        938*m3^4*mQ3^4*mU3^4 - 231*m3^2*mQ3^6*mU3^4 - 140*mQ3^8*mU3^4 + 
        180*m3^4*mQ3^2*msq^2*mU3^4 + 150*m3^2*mQ3^4*msq^2*mU3^4 + 
        30*mQ3^6*msq^2*mU3^4 + 53*m3^6*mU3^6 + 152*m3^4*mQ3^2*mU3^6 - 
        295*m3^2*mQ3^4*mU3^6 + 146*mQ3^6*mU3^6 - 90*m3^2*mQ3^2*msq^2*mU3^6 - 
        30*mQ3^4*msq^2*mU3^6 - 9*m3^2*mQ3^2*mU3^8 - 3*mQ3^4*mU3^8))/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) + 
     (4*(128*m3^16 - 556*m3^14*mQ3^2 + 714*m3^12*mQ3^4 - 373*m3^10*mQ3^6 + 
        96*m3^8*mQ3^8 - 42*m3^6*mQ3^10 + 38*m3^4*mQ3^12 - 9*m3^2*mQ3^14 + 
        60*m3^12*mQ3^2*msq^2 + 120*m3^10*mQ3^4*msq^2 - 180*m3^8*mQ3^6*msq^2 - 
        90*m3^10*mQ3^2*msq^4 + 60*m3^8*mQ3^4*msq^4 + 30*m3^6*mQ3^6*msq^4 - 
        340*m3^14*mU3^2 + 1712*m3^12*mQ3^2*mU3^2 - 2209*m3^10*mQ3^4*mU3^2 + 
        933*m3^8*mQ3^6*mU3^2 + 106*m3^6*mQ3^8*mU3^2 - 230*m3^4*mQ3^10*mU3^2 + 
        47*m3^2*mQ3^12*mU3^2 - 3*mQ3^14*mU3^2 - 60*m3^12*msq^2*mU3^2 - 
        300*m3^10*mQ3^2*msq^2*mU3^2 - 180*m3^8*mQ3^4*msq^2*mU3^2 + 
        540*m3^6*mQ3^6*msq^2*mU3^2 + 90*m3^10*msq^4*mU3^2 + 
        210*m3^8*mQ3^2*msq^4*mU3^2 - 210*m3^6*mQ3^4*msq^4*mU3^2 - 
        90*m3^4*mQ3^6*msq^4*mU3^2 + 262*m3^12*mU3^4 - 1903*m3^10*mQ3^2*
         mU3^4 + 2771*m3^8*mQ3^4*mU3^4 - 1506*m3^6*mQ3^6*mU3^4 + 
        316*m3^4*mQ3^8*mU3^4 + 37*m3^2*mQ3^10*mU3^4 - mQ3^12*mU3^4 + 
        180*m3^10*msq^2*mU3^4 + 540*m3^8*mQ3^2*msq^2*mU3^4 - 
        180*m3^6*mQ3^4*msq^2*mU3^4 - 540*m3^4*mQ3^6*msq^2*mU3^4 - 
        270*m3^8*msq^4*mU3^4 - 90*m3^6*mQ3^2*msq^4*mU3^4 + 
        270*m3^4*mQ3^4*msq^4*mU3^4 + 90*m3^2*mQ3^6*msq^4*mU3^4 + 
        5*m3^10*mU3^6 + 723*m3^8*mQ3^2*mU3^6 - 1186*m3^6*mQ3^4*mU3^6 + 
        610*m3^4*mQ3^6*mU3^6 - 119*m3^2*mQ3^8*mU3^6 - 17*mQ3^10*mU3^6 - 
        180*m3^8*msq^2*mU3^6 - 420*m3^6*mQ3^2*msq^2*mU3^6 + 
        420*m3^4*mQ3^4*msq^2*mU3^6 + 180*m3^2*mQ3^6*msq^2*mU3^6 + 
        270*m3^6*msq^4*mU3^6 - 90*m3^4*mQ3^2*msq^4*mU3^6 - 
        150*m3^2*mQ3^4*msq^4*mU3^6 - 30*mQ3^6*msq^4*mU3^6 - 43*m3^8*mU3^8 - 
        60*m3^6*mQ3^2*mU3^8 + 162*m3^4*mQ3^4*mU3^8 - 84*m3^2*mQ3^6*mU3^8 + 
        21*mQ3^8*mU3^8 + 60*m3^6*msq^2*mU3^8 + 120*m3^4*mQ3^2*msq^2*mU3^8 - 
        180*m3^2*mQ3^4*msq^2*mU3^8 - 90*m3^4*msq^4*mU3^8 + 
        60*m3^2*mQ3^2*msq^4*mU3^8 + 30*mQ3^4*msq^4*mU3^8))/
      (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) + 
     (8*(m3^4 - 34*m3^2*mQ3^2 + 17*mQ3^4)*(7*m3^6 - 2*mQ3^4*mU3^2 - 
        m3^4*(7*mQ3^2 + 6*mU3^2) + m3^2*(3*mQ3^4 + 5*mQ3^2*mU3^2)))/
      (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)) - 
     (8*(128*m3^16 - 556*m3^14*mQ3^2 + 718*m3^12*mQ3^4 - 463*m3^10*mQ3^6 + 
        184*m3^8*mQ3^8 - 64*m3^6*mQ3^10 + 38*m3^4*mQ3^12 - 9*m3^2*mQ3^14 + 
        60*m3^12*mQ3^2*msq^2 + 120*m3^10*mQ3^4*msq^2 - 180*m3^8*mQ3^6*msq^2 - 
        90*m3^10*mQ3^2*msq^4 + 60*m3^8*mQ3^4*msq^4 + 30*m3^6*mQ3^6*msq^4 - 
        340*m3^14*mU3^2 + 1708*m3^12*mQ3^2*mU3^2 - 2131*m3^10*mQ3^4*mU3^2 + 
        1115*m3^8*mQ3^6*mU3^2 - 136*m3^6*mQ3^8*mU3^2 - 
        164*m3^4*mQ3^10*mU3^2 + 47*m3^2*mQ3^12*mU3^2 - 3*mQ3^14*mU3^2 - 
        60*m3^12*msq^2*mU3^2 - 300*m3^10*mQ3^2*msq^2*mU3^2 - 
        180*m3^8*mQ3^4*msq^2*mU3^2 + 540*m3^6*mQ3^6*msq^2*mU3^2 + 
        90*m3^10*msq^4*mU3^2 + 210*m3^8*mQ3^2*msq^4*mU3^2 - 
        210*m3^6*mQ3^4*msq^4*mU3^2 - 90*m3^4*mQ3^6*msq^4*mU3^2 + 
        262*m3^12*mU3^4 - 1891*m3^10*mQ3^2*mU3^4 + 2513*m3^8*mQ3^4*mU3^4 - 
        1512*m3^6*mQ3^6*mU3^4 + 514*m3^4*mQ3^8*mU3^4 - 29*m3^2*mQ3^10*mU3^4 - 
        mQ3^12*mU3^4 + 180*m3^10*msq^2*mU3^4 + 540*m3^8*mQ3^2*msq^2*mU3^4 - 
        180*m3^6*mQ3^4*msq^2*mU3^4 - 540*m3^4*mQ3^6*msq^2*mU3^4 - 
        270*m3^8*msq^4*mU3^4 - 90*m3^6*mQ3^2*msq^4*mU3^4 + 
        270*m3^4*mQ3^4*msq^4*mU3^4 + 90*m3^2*mQ3^6*msq^4*mU3^4 + 
        5*m3^10*mU3^6 + 711*m3^8*mQ3^2*mU3^6 - 920*m3^6*mQ3^4*mU3^6 + 
        436*m3^4*mQ3^6*mU3^6 - 141*m3^2*mQ3^8*mU3^6 + 5*mQ3^10*mU3^6 - 
        180*m3^8*msq^2*mU3^6 - 420*m3^6*mQ3^2*msq^2*mU3^6 + 
        420*m3^4*mQ3^4*msq^2*mU3^6 + 180*m3^2*mQ3^6*msq^2*mU3^6 + 
        270*m3^6*msq^4*mU3^6 - 90*m3^4*mQ3^2*msq^4*mU3^6 - 
        150*m3^2*mQ3^4*msq^4*mU3^6 - 30*mQ3^6*msq^4*mU3^6 - 43*m3^8*mU3^8 - 
        56*m3^6*mQ3^2*mU3^8 + 72*m3^4*mQ3^4*mU3^8 + 4*m3^2*mQ3^6*mU3^8 - 
        mQ3^8*mU3^8 + 60*m3^6*msq^2*mU3^8 + 120*m3^4*mQ3^2*msq^2*mU3^8 - 
        180*m3^2*mQ3^4*msq^2*mU3^8 - 90*m3^4*msq^4*mU3^8 + 
        60*m3^2*mQ3^2*msq^4*mU3^8 + 30*mQ3^4*msq^4*mU3^8)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)))) + 
 Log[mU3^2/MR^2]^2*((-16*a3*at*k*mt^2*(3*m3^4 - 2*m3^2*mU3^2 + mU3^4))/
    (m3^2 - mU3^2)^2 - (16*a3^2*at*k^(3/2)*mt^2*(64*m3^16 - 139*m3^14*mQ3^2 + 
      87*m3^12*mQ3^4 + 22*m3^10*mQ3^6 - 28*m3^8*mQ3^8 - 245*m3^14*mU3^2 + 
      633*m3^12*mQ3^2*mU3^2 - 700*m3^10*mQ3^4*mU3^2 + 316*m3^8*mQ3^6*mU3^2 - 
      34*m3^6*mQ3^8*mU3^2 + 240*m3^12*mU3^4 - 432*m3^10*mQ3^2*mU3^4 + 
      400*m3^8*mQ3^4*mU3^4 - 146*m3^6*mQ3^6*mU3^4 - 2*m3^4*mQ3^8*mU3^4 - 
      170*m3^10*mU3^6 + 162*m3^8*mQ3^2*mU3^6 - 20*m3^6*mQ3^4*mU3^6 - 
      80*m3^4*mQ3^6*mU3^6 + 48*m3^2*mQ3^8*mU3^6 + 110*m3^8*mU3^8 - 
      157*m3^6*mQ3^2*mU3^8 + 101*m3^4*mQ3^4*mU3^8 - 12*m3^2*mQ3^6*mU3^8 - 
      12*mQ3^8*mU3^8 - 27*m3^6*mU3^10 + 45*m3^4*mQ3^2*mU3^10 - 
      36*m3^2*mQ3^4*mU3^10 + 12*mQ3^6*mU3^10)*Log[m3^2/MR^2])/
    (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^4*(mQ3^2 - mU3^2)) + 
   a3^2*at*k^(3/2)*mt^2*((-8*(m3^4 - 34*m3^2*mQ3^2 + 17*mQ3^4)*
       (3*m3^4 - 4*m3^2*mU3^2 + 2*mU3^4))/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^2) + (8*(12*m3^10*mQ3^2 - 14*m3^8*mQ3^4 + 
        34*m3^6*mQ3^6 - 29*m3^4*mQ3^8 + 9*m3^2*mQ3^10 + 64*m3^8*mQ3^2*mU3^2 - 
        188*m3^6*mQ3^4*mU3^2 + 120*m3^4*mQ3^6*mU3^2 - 35*m3^2*mQ3^8*mU3^2 + 
        3*mQ3^10*mU3^2 - 78*m3^6*mQ3^2*mU3^4 + 175*m3^4*mQ3^4*mU3^4 - 
        65*m3^2*mQ3^6*mU3^4 + 4*mQ3^8*mU3^4 + 26*m3^4*mQ3^2*mU3^6 - 
        57*m3^2*mQ3^4*mU3^6 + 19*mQ3^6*mU3^6))/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^3) - (4*(128*m3^10 + 44*m3^8*mQ3^2 - 2*m3^6*mQ3^4 + 
        9*m3^4*mQ3^6 - 60*m3^6*mQ3^2*msq^2 + 90*m3^4*mQ3^2*msq^4 - 
        556*m3^8*mU3^2 + 64*m3^6*mQ3^2*mU3^2 - 33*m3^4*mQ3^4*mU3^2 - 
        6*m3^2*mQ3^6*mU3^2 + 60*m3^6*msq^2*mU3^2 - 120*m3^4*mQ3^2*msq^2*
         mU3^2 - 90*m3^4*msq^4*mU3^2 - 60*m3^2*mQ3^2*msq^4*mU3^2 + 
        706*m3^6*mU3^4 - 51*m3^4*mQ3^2*mU3^4 + 36*m3^2*mQ3^4*mU3^4 - 
        3*mQ3^6*mU3^4 + 120*m3^4*msq^2*mU3^4 + 180*m3^2*mQ3^2*msq^2*mU3^4 + 
        60*m3^2*msq^4*mU3^4 - 30*mQ3^2*msq^4*mU3^4 - 437*m3^4*mU3^6 - 
        38*m3^2*mQ3^2*mU3^6 - mQ3^4*mU3^6 - 180*m3^2*msq^2*mU3^6 + 
        30*msq^4*mU3^6 + 136*m3^2*mU3^8 + 5*mQ3^2*mU3^8 - mU3^10))/
      (3*(m3^2 - mU3^2)^4*(-mQ3^2 + mU3^2)) + 
     (8*(m3^4 - 34*m3^2*mU3^2 + 17*mU3^4)*(-2*mQ3^4*mU3^4 + 
        2*m3^6*(mQ3^2 + mU3^2) + 4*m3^2*mQ3^2*mU3^2*(mQ3^2 + mU3^2) - 
        m3^4*(mQ3^4 + 8*mQ3^2*mU3^2 + mU3^4)))/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^4))*Log[mQ3^2/MR^2] + a3^2*at*k^(3/2)*mt^2*
    ((8*(2*m3^2*mU3^2 - mU3^4))/(3*(m3^2 - mU3^2)^2) - 
     (48*(2*m3^2*mU3^2 - mU3^4)^2)/(m3^2 - mU3^2)^4 - 
     (16*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2)*
       (3*m3^4 - 4*m3^2*mU3^2 + 2*mU3^4))/(3*(m3^2 - mQ3^2)*
       (m3^2 - mU3^2)^3) + (4*(128*m3^10 + 44*m3^8*mQ3^2 - 2*m3^6*mQ3^4 + 
        9*m3^4*mQ3^6 - 60*m3^6*mQ3^2*msq^2 + 90*m3^4*mQ3^2*msq^4 - 
        556*m3^8*mU3^2 + 68*m3^6*mQ3^2*mU3^2 - 33*m3^4*mQ3^4*mU3^2 - 
        6*m3^2*mQ3^6*mU3^2 + 60*m3^6*msq^2*mU3^2 - 120*m3^4*mQ3^2*msq^2*
         mU3^2 - 90*m3^4*msq^4*mU3^2 - 60*m3^2*mQ3^2*msq^4*mU3^2 + 
        702*m3^6*mU3^4 - 141*m3^4*mQ3^2*mU3^4 + 36*m3^2*mQ3^4*mU3^4 - 
        3*mQ3^6*mU3^4 + 120*m3^4*msq^2*mU3^4 + 180*m3^2*mQ3^2*msq^2*mU3^4 + 
        60*m3^2*msq^4*mU3^4 - 30*mQ3^2*msq^4*mU3^4 - 347*m3^4*mU3^6 + 
        50*m3^2*mQ3^2*mU3^6 - mQ3^4*mU3^6 - 180*m3^2*msq^2*mU3^6 + 
        30*msq^4*mU3^6 + 48*m3^2*mU3^8 - 17*mQ3^2*mU3^8 + 21*mU3^10))/
      (3*(m3^2 - mU3^2)^4*(-mQ3^2 + mU3^2)) + 
     (8*(53*m3^10*mQ3^2 - 106*m3^8*mQ3^4 + 53*m3^6*mQ3^6 - 309*m3^10*mU3^2 + 
        711*m3^8*mQ3^2*mU3^2 - 557*m3^6*mQ3^4*mU3^2 + 152*m3^4*mQ3^6*mU3^2 - 
        9*m3^2*mQ3^8*mU3^2 - 90*m3^6*mQ3^2*msq^2*mU3^2 + 
        180*m3^4*mQ3^4*msq^2*mU3^2 - 90*m3^2*mQ3^6*msq^2*mU3^2 + 
        419*m3^8*mU3^4 - 1011*m3^6*mQ3^2*mU3^4 + 938*m3^4*mQ3^4*mU3^4 - 
        295*m3^2*mQ3^6*mU3^4 - 3*mQ3^8*mU3^4 + 90*m3^6*msq^2*mU3^4 - 
        210*m3^4*mQ3^2*msq^2*mU3^4 + 150*m3^2*mQ3^4*msq^2*mU3^4 - 
        30*mQ3^6*msq^2*mU3^4 - 21*m3^6*mU3^6 + 34*m3^4*mQ3^2*mU3^6 - 
        231*m3^2*mQ3^4*mU3^6 + 146*mQ3^6*mU3^6 + 30*m3^4*msq^2*mU3^6 - 
        60*m3^2*mQ3^2*msq^2*mU3^6 + 30*mQ3^4*msq^2*mU3^6 - 100*m3^4*mU3^8 + 
        288*m3^2*mQ3^2*mU3^8 - 140*mQ3^4*mU3^8 - 9*m3^2*mU3^10 - 
        3*mQ3^2*mU3^10))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*
       (-mQ3^2 + mU3^2)) + (8*(m3^4 - 34*m3^2*mU3^2 + 17*mU3^4)*
       (7*m3^6 - 2*mQ3^2*mU3^4 - m3^4*(6*mQ3^2 + 7*mU3^2) + 
        m3^2*(5*mQ3^2*mU3^2 + 3*mU3^4)))/(3*(m3^2 - mQ3^2)*
       (m3^2 - mU3^2)^4) - (8*(128*m3^10 + 44*m3^8*mQ3^2 - 2*m3^6*mQ3^4 + 
        9*m3^4*mQ3^6 - 60*m3^6*mQ3^2*msq^2 + 90*m3^4*mQ3^2*msq^4 - 
        556*m3^8*mU3^2 + 64*m3^6*mQ3^2*mU3^2 - 33*m3^4*mQ3^4*mU3^2 - 
        6*m3^2*mQ3^6*mU3^2 + 60*m3^6*msq^2*mU3^2 - 120*m3^4*mQ3^2*msq^2*
         mU3^2 - 90*m3^4*msq^4*mU3^2 - 60*m3^2*mQ3^2*msq^4*mU3^2 + 
        706*m3^6*mU3^4 - 51*m3^4*mQ3^2*mU3^4 + 36*m3^2*mQ3^4*mU3^4 - 
        3*mQ3^6*mU3^4 + 120*m3^4*msq^2*mU3^4 + 180*m3^2*mQ3^2*msq^2*mU3^4 + 
        60*m3^2*msq^4*mU3^4 - 30*mQ3^2*msq^4*mU3^4 - 437*m3^4*mU3^6 - 
        38*m3^2*mQ3^2*mU3^6 - mQ3^4*mU3^6 - 180*m3^2*msq^2*mU3^6 + 
        30*msq^4*mU3^6 + 136*m3^2*mU3^8 + 5*mQ3^2*mU3^8 - mU3^10)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      (3*(m3^2 - mU3^2)^4*(-mQ3^2 + mU3^2)))) + 
 Log[msq^2/MR^2]^2*(-24*a3^2*at*k^(3/2)*mt^2*
    ((-5*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 3*m3^2*msq^2 + 
        mQ3^2*msq^2))/(6*(-m3^2 + mQ3^2)^3) - 
     (5*(-2*msq^2 + 2*mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
        msq^2*mU3^2 + mU3^4))/(6*(-m3^2 + mU3^2)^3))*Log[mQ3^2/MR^2] - 
   24*a3^2*at*k^(3/2)*mt^2*
    ((-5*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 3*m3^2*msq^2 + 
        mQ3^2*msq^2))/(6*(-m3^2 + mQ3^2)^3) - 
     (5*(-2*msq^2 + 2*mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
        msq^2*mU3^2 + mU3^4))/(6*(-m3^2 + mU3^2)^3))*Log[mU3^2/MR^2] + 
   a3^2*at*k^(3/2)*mt^2*
    (24*((-5*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 
          3*m3^2*msq^2 + mQ3^2*msq^2))/(6*(-m3^2 + mQ3^2)^3) - 
       (5*(-2*msq^2 + 2*mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
          msq^2*mU3^2 + mU3^4))/(6*(-m3^2 + mU3^2)^3)) - 
     48*((-5*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 
          3*m3^2*msq^2 + mQ3^2*msq^2))/(6*(-m3^2 + mQ3^2)^3) - 
       (5*(-2*msq^2 + 2*mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
          msq^2*mU3^2 + mU3^4))/(6*(-m3^2 + mU3^2)^3))*
      (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))) + 
 Log[msq^2/MR^2]*((-80*a3^2*at*k^(3/2)*(3*m3^6 - 7*m3^4*mQ3^2 + 
      6*m3^2*mQ3^4 - 2*mQ3^6 - 6*m3^4*msq^2 - 18*m3^2*mQ3^2*msq^2 + 
      9*m3^2*msq^4 + 3*mQ3^2*msq^4)*mt^2*Log[mQ3^2/MR^2]^2)/
    (3*(m3^2 - mQ3^2)^3) - (80*a3^2*at*k^(3/2)*mt^2*
     (3*m3^6 - 6*m3^4*msq^2 + 9*m3^2*msq^4 - 7*m3^4*mU3^2 - 
      18*m3^2*msq^2*mU3^2 + 3*msq^4*mU3^2 + 6*m3^2*mU3^4 - 2*mU3^6)*
     Log[mU3^2/MR^2]^2)/(3*(m3^2 - mU3^2)^3) + 
   Log[mQ3^2/MR^2]*((160*a3^2*at*k^(3/2)*mt^2*(3*m3^12 - 11*m3^10*mQ3^2 + 
        3*m3^8*mQ3^4 - m3^6*mQ3^6 - 13*m3^10*mU3^2 + 45*m3^8*mQ3^2*mU3^2 - 
        21*m3^6*mQ3^4*mU3^2 + 7*m3^4*mQ3^6*mU3^2 + 6*m3^8*mU3^4 - 
        24*m3^6*mQ3^2*mU3^4 - 2*m3^6*mU3^6 + 8*m3^4*mQ3^2*mU3^6)*
       Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) + 
     a3^2*at*k^(3/2)*mt^2*((160*m3^2*mQ3^2)/(3*(m3^2 - mQ3^2)^2) - 
       (80*mQ3^4)/(3*(m3^2 - mQ3^2)^2) - 
       (80*(2*m3^4*mQ3^2 - 3*m3^2*mQ3^4 + mQ3^6 + 6*m3^4*msq^2 + 
          18*m3^2*mQ3^2*msq^2 - 9*m3^2*msq^4 - 3*mQ3^2*msq^4))/
        (3*(m3^2 - mQ3^2)^3) + (80*(11*m3^8 - 22*m3^6*mQ3^2 + 11*m3^4*mQ3^4 + 
          18*m3^6*msq^2 - 15*m3^4*mQ3^2*msq^2 + 9*m3^2*mQ3^4*msq^2 - 
          22*m3^6*mU3^2 + 44*m3^4*mQ3^2*mU3^2 - 22*m3^2*mQ3^4*mU3^2 - 
          15*m3^4*msq^2*mU3^2 - 12*m3^2*mQ3^2*msq^2*mU3^2 + 
          3*mQ3^4*msq^2*mU3^2 + 11*m3^4*mU3^4 - 22*m3^2*mQ3^2*mU3^4 + 
          11*mQ3^4*mU3^4 + 9*m3^2*msq^2*mU3^4 + 3*mQ3^2*msq^2*mU3^4))/
        (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
       (80*(7*m3^6 - 2*mQ3^4*mU3^2 - m3^4*(7*mQ3^2 + 6*mU3^2) + 
          m3^2*(3*mQ3^4 + 5*mQ3^2*mU3^2)))/(3*(m3^2 - mQ3^2)^2*
         (m3^2 - mU3^2)) + (160*(2*m3^4*msq^2 + 6*m3^2*mQ3^2*msq^2 - 
          3*m3^2*msq^4 - mQ3^2*msq^4)*(Log[MR^2] - 2*Log[mt] - 
          2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/(m3^2 - mQ3^2)^3)) + 
   Log[mU3^2/MR^2]*((160*a3^2*at*k^(3/2)*mt^2*(3*m3^12 - 13*m3^10*mQ3^2 + 
        6*m3^8*mQ3^4 - 2*m3^6*mQ3^6 - 11*m3^10*mU3^2 + 45*m3^8*mQ3^2*mU3^2 - 
        24*m3^6*mQ3^4*mU3^2 + 8*m3^4*mQ3^6*mU3^2 + 3*m3^8*mU3^4 - 
        21*m3^6*mQ3^2*mU3^4 - m3^6*mU3^6 + 7*m3^4*mQ3^2*mU3^6)*
       Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) + 
     a3^2*at*k^(3/2)*mt^2*((80*(2*m3^4*msq^2 + 6*m3^2*mQ3^2*msq^2 - 
          3*m3^2*msq^4 - mQ3^2*msq^4))/(m3^2 - mQ3^2)^3 + 
       (80*(2*m3^4*msq^2 - 3*m3^2*msq^4 + 6*m3^2*msq^2*mU3^2 - msq^4*mU3^2))/
        (m3^2 - mU3^2)^3 + (80*(-2*mQ3^4*mU3^4 + 2*m3^6*(mQ3^2 + mU3^2) + 
          4*m3^2*mQ3^2*mU3^2*(mQ3^2 + mU3^2) - m3^4*(mQ3^4 + 8*mQ3^2*mU3^2 + 
            mU3^4)))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*Log[mQ3^2/MR^2] + 
     a3^2*at*k^(3/2)*mt^2*((160*m3^2*mU3^2)/(3*(m3^2 - mU3^2)^2) - 
       (80*mU3^4)/(3*(m3^2 - mU3^2)^2) + 
       (80*(11*m3^8 - 22*m3^6*mQ3^2 + 11*m3^4*mQ3^4 + 18*m3^6*msq^2 - 
          15*m3^4*mQ3^2*msq^2 + 9*m3^2*mQ3^4*msq^2 - 22*m3^6*mU3^2 + 
          44*m3^4*mQ3^2*mU3^2 - 22*m3^2*mQ3^4*mU3^2 - 15*m3^4*msq^2*mU3^2 - 
          12*m3^2*mQ3^2*msq^2*mU3^2 + 3*mQ3^4*msq^2*mU3^2 + 11*m3^4*mU3^4 - 
          22*m3^2*mQ3^2*mU3^4 + 11*mQ3^4*mU3^4 + 9*m3^2*msq^2*mU3^4 + 
          3*mQ3^2*msq^2*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) - 
       (80*(6*m3^4*msq^2 - 9*m3^2*msq^4 + 2*m3^4*mU3^2 + 
          18*m3^2*msq^2*mU3^2 - 3*msq^4*mU3^2 - 3*m3^2*mU3^4 + mU3^6))/
        (3*(m3^2 - mU3^2)^3) + (80*(7*m3^6 - 2*mQ3^2*mU3^4 - 
          m3^4*(6*mQ3^2 + 7*mU3^2) + m3^2*(5*mQ3^2*mU3^2 + 3*mU3^4)))/
        (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2) + 
       (160*(2*m3^4*msq^2 - 3*m3^2*msq^4 + 6*m3^2*msq^2*mU3^2 - msq^4*mU3^2)*
         (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
        (m3^2 - mU3^2)^3)) + a3^2*at*k^(3/2)*mt^2*Log[m3^2/MR^2]*
    ((-80*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + m3^4*mU3^4))/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
     (8*(-40*m3^12 + 200*m3^10*mQ3^2 - 60*m3^8*mQ3^4 + 20*m3^6*mQ3^6 + 
        200*m3^10*mU3^2 - 840*m3^8*mQ3^2*mU3^2 + 420*m3^6*mQ3^4*mU3^2 - 
        140*m3^4*mQ3^6*mU3^2 - 60*m3^8*mU3^4 + 420*m3^6*mQ3^2*mU3^4 + 
        20*m3^6*mU3^6 - 140*m3^4*mQ3^2*mU3^6))/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^3) - (80*m3^4*(2*m3^6*(mQ3^2 + mU3^2) + 
        2*m3^2*(mQ3^2 - mU3^2)^2*(mQ3^2 + mU3^2) + 
        m3^4*(-4*mQ3^4 + 2*mQ3^2*mU3^2 - 4*mU3^4) + 
        mQ3^2*mU3^2*(mQ3^4 + mU3^4)))/(3*mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*
       (m3^2 - mU3^2)^2) + (160*(2*m3^12 - 8*m3^10*mQ3^2 + 3*m3^8*mQ3^4 - 
        m3^6*mQ3^6 - 8*m3^10*mU3^2 + 30*m3^8*mQ3^2*mU3^2 - 
        15*m3^6*mQ3^4*mU3^2 + 5*m3^4*mQ3^6*mU3^2 + 3*m3^8*mU3^4 - 
        15*m3^6*mQ3^2*mU3^4 - m3^6*mU3^6 + 5*m3^4*mQ3^2*mU3^6)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      ((m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3)) + a3^2*at*k^(3/2)*mt^2*
    ((80*(m3^4 - mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)) + 
     (80*(-2*m3^6*mQ3^2 + 2*m3^4*mQ3^4 - 2*m3^6*mU3^2 + 9*m3^4*mQ3^2*mU3^2 - 
        6*m3^2*mQ3^4*mU3^2 + 2*m3^4*mU3^4 - 6*m3^2*mQ3^2*mU3^4 + 
        3*mQ3^4*mU3^4))/(3*mQ3^2*(-m3^2 + mQ3^2)*mU3^2*(m3^2 - mU3^2)) - 
     (80*(12*m3^8 - 23*m3^6*mQ3^2 + 11*m3^4*mQ3^4 + 18*m3^6*msq^2 - 
        15*m3^4*mQ3^2*msq^2 + 9*m3^2*mQ3^4*msq^2 - 23*m3^6*mU3^2 + 
        44*m3^4*mQ3^2*mU3^2 - 21*m3^2*mQ3^4*mU3^2 - 15*m3^4*msq^2*mU3^2 - 
        12*m3^2*mQ3^2*msq^2*mU3^2 + 3*mQ3^4*msq^2*mU3^2 + 11*m3^4*mU3^4 - 
        21*m3^2*mQ3^2*mU3^4 + 10*mQ3^4*mU3^4 + 9*m3^2*msq^2*mU3^4 + 
        3*mQ3^2*msq^2*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) - 
     (160*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/3 + 
     (160*(11*m3^8 - 22*m3^6*mQ3^2 + 11*m3^4*mQ3^4 + 18*m3^6*msq^2 - 
        15*m3^4*mQ3^2*msq^2 + 9*m3^2*mQ3^4*msq^2 - 22*m3^6*mU3^2 + 
        44*m3^4*mQ3^2*mU3^2 - 22*m3^2*mQ3^4*mU3^2 - 15*m3^4*msq^2*mU3^2 - 
        12*m3^2*mQ3^2*msq^2*mU3^2 + 3*mQ3^4*msq^2*mU3^2 + 11*m3^4*mU3^4 - 
        22*m3^2*mQ3^2*mU3^4 + 11*mQ3^4*mU3^4 + 9*m3^2*msq^2*mU3^4 + 
        3*mQ3^2*msq^2*mU3^4)*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
        2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
     160*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]])^2)) + 
 a3^2*at*k^(3/2)*mt^2*Log[m3^2/MR^2]^2*
  ((-32*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + m3^4*mU3^4))/
    ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) - 
   (48*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + m3^4*mU3^4)^2)/
    ((m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^4) + 
   (8*(262*m3^16 - 878*m3^14*mQ3^2 + 885*m3^12*mQ3^4 - 320*m3^10*mQ3^6 + 
      49*m3^8*mQ3^8 - 878*m3^14*mU3^2 + 2972*m3^12*mQ3^2*mU3^2 - 
      2932*m3^10*mQ3^4*mU3^2 + 1004*m3^8*mQ3^6*mU3^2 - 158*m3^6*mQ3^8*mU3^2 + 
      885*m3^12*mU3^4 - 2932*m3^10*mQ3^2*mU3^4 + 2404*m3^8*mQ3^4*mU3^4 - 
      368*m3^6*mQ3^6*mU3^4 - m3^4*mQ3^8*mU3^4 - 320*m3^10*mU3^6 + 
      1004*m3^8*mQ3^2*mU3^6 - 368*m3^6*mQ3^4*mU3^6 - 452*m3^4*mQ3^6*mU3^6 + 
      144*m3^2*mQ3^8*mU3^6 + 49*m3^8*mU3^8 - 158*m3^6*mQ3^2*mU3^8 - 
      m3^4*mQ3^4*mU3^8 + 144*m3^2*mQ3^6*mU3^8 - 36*mQ3^8*mU3^8))/
    (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^4) - 
   (16*m3^4*(2*m3^6*(mQ3^2 + mU3^2) + 2*m3^2*(mQ3^2 - mU3^2)^2*
       (mQ3^2 + mU3^2) + m3^4*(-4*mQ3^4 + 2*mQ3^2*mU3^2 - 4*mU3^4) + 
      mQ3^2*mU3^2*(mQ3^4 + mU3^4))*
     (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
         m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)))/
    (mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*(m3^2 - mU3^2)^2) - 
   (16*(198*m3^16 - 726*m3^14*mQ3^2 + 721*m3^12*mQ3^4 - 232*m3^10*mQ3^6 + 
      27*m3^8*mQ3^8 - 726*m3^14*mU3^2 + 2700*m3^12*mQ3^2*mU3^2 - 
      2700*m3^10*mQ3^4*mU3^2 + 908*m3^8*mQ3^6*mU3^2 - 134*m3^6*mQ3^8*mU3^2 + 
      721*m3^12*mU3^4 - 2700*m3^10*mQ3^2*mU3^4 + 2240*m3^8*mQ3^4*mU3^4 - 
      320*m3^6*mQ3^6*mU3^4 - 13*m3^4*mQ3^8*mU3^4 - 232*m3^10*mU3^6 + 
      908*m3^8*mQ3^2*mU3^6 - 320*m3^6*mQ3^4*mU3^6 - 452*m3^4*mQ3^6*mU3^6 + 
      144*m3^2*mQ3^8*mU3^6 + 27*m3^8*mU3^8 - 134*m3^6*mQ3^2*mU3^8 - 
      13*m3^4*mQ3^4*mU3^8 + 144*m3^2*mQ3^6*mU3^8 - 36*mQ3^8*mU3^8)*
     (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
    (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^4)) + 
 at*Sqrt[k]*mt^2*(12*Log[MR^2] - 24*Log[mt] - 24*Log[Csc[beta]] - 
   24*Log[Sin[beta]]) + a3^2*at*k^(3/2)*mt^2*z3*
  (-128*Log[MR^2] + 256*Log[mt] + 256*Log[Csc[beta]] + 256*Log[Sin[beta]]) + 
 a3^2*at*k^(3/2)*mt^2*(dlatas2 - (128*m3^4)/(3*(m3^2 - mQ3^2)*
     (m3^2 - mU3^2)) + (128*mQ3^2*mU3^2)/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
   (16*(m3^4 - mQ3^2*mU3^2))/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
   (48*(m3^4 - mQ3^2*mU3^2)^2)/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) - 
   (16*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2)*
     (-2*m3^6*mQ3^2 + 2*m3^4*mQ3^4 - 2*m3^6*mU3^2 + 9*m3^4*mQ3^2*mU3^2 - 
      6*m3^2*mQ3^4*mU3^2 + 2*m3^4*mU3^4 - 6*m3^2*mQ3^2*mU3^4 + 
      3*mQ3^4*mU3^4))/(3*mQ3^2*(-m3^2 + mQ3^2)^2*mU3^2*(m3^2 - mU3^2)^2) + 
   (4*(291*m3^8 - 282*m3^6*mQ3^2 + 93*m3^4*mQ3^4 - 18*m3^2*mQ3^6 - 
      360*m3^6*msq^2 + 300*m3^4*mQ3^2*msq^2 - 180*m3^2*mQ3^4*msq^2 - 
      282*m3^6*mU3^2 - 112*m3^4*mQ3^2*mU3^2 + 208*m3^2*mQ3^4*mU3^2 - 
      6*mQ3^6*mU3^2 + 300*m3^4*msq^2*mU3^2 + 240*m3^2*mQ3^2*msq^2*mU3^2 - 
      60*mQ3^4*msq^2*mU3^2 + 93*m3^4*mU3^4 + 208*m3^2*mQ3^2*mU3^4 - 
      169*mQ3^4*mU3^4 - 180*m3^2*msq^2*mU3^4 - 60*mQ3^2*msq^2*mU3^4 - 
      18*m3^2*mU3^6 - 6*mQ3^2*mU3^6))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
   SMConstatas2 + (256*m3^4*Log[MR^2])/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
   (256*mQ3^2*mU3^2*Log[MR^2])/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
   (128*(m3^4 - mQ3^2*mU3^2)*(2*Log[at] - Log[k] + 4*Log[Csc[beta]]))/
    ((m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
   (2056*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/3 + 
   (32*(11*m3^4 - 3*m3^2*mQ3^2 - 3*m3^2*mU3^2 - 5*mQ3^2*mU3^2)*
     (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
    (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
   (8*(259*m3^8 - 270*m3^6*mQ3^2 + 93*m3^4*mQ3^4 - 18*m3^2*mQ3^6 - 
      360*m3^6*msq^2 + 300*m3^4*mQ3^2*msq^2 - 180*m3^2*mQ3^4*msq^2 - 
      270*m3^6*mU3^2 - 72*m3^4*mQ3^2*mU3^2 + 196*m3^2*mQ3^4*mU3^2 - 
      6*mQ3^6*mU3^2 + 300*m3^4*msq^2*mU3^2 + 240*m3^2*mQ3^2*msq^2*mU3^2 - 
      60*mQ3^4*msq^2*mU3^2 + 93*m3^4*mU3^4 + 196*m3^2*mQ3^2*mU3^4 - 
      177*mQ3^4*mU3^4 - 180*m3^2*msq^2*mU3^4 - 60*mQ3^2*msq^2*mU3^4 - 
      18*m3^2*mU3^6 - 6*mQ3^2*mU3^6)*(Log[MR^2] - 2*Log[mt] - 
      2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^2*
     (m3^2 - mU3^2)^2) + 160*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
      2*Log[Sin[beta]])^2 - (32*(11*m3^4 - 3*m3^2*mQ3^2 - 3*m3^2*mU3^2 - 
      5*mQ3^2*mU3^2)*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
       2*Log[Sin[beta]])^2)/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) + 
   736*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]])^3 - 
   (512*(m3^4 - mQ3^2*mU3^2)*(-Log[at]/2 + Log[k]/4 + Log[mt] + 
      Log[Sin[beta]]))/((m3^2 - mQ3^2)*(m3^2 - mU3^2))) + 
 a3*at*k*mt^2*((16*m3^4)/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
   (16*mQ3^2*mU3^2)/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) + 
   (16*(-2*m3^6*mQ3^2 + 2*m3^4*mQ3^4 - 2*m3^6*mU3^2 + 9*m3^4*mQ3^2*mU3^2 - 
      6*m3^2*mQ3^4*mU3^2 + 2*m3^4*mU3^4 - 6*m3^2*mQ3^2*mU3^4 + 
      3*mQ3^4*mU3^4))/(mQ3^2*(-m3^2 + mQ3^2)*mU3^2*(m3^2 - mU3^2)) - 
   32*Log[MR^2] - (32*m3^4*Log[MR^2])/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) + 
   (32*mQ3^2*mU3^2*Log[MR^2])/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) + 64*Log[mt] + 
   64*Log[Csc[beta]] + (16*(m3^4 - mQ3^2*mU3^2)*(2*Log[at] - Log[k] + 
      4*Log[Csc[beta]]))/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) + 
   96*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]])^2 + 
   64*Log[Sin[beta]] + (64*(m3^4 - mQ3^2*mU3^2)*(-Log[at]/2 + Log[k]/4 + 
      Log[mt] + Log[Sin[beta]]))/((m3^2 - mQ3^2)*(m3^2 - mU3^2))) + 
 Log[mQ3^2/MR^2]*(6*at*Sqrt[k]*mt^2 - 
   (8*a3^2*at*k^(3/2)*mt^2*(198*m3^16 - 814*m3^14*mQ3^2 + 877*m3^12*mQ3^4 - 
      344*m3^10*mQ3^6 + 55*m3^8*mQ3^8 - 638*m3^14*mU3^2 + 
      2700*m3^12*mQ3^2*mU3^2 - 2796*m3^10*mQ3^4*mU3^2 + 
      1004*m3^8*mQ3^6*mU3^2 - 158*m3^6*mQ3^8*mU3^2 + 565*m3^12*mU3^4 - 
      2604*m3^10*mQ3^2*mU3^4 + 2240*m3^8*mQ3^4*mU3^4 - 368*m3^6*mQ3^6*mU3^4 - 
      m3^4*mQ3^8*mU3^4 - 120*m3^10*mU3^6 + 812*m3^8*mQ3^2*mU3^6 - 
      272*m3^6*mQ3^4*mU3^6 - 452*m3^4*mQ3^6*mU3^6 + 144*m3^2*mQ3^8*mU3^6 - 
      m3^8*mU3^8 - 110*m3^6*mQ3^2*mU3^8 - 25*m3^4*mQ3^4*mU3^8 + 
      144*m3^2*mQ3^6*mU3^8 - 36*mQ3^8*mU3^8)*Log[m3^2/MR^2]^2)/
    (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^4) + 
   Log[m3^2/MR^2]*((32*a3*at*k*m3^4*mt^2)/(m3^2 - mQ3^2)^2 + 
     a3^2*at*k^(3/2)*mt^2*((16*m3^4*(2*m3^2 - mQ3^2 - mU3^2)*(mQ3^2 - mU3^2)*
         (-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2))/
        (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) + 
       (96*(2*m3^2*mQ3^2 - mQ3^4)*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 
          2*m3^6*mU3^2 + m3^4*mU3^4))/((m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^2) + 
       (8*(800*m3^12 - 1854*m3^10*mQ3^2 + 1410*m3^8*mQ3^4 - 345*m3^6*mQ3^6 + 
          9*m3^4*mQ3^8 - 180*m3^10*msq^2 + 240*m3^8*mQ3^2*msq^2 - 
          270*m3^6*mQ3^4*msq^2 + 90*m3^4*mQ3^6*msq^2 - 1854*m3^10*mU3^2 + 
          4056*m3^8*mQ3^2*mU3^2 - 2875*m3^6*mQ3^4*mU3^2 + 
          598*m3^4*mQ3^6*mU3^2 + 3*m3^2*mQ3^8*mU3^2 + 240*m3^8*msq^2*mU3^2 + 
          180*m3^6*mQ3^2*msq^2*mU3^2 - 90*m3^4*mQ3^4*msq^2*mU3^2 + 
          30*m3^2*mQ3^6*msq^2*mU3^2 + 1410*m3^8*mU3^4 - 2875*m3^6*mQ3^2*
           mU3^4 + 1930*m3^4*mQ3^4*mU3^4 - 369*m3^2*mQ3^6*mU3^4 - 
          270*m3^6*msq^2*mU3^4 - 90*m3^4*mQ3^2*msq^2*mU3^4 - 345*m3^6*mU3^6 + 
          598*m3^4*mQ3^2*mU3^6 - 369*m3^2*mQ3^4*mU3^6 + 60*mQ3^6*mU3^6 + 
          90*m3^4*msq^2*mU3^6 + 30*m3^2*mQ3^2*msq^2*mU3^6 + 9*m3^4*mU3^8 + 
          3*m3^2*mQ3^2*mU3^8))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) - 
       (16*(64*m3^16 - 180*m3^14*mQ3^2 + 122*m3^12*mQ3^4 - 31*m3^10*mQ3^6 + 
          30*m3^8*mQ3^8 - 7*m3^6*mQ3^10 - 204*m3^14*mU3^2 + 
          604*m3^12*mQ3^2*mU3^2 - 369*m3^10*mQ3^4*mU3^2 + 
          13*m3^8*mQ3^6*mU3^2 - 49*m3^6*mQ3^8*mU3^2 + 13*m3^4*mQ3^10*mU3^2 + 
          234*m3^12*mU3^4 - 779*m3^10*mQ3^2*mU3^4 + 503*m3^8*mQ3^4*mU3^4 - 
          13*m3^6*mQ3^6*mU3^4 + 61*m3^4*mQ3^8*mU3^4 - 18*m3^2*mQ3^10*mU3^4 - 
          101*m3^10*mU3^6 + 401*m3^8*mQ3^2*mU3^6 - 239*m3^6*mQ3^4*mU3^6 - 
          53*m3^4*mQ3^6*mU3^6 - 6*m3^2*mQ3^8*mU3^6 + 6*mQ3^10*mU3^6 + 
          13*m3^8*mU3^8 - 76*m3^6*mQ3^2*mU3^8 + 43*m3^4*mQ3^4*mU3^8 + 
          24*m3^2*mQ3^6*mU3^8 - 6*mQ3^8*mU3^8))/(3*(m3^2 - mQ3^2)^4*
         (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) - 
       (8*m3^4*(m3^4 - 34*m3^2*mQ3^2 + 17*mQ3^4)*(2*m3^6*(mQ3^2 + mU3^2) + 
          2*m3^2*(mQ3^2 - mU3^2)^2*(mQ3^2 + mU3^2) + 
          m3^4*(-4*mQ3^4 + 2*mQ3^2*mU3^2 - 4*mU3^4) + mQ3^2*mU3^2*
           (mQ3^4 + mU3^4)))/(3*mQ3^2*(m3^2 - mQ3^2)^4*mU3^2*
         (m3^2 - mU3^2)^2) - 48*((-4*m3^2*mQ3^2)/(3*(m3^2 - mQ3^2)^2) + 
         (2*mQ3^4)/(3*(m3^2 - mQ3^2)^2) + (2*m3^8 - 2*m3^6*mQ3^2 + 
           m3^4*mQ3^4 - 2*m3^6*mU3^2 + m3^4*mU3^4)/(18*(m3^2 - mQ3^2)^2*
           (m3^2 - mU3^2)^2)) + (16*(7*m3^6 - 2*mQ3^4*mU3^2 - 
          m3^4*(7*mQ3^2 + 6*mU3^2) + m3^2*(3*mQ3^4 + 5*mQ3^2*mU3^2))*
         (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
             m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)))/
        ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)) + 
       (16*(128*m3^16 - 358*m3^14*mQ3^2 + 134*m3^12*mQ3^4 + 125*m3^10*mQ3^6 - 
          72*m3^8*mQ3^8 + 19*m3^6*mQ3^10 - 410*m3^14*mU3^2 + 
          1314*m3^12*mQ3^2*mU3^2 - 683*m3^10*mQ3^4*mU3^2 - 
          191*m3^8*mQ3^6*mU3^2 + 97*m3^6*mQ3^8*mU3^2 - 31*m3^4*mQ3^10*mU3^2 + 
          472*m3^12*mU3^4 - 1797*m3^10*mQ3^2*mU3^4 + 1157*m3^8*mQ3^4*mU3^4 - 
          11*m3^6*mQ3^6*mU3^4 + 35*m3^4*mQ3^8*mU3^4 - 205*m3^10*mU3^6 + 
          999*m3^8*mQ3^2*mU3^6 - 655*m3^6*mQ3^4*mU3^6 - 43*m3^4*mQ3^6*mU3^6 + 
          27*m3^8*mU3^8 - 218*m3^6*mQ3^2*mU3^8 + 167*m3^4*mQ3^4*mU3^8)*
         (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
        (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)))) + 
   a3^2*at*k^(3/2)*mt^2*((-352*m3^2*mQ3^2)/(3*(m3^2 - mQ3^2)^2) + 
     (176*mQ3^4)/(3*(m3^2 - mQ3^2)^2) + (8*m3^4)/(3*(m3^2 - mQ3^2)*
       (m3^2 - mU3^2)) - (8*mQ3^2*mU3^2)/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
     (96*(2*m3^2*mQ3^2 - mQ3^4)*(m3^4 - mQ3^2*mU3^2))/
      ((m3^2 - mQ3^2)^3*(m3^2 - mU3^2)) + 
     (8*(m3^4 - 34*m3^2*mQ3^2 + 17*mQ3^4)*(-2*m3^6*mQ3^2 + 2*m3^4*mQ3^4 - 
        2*m3^6*mU3^2 + 9*m3^4*mQ3^2*mU3^2 - 6*m3^2*mQ3^4*mU3^2 + 
        2*m3^4*mU3^4 - 6*m3^2*mQ3^2*mU3^4 + 3*mQ3^4*mU3^4))/
      (3*mQ3^2*(-m3^2 + mQ3^2)^3*mU3^2*(m3^2 - mU3^2)) - 
     (4*(259*m3^8 - 270*m3^6*mQ3^2 + 93*m3^4*mQ3^4 - 18*m3^2*mQ3^6 - 
        360*m3^6*msq^2 + 300*m3^4*mQ3^2*msq^2 - 180*m3^2*mQ3^4*msq^2 - 
        270*m3^6*mU3^2 - 72*m3^4*mQ3^2*mU3^2 + 196*m3^2*mQ3^4*mU3^2 - 
        6*mQ3^6*mU3^2 + 300*m3^4*msq^2*mU3^2 + 240*m3^2*mQ3^2*msq^2*mU3^2 - 
        60*mQ3^4*msq^2*mU3^2 + 93*m3^4*mU3^4 + 196*m3^2*mQ3^2*mU3^4 - 
        177*mQ3^4*mU3^4 - 180*m3^2*msq^2*mU3^4 - 60*mQ3^2*msq^2*mU3^4 - 
        18*m3^2*mU3^6 - 6*mQ3^2*mU3^6))/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^2) + (8*(308*m3^10*mQ3^2 - 365*m3^8*mQ3^4 - 
        18*m3^6*mQ3^6 + 106*m3^4*mQ3^8 + 9*m3^2*mQ3^10 - 
        90*m3^6*mQ3^4*msq^2 - 30*m3^4*mQ3^6*msq^2 - 52*m3^10*mU3^2 - 
        764*m3^8*mQ3^2*mU3^2 + 985*m3^6*mQ3^4*mU3^2 - 25*m3^4*mQ3^6*mU3^2 - 
        279*m3^2*mQ3^8*mU3^2 + 3*mQ3^10*mU3^2 + 90*m3^6*mQ3^2*msq^2*mU3^2 + 
        210*m3^4*mQ3^4*msq^2*mU3^2 + 60*m3^2*mQ3^6*msq^2*mU3^2 + 
        105*m3^8*mU3^4 + 622*m3^6*mQ3^2*mU3^4 - 942*m3^4*mQ3^4*mU3^4 + 
        246*m3^2*mQ3^6*mU3^4 + 125*mQ3^8*mU3^4 - 180*m3^4*mQ3^2*msq^2*mU3^4 - 
        150*m3^2*mQ3^4*msq^2*mU3^4 - 30*mQ3^6*msq^2*mU3^4 - 53*m3^6*mU3^6 - 
        163*m3^4*mQ3^2*mU3^6 + 271*m3^2*mQ3^4*mU3^6 - 131*mQ3^6*mU3^6 + 
        90*m3^2*mQ3^2*msq^2*mU3^6 + 30*mQ3^4*msq^2*mU3^6 + 
        9*m3^2*mQ3^2*mU3^8 + 3*mQ3^4*mU3^8))/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) + 
     (16*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2)*
       (7*m3^6 - 2*mQ3^4*mU3^2 - m3^4*(7*mQ3^2 + 6*mU3^2) + 
        m3^2*(3*mQ3^4 + 5*mQ3^2*mU3^2)))/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^2) + (512*m3^2*mQ3^2*Log[MR^2])/(m3^2 - mQ3^2)^2 - 
     (256*mQ3^4*Log[MR^2])/(m3^2 - mQ3^2)^2 - 
     (128*(2*m3^2*mQ3^2 - mQ3^4)*(2*Log[at] - Log[k] + 4*Log[Csc[beta]]))/
      (m3^2 - mQ3^2)^2 + (128*(2*m3^2*mQ3^2 - mQ3^4)*(2*Log[at] - Log[k] - 
        4*Log[mt] - 4*Log[Sin[beta]]))/(m3^2 - mQ3^2)^2 - 
     (16*(m3^4 - 34*m3^2*mQ3^2 + 17*mQ3^4)*(Log[MR^2] - 2*Log[mt] - 
        2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^2) + 
     (16*(-309*m3^10*mQ3^2 + 419*m3^8*mQ3^4 - 21*m3^6*mQ3^6 - 
        100*m3^4*mQ3^8 - 9*m3^2*mQ3^10 + 90*m3^6*mQ3^4*msq^2 + 
        30*m3^4*mQ3^6*msq^2 + 53*m3^10*mU3^2 + 711*m3^8*mQ3^2*mU3^2 - 
        1011*m3^6*mQ3^4*mU3^2 + 34*m3^4*mQ3^6*mU3^2 + 288*m3^2*mQ3^8*mU3^2 - 
        3*mQ3^10*mU3^2 - 90*m3^6*mQ3^2*msq^2*mU3^2 - 210*m3^4*mQ3^4*msq^2*
         mU3^2 - 60*m3^2*mQ3^6*msq^2*mU3^2 - 106*m3^8*mU3^4 - 
        557*m3^6*mQ3^2*mU3^4 + 938*m3^4*mQ3^4*mU3^4 - 231*m3^2*mQ3^6*mU3^4 - 
        140*mQ3^8*mU3^4 + 180*m3^4*mQ3^2*msq^2*mU3^4 + 150*m3^2*mQ3^4*msq^2*
         mU3^4 + 30*mQ3^6*msq^2*mU3^4 + 53*m3^6*mU3^6 + 
        152*m3^4*mQ3^2*mU3^6 - 295*m3^2*mQ3^4*mU3^6 + 146*mQ3^6*mU3^6 - 
        90*m3^2*mQ3^2*msq^2*mU3^6 - 30*mQ3^4*msq^2*mU3^6 - 
        9*m3^2*mQ3^2*mU3^8 - 3*mQ3^4*mU3^8)*(Log[MR^2] - 2*Log[mt] - 
        2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) + 
     (16*(m3^4 - 34*m3^2*mQ3^2 + 17*mQ3^4)*(Log[MR^2] - 2*Log[mt] - 
         2*Log[Csc[beta]] - 2*Log[Sin[beta]])^2)/(m3^2 - mQ3^2)^2) + 
   (32*a3*at*k*mt^2*(3*m3^4 - 2*m3^2*mQ3^2 + mQ3^4 - 2*m3^2*mQ3^2*Log[MR^2] + 
      mQ3^4*Log[MR^2] + 4*m3^2*mQ3^2*Log[mt] - 2*mQ3^4*Log[mt] + 
      4*m3^2*mQ3^2*Log[Csc[beta]] - 2*mQ3^4*Log[Csc[beta]] + 
      4*m3^2*mQ3^2*Log[Sin[beta]] - 2*mQ3^4*Log[Sin[beta]]))/
    (m3^2 - mQ3^2)^2) + Log[m3^2/MR^2]*
  (a3*at*k*mt^2*((-16*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
        m3^4*mU3^4))/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) - 
     (16*m3^4*(2*m3^6*(mQ3^2 + mU3^2) + 2*m3^2*(mQ3^2 - mU3^2)^2*
         (mQ3^2 + mU3^2) + m3^4*(-4*mQ3^4 + 2*mQ3^2*mU3^2 - 4*mU3^4) + 
        mQ3^2*mU3^2*(mQ3^4 + mU3^4)))/(mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*
       (m3^2 - mU3^2)^2) + (32*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 
        2*m3^6*mU3^2 + m3^4*mU3^4)*Log[MR^2])/((m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^2) - (16*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 
        2*m3^6*mU3^2 + m3^4*mU3^4)*(2*Log[at] - Log[k] + 4*Log[Csc[beta]]))/
      ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) - 
     (64*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + m3^4*mU3^4)*
       (-Log[at]/2 + Log[k]/4 + Log[mt] + Log[Sin[beta]]))/
      ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)) + a3^2*at*k^(3/2)*mt^2*
    ((128*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + m3^4*mU3^4))/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
     (96*(m3^4 - mQ3^2*mU3^2)*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 
        2*m3^6*mU3^2 + m3^4*mU3^4))/((m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) + 
     (16*(4*m3^8 - 4*m3^6*mQ3^2 + m3^4*mQ3^4 - 4*m3^6*mU3^2 + 
        2*m3^2*mQ3^4*mU3^2 + m3^4*mU3^4 + 2*m3^2*mQ3^2*mU3^4 - 
        2*mQ3^4*mU3^4))/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
     (8*(-864*m3^12 + 1942*m3^10*mQ3^2 - 1460*m3^8*mQ3^4 + 351*m3^6*mQ3^6 - 
        9*m3^4*mQ3^8 + 180*m3^10*msq^2 - 240*m3^8*mQ3^2*msq^2 + 
        270*m3^6*mQ3^4*msq^2 - 90*m3^4*mQ3^6*msq^2 + 1942*m3^10*mU3^2 - 
        4088*m3^8*mQ3^2*mU3^2 + 2853*m3^6*mQ3^4*mU3^2 - 
        572*m3^4*mQ3^6*mU3^2 - 3*m3^2*mQ3^8*mU3^2 - 240*m3^8*msq^2*mU3^2 - 
        180*m3^6*mQ3^2*msq^2*mU3^2 + 90*m3^4*mQ3^4*msq^2*mU3^2 - 
        30*m3^2*mQ3^6*msq^2*mU3^2 - 1460*m3^8*mU3^4 + 2853*m3^6*mQ3^2*mU3^4 - 
        1894*m3^4*mQ3^4*mU3^4 + 345*m3^2*mQ3^6*mU3^4 + 270*m3^6*msq^2*mU3^4 + 
        90*m3^4*mQ3^2*msq^2*mU3^4 + 351*m3^6*mU3^6 - 572*m3^4*mQ3^2*mU3^6 + 
        345*m3^2*mQ3^4*mU3^6 - 48*mQ3^6*mU3^6 - 90*m3^4*msq^2*mU3^6 - 
        30*m3^2*mQ3^2*msq^2*mU3^6 - 9*m3^4*mU3^8 - 3*m3^2*mQ3^2*mU3^8))/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) - 
     (16*m3^4*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2)*
       (2*m3^6*(mQ3^2 + mU3^2) + 2*m3^2*(mQ3^2 - mU3^2)^2*(mQ3^2 + mU3^2) + 
        m3^4*(-4*mQ3^4 + 2*mQ3^2*mU3^2 - 4*mU3^4) + 
        mQ3^2*mU3^2*(mQ3^4 + mU3^4)))/(3*mQ3^2*(m3^2 - mQ3^2)^3*mU3^2*
       (m3^2 - mU3^2)^3) + (16*(-2*m3^6*mQ3^2 + 2*m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
        9*m3^4*mQ3^2*mU3^2 - 6*m3^2*mQ3^4*mU3^2 + 2*m3^4*mU3^4 - 
        6*m3^2*mQ3^2*mU3^4 + 3*mQ3^4*mU3^4)*
       (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
           m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)))/
      (mQ3^2*(-m3^2 + mQ3^2)*mU3^2*(m3^2 - mU3^2)) - 
     (256*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + m3^4*mU3^4)*
       Log[MR^2])/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
     (128*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + m3^4*mU3^4)*
       (2*Log[at] - Log[k] + 4*Log[Csc[beta]]))/((m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^2) + (16*(800*m3^12 - 1854*m3^10*mQ3^2 + 
        1410*m3^8*mQ3^4 - 345*m3^6*mQ3^6 + 9*m3^4*mQ3^8 - 180*m3^10*msq^2 + 
        240*m3^8*mQ3^2*msq^2 - 270*m3^6*mQ3^4*msq^2 + 90*m3^4*mQ3^6*msq^2 - 
        1854*m3^10*mU3^2 + 4056*m3^8*mQ3^2*mU3^2 - 2875*m3^6*mQ3^4*mU3^2 + 
        598*m3^4*mQ3^6*mU3^2 + 3*m3^2*mQ3^8*mU3^2 + 240*m3^8*msq^2*mU3^2 + 
        180*m3^6*mQ3^2*msq^2*mU3^2 - 90*m3^4*mQ3^4*msq^2*mU3^2 + 
        30*m3^2*mQ3^6*msq^2*mU3^2 + 1410*m3^8*mU3^4 - 2875*m3^6*mQ3^2*mU3^4 + 
        1930*m3^4*mQ3^4*mU3^4 - 369*m3^2*mQ3^6*mU3^4 - 270*m3^6*msq^2*mU3^4 - 
        90*m3^4*mQ3^2*msq^2*mU3^4 - 345*m3^6*mU3^6 + 598*m3^4*mQ3^2*mU3^6 - 
        369*m3^2*mQ3^4*mU3^6 + 60*mQ3^6*mU3^6 + 90*m3^4*msq^2*mU3^6 + 
        30*m3^2*mQ3^2*msq^2*mU3^6 + 9*m3^4*mU3^8 + 3*m3^2*mQ3^2*mU3^8)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) + 
     64*(-1 - (4*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
          m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*
      (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]) - 
     192*(-1 - (4*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
          m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*
      (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]])^2 + 
     (512*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + m3^4*mU3^4)*
       (-Log[at]/2 + Log[k]/4 + Log[mt] + Log[Sin[beta]]))/
      ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))) + 
 Log[mU3^2/MR^2]*(6*at*Sqrt[k]*mt^2 - 
   (8*a3^2*at*k^(3/2)*mt^2*(198*m3^16 - 638*m3^14*mQ3^2 + 565*m3^12*mQ3^4 - 
      120*m3^10*mQ3^6 - m3^8*mQ3^8 - 814*m3^14*mU3^2 + 
      2700*m3^12*mQ3^2*mU3^2 - 2604*m3^10*mQ3^4*mU3^2 + 
      812*m3^8*mQ3^6*mU3^2 - 110*m3^6*mQ3^8*mU3^2 + 877*m3^12*mU3^4 - 
      2796*m3^10*mQ3^2*mU3^4 + 2240*m3^8*mQ3^4*mU3^4 - 272*m3^6*mQ3^6*mU3^4 - 
      25*m3^4*mQ3^8*mU3^4 - 344*m3^10*mU3^6 + 1004*m3^8*mQ3^2*mU3^6 - 
      368*m3^6*mQ3^4*mU3^6 - 452*m3^4*mQ3^6*mU3^6 + 144*m3^2*mQ3^8*mU3^6 + 
      55*m3^8*mU3^8 - 158*m3^6*mQ3^2*mU3^8 - m3^4*mQ3^4*mU3^8 + 
      144*m3^2*mQ3^6*mU3^8 - 36*mQ3^8*mU3^8)*Log[m3^2/MR^2]^2)/
    (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^4) - 
   (4*a3^2*at*k^(3/2)*mt^2*(128*m3^16 - 550*m3^14*mQ3^2 + 670*m3^12*mQ3^4 - 
      247*m3^10*mQ3^6 - 64*m3^8*mQ3^8 + 100*m3^6*mQ3^10 - 38*m3^4*mQ3^12 + 
      9*m3^2*mQ3^14 + 60*m3^12*mQ3^2*msq^2 + 120*m3^10*mQ3^4*msq^2 - 
      180*m3^8*mQ3^6*msq^2 - 90*m3^10*mQ3^2*msq^4 + 60*m3^8*mQ3^4*msq^4 + 
      30*m3^6*mQ3^6*msq^4 - 346*m3^14*mU3^2 + 1542*m3^12*mQ3^2*mU3^2 - 
      1619*m3^10*mQ3^4*mU3^2 + 199*m3^8*mQ3^6*mU3^2 + 324*m3^6*mQ3^8*mU3^2 - 
      88*m3^4*mQ3^10*mU3^2 - 47*m3^2*mQ3^12*mU3^2 + 3*mQ3^14*mU3^2 - 
      60*m3^12*msq^2*mU3^2 - 300*m3^10*mQ3^2*msq^2*mU3^2 - 
      180*m3^8*mQ3^4*msq^2*mU3^2 + 540*m3^6*mQ3^6*msq^2*mU3^2 + 
      90*m3^10*msq^4*mU3^2 + 210*m3^8*mQ3^2*msq^4*mU3^2 - 
      210*m3^6*mQ3^4*msq^4*mU3^2 - 90*m3^4*mQ3^6*msq^4*mU3^2 + 
      476*m3^12*mU3^4 - 2307*m3^10*mQ3^2*mU3^4 + 2585*m3^8*mQ3^4*mU3^4 - 
      386*m3^6*mQ3^6*mU3^4 - 638*m3^4*mQ3^8*mU3^4 + 317*m3^2*mQ3^10*mU3^4 + 
      mQ3^12*mU3^4 + 180*m3^10*msq^2*mU3^4 + 540*m3^8*mQ3^2*msq^2*mU3^4 - 
      180*m3^6*mQ3^4*msq^2*mU3^4 - 540*m3^4*mQ3^6*msq^2*mU3^4 - 
      270*m3^8*msq^4*mU3^4 - 90*m3^6*mQ3^2*msq^4*mU3^4 + 
      270*m3^4*mQ3^4*msq^4*mU3^4 + 90*m3^2*mQ3^6*msq^4*mU3^4 - 
      307*m3^10*mU3^6 + 1699*m3^8*mQ3^2*mU3^6 - 2306*m3^6*mQ3^4*mU3^6 + 
      1002*m3^4*mQ3^6*mU3^6 - 19*m3^2*mQ3^8*mU3^6 - 101*mQ3^10*mU3^6 - 
      180*m3^8*msq^2*mU3^6 - 420*m3^6*mQ3^2*msq^2*mU3^6 + 
      420*m3^4*mQ3^4*msq^2*mU3^6 + 180*m3^2*mQ3^6*msq^2*mU3^6 + 
      270*m3^6*msq^4*mU3^6 - 90*m3^4*mQ3^2*msq^4*mU3^6 - 
      150*m3^2*mQ3^4*msq^4*mU3^6 - 30*mQ3^6*msq^4*mU3^6 + 61*m3^8*mU3^8 - 
      420*m3^6*mQ3^2*mU3^8 + 658*m3^4*mQ3^4*mU3^8 - 388*m3^2*mQ3^6*mU3^8 + 
      97*mQ3^8*mU3^8 + 60*m3^6*msq^2*mU3^8 + 120*m3^4*mQ3^2*msq^2*mU3^8 - 
      180*m3^2*mQ3^4*msq^2*mU3^8 - 90*m3^4*msq^4*mU3^8 + 
      60*m3^2*mQ3^2*msq^4*mU3^8 + 30*mQ3^4*msq^4*mU3^8)*Log[mQ3^2/MR^2]^2)/
    (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) + 
   Log[m3^2/MR^2]*((32*a3*at*k*m3^4*mt^2)/(m3^2 - mU3^2)^2 + 
     a3^2*at*k^(3/2)*mt^2*((-16*m3^4*(2*m3^2 - mQ3^2 - mU3^2)*(mQ3^2 - mU3^2)*
         (-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2))/
        (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) + 
       (96*(2*m3^2*mU3^2 - mU3^4)*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 
          2*m3^6*mU3^2 + m3^4*mU3^4))/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^4) + 
       (8*(800*m3^12 - 1854*m3^10*mQ3^2 + 1410*m3^8*mQ3^4 - 345*m3^6*mQ3^6 + 
          9*m3^4*mQ3^8 - 180*m3^10*msq^2 + 240*m3^8*mQ3^2*msq^2 - 
          270*m3^6*mQ3^4*msq^2 + 90*m3^4*mQ3^6*msq^2 - 1854*m3^10*mU3^2 + 
          4056*m3^8*mQ3^2*mU3^2 - 2875*m3^6*mQ3^4*mU3^2 + 
          598*m3^4*mQ3^6*mU3^2 + 3*m3^2*mQ3^8*mU3^2 + 240*m3^8*msq^2*mU3^2 + 
          180*m3^6*mQ3^2*msq^2*mU3^2 - 90*m3^4*mQ3^4*msq^2*mU3^2 + 
          30*m3^2*mQ3^6*msq^2*mU3^2 + 1410*m3^8*mU3^4 - 2875*m3^6*mQ3^2*
           mU3^4 + 1930*m3^4*mQ3^4*mU3^4 - 369*m3^2*mQ3^6*mU3^4 - 
          270*m3^6*msq^2*mU3^4 - 90*m3^4*mQ3^2*msq^2*mU3^4 - 345*m3^6*mU3^6 + 
          598*m3^4*mQ3^2*mU3^6 - 369*m3^2*mQ3^4*mU3^6 + 60*mQ3^6*mU3^6 + 
          90*m3^4*msq^2*mU3^6 + 30*m3^2*mQ3^2*msq^2*mU3^6 + 9*m3^4*mU3^8 + 
          3*m3^2*mQ3^2*mU3^8))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) + 
       (16*(64*m3^16 - 204*m3^14*mQ3^2 + 234*m3^12*mQ3^4 - 101*m3^10*mQ3^6 + 
          13*m3^8*mQ3^8 - 180*m3^14*mU3^2 + 604*m3^12*mQ3^2*mU3^2 - 
          779*m3^10*mQ3^4*mU3^2 + 401*m3^8*mQ3^6*mU3^2 - 
          76*m3^6*mQ3^8*mU3^2 + 122*m3^12*mU3^4 - 369*m3^10*mQ3^2*mU3^4 + 
          503*m3^8*mQ3^4*mU3^4 - 239*m3^6*mQ3^6*mU3^4 + 43*m3^4*mQ3^8*mU3^4 - 
          31*m3^10*mU3^6 + 13*m3^8*mQ3^2*mU3^6 - 13*m3^6*mQ3^4*mU3^6 - 
          53*m3^4*mQ3^6*mU3^6 + 24*m3^2*mQ3^8*mU3^6 + 30*m3^8*mU3^8 - 
          49*m3^6*mQ3^2*mU3^8 + 61*m3^4*mQ3^4*mU3^8 - 6*m3^2*mQ3^6*mU3^8 - 
          6*mQ3^8*mU3^8 - 7*m3^6*mU3^10 + 13*m3^4*mQ3^2*mU3^10 - 
          18*m3^2*mQ3^4*mU3^10 + 6*mQ3^6*mU3^10))/(3*(m3^2 - mQ3^2)^3*
         (m3^2 - mU3^2)^4*(mQ3^2 - mU3^2)) - 
       (8*m3^4*(m3^4 - 34*m3^2*mU3^2 + 17*mU3^4)*(2*m3^6*(mQ3^2 + mU3^2) + 
          2*m3^2*(mQ3^2 - mU3^2)^2*(mQ3^2 + mU3^2) + 
          m3^4*(-4*mQ3^4 + 2*mQ3^2*mU3^2 - 4*mU3^4) + mQ3^2*mU3^2*
           (mQ3^4 + mU3^4)))/(3*mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*
         (m3^2 - mU3^2)^4) - 48*((-4*m3^2*mU3^2)/(3*(m3^2 - mU3^2)^2) + 
         (2*mU3^4)/(3*(m3^2 - mU3^2)^2) + (2*m3^8 - 2*m3^6*mQ3^2 + 
           m3^4*mQ3^4 - 2*m3^6*mU3^2 + m3^4*mU3^4)/(18*(m3^2 - mQ3^2)^2*
           (m3^2 - mU3^2)^2)) + (16*(7*m3^6 - 2*mQ3^2*mU3^4 - 
          m3^4*(6*mQ3^2 + 7*mU3^2) + m3^2*(5*mQ3^2*mU3^2 + 3*mU3^4))*
         (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
             m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)))/
        ((m3^2 - mQ3^2)*(m3^2 - mU3^2)^2) - 
       (16*(128*m3^16 - 410*m3^14*mQ3^2 + 472*m3^12*mQ3^4 - 205*m3^10*mQ3^6 + 
          27*m3^8*mQ3^8 - 358*m3^14*mU3^2 + 1314*m3^12*mQ3^2*mU3^2 - 
          1797*m3^10*mQ3^4*mU3^2 + 999*m3^8*mQ3^6*mU3^2 - 
          218*m3^6*mQ3^8*mU3^2 + 134*m3^12*mU3^4 - 683*m3^10*mQ3^2*mU3^4 + 
          1157*m3^8*mQ3^4*mU3^4 - 655*m3^6*mQ3^6*mU3^4 + 
          167*m3^4*mQ3^8*mU3^4 + 125*m3^10*mU3^6 - 191*m3^8*mQ3^2*mU3^6 - 
          11*m3^6*mQ3^4*mU3^6 - 43*m3^4*mQ3^6*mU3^6 - 72*m3^8*mU3^8 + 
          97*m3^6*mQ3^2*mU3^8 + 35*m3^4*mQ3^4*mU3^8 + 19*m3^6*mU3^10 - 
          31*m3^4*mQ3^2*mU3^10)*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
          2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^4*
         (mQ3^2 - mU3^2)))) + a3^2*at*k^(3/2)*mt^2*
    ((-352*m3^2*mU3^2)/(3*(m3^2 - mU3^2)^2) + 
     (176*mU3^4)/(3*(m3^2 - mU3^2)^2) + (8*m3^4)/(3*(m3^2 - mQ3^2)*
       (m3^2 - mU3^2)) - (8*mQ3^2*mU3^2)/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
     (96*(m3^4 - mQ3^2*mU3^2)*(2*m3^2*mU3^2 - mU3^4))/
      ((m3^2 - mQ3^2)*(m3^2 - mU3^2)^3) + 
     (8*(m3^4 - 34*m3^2*mU3^2 + 17*mU3^4)*(-2*m3^6*mQ3^2 + 2*m3^4*mQ3^4 - 
        2*m3^6*mU3^2 + 9*m3^4*mQ3^2*mU3^2 - 6*m3^2*mQ3^4*mU3^2 + 
        2*m3^4*mU3^4 - 6*m3^2*mQ3^2*mU3^4 + 3*mQ3^4*mU3^4))/
      (3*mQ3^2*(-m3^2 + mQ3^2)*mU3^2*(m3^2 - mU3^2)^3) - 
     (4*(259*m3^8 - 270*m3^6*mQ3^2 + 93*m3^4*mQ3^4 - 18*m3^2*mQ3^6 - 
        360*m3^6*msq^2 + 300*m3^4*mQ3^2*msq^2 - 180*m3^2*mQ3^4*msq^2 - 
        270*m3^6*mU3^2 - 72*m3^4*mQ3^2*mU3^2 + 196*m3^2*mQ3^4*mU3^2 - 
        6*mQ3^6*mU3^2 + 300*m3^4*msq^2*mU3^2 + 240*m3^2*mQ3^2*msq^2*mU3^2 - 
        60*mQ3^4*msq^2*mU3^2 + 93*m3^4*mU3^4 + 196*m3^2*mQ3^2*mU3^4 - 
        177*mQ3^4*mU3^4 - 180*m3^2*msq^2*mU3^4 - 60*mQ3^2*msq^2*mU3^4 - 
        18*m3^2*mU3^6 - 6*mQ3^2*mU3^6))/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^2) + (8*(52*m3^10*mQ3^2 - 105*m3^8*mQ3^4 + 
        53*m3^6*mQ3^6 - 308*m3^10*mU3^2 + 764*m3^8*mQ3^2*mU3^2 - 
        622*m3^6*mQ3^4*mU3^2 + 163*m3^4*mQ3^6*mU3^2 - 9*m3^2*mQ3^8*mU3^2 - 
        90*m3^6*mQ3^2*msq^2*mU3^2 + 180*m3^4*mQ3^4*msq^2*mU3^2 - 
        90*m3^2*mQ3^6*msq^2*mU3^2 + 365*m3^8*mU3^4 - 985*m3^6*mQ3^2*mU3^4 + 
        942*m3^4*mQ3^4*mU3^4 - 271*m3^2*mQ3^6*mU3^4 - 3*mQ3^8*mU3^4 + 
        90*m3^6*msq^2*mU3^4 - 210*m3^4*mQ3^2*msq^2*mU3^4 + 
        150*m3^2*mQ3^4*msq^2*mU3^4 - 30*mQ3^6*msq^2*mU3^4 + 18*m3^6*mU3^6 + 
        25*m3^4*mQ3^2*mU3^6 - 246*m3^2*mQ3^4*mU3^6 + 131*mQ3^6*mU3^6 + 
        30*m3^4*msq^2*mU3^6 - 60*m3^2*mQ3^2*msq^2*mU3^6 + 
        30*mQ3^4*msq^2*mU3^6 - 106*m3^4*mU3^8 + 279*m3^2*mQ3^2*mU3^8 - 
        125*mQ3^4*mU3^8 - 9*m3^2*mU3^10 - 3*mQ3^2*mU3^10))/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) + 
     (16*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2)*
       (7*m3^6 - 2*mQ3^2*mU3^4 - m3^4*(6*mQ3^2 + 7*mU3^2) + 
        m3^2*(5*mQ3^2*mU3^2 + 3*mU3^4)))/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^3) + (512*m3^2*mU3^2*Log[MR^2])/(m3^2 - mU3^2)^2 - 
     (256*mU3^4*Log[MR^2])/(m3^2 - mU3^2)^2 - 
     (128*(2*m3^2*mU3^2 - mU3^4)*(2*Log[at] - Log[k] + 4*Log[Csc[beta]]))/
      (m3^2 - mU3^2)^2 + (128*(2*m3^2*mU3^2 - mU3^4)*(2*Log[at] - Log[k] - 
        4*Log[mt] - 4*Log[Sin[beta]]))/(m3^2 - mU3^2)^2 - 
     (16*(m3^4 - 34*m3^2*mU3^2 + 17*mU3^4)*(Log[MR^2] - 2*Log[mt] - 
        2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/(3*(m3^2 - mU3^2)^2) + 
     (16*(53*m3^10*mQ3^2 - 106*m3^8*mQ3^4 + 53*m3^6*mQ3^6 - 309*m3^10*mU3^2 + 
        711*m3^8*mQ3^2*mU3^2 - 557*m3^6*mQ3^4*mU3^2 + 152*m3^4*mQ3^6*mU3^2 - 
        9*m3^2*mQ3^8*mU3^2 - 90*m3^6*mQ3^2*msq^2*mU3^2 + 
        180*m3^4*mQ3^4*msq^2*mU3^2 - 90*m3^2*mQ3^6*msq^2*mU3^2 + 
        419*m3^8*mU3^4 - 1011*m3^6*mQ3^2*mU3^4 + 938*m3^4*mQ3^4*mU3^4 - 
        295*m3^2*mQ3^6*mU3^4 - 3*mQ3^8*mU3^4 + 90*m3^6*msq^2*mU3^4 - 
        210*m3^4*mQ3^2*msq^2*mU3^4 + 150*m3^2*mQ3^4*msq^2*mU3^4 - 
        30*mQ3^6*msq^2*mU3^4 - 21*m3^6*mU3^6 + 34*m3^4*mQ3^2*mU3^6 - 
        231*m3^2*mQ3^4*mU3^6 + 146*mQ3^6*mU3^6 + 30*m3^4*msq^2*mU3^6 - 
        60*m3^2*mQ3^2*msq^2*mU3^6 + 30*mQ3^4*msq^2*mU3^6 - 100*m3^4*mU3^8 + 
        288*m3^2*mQ3^2*mU3^8 - 140*mQ3^4*mU3^8 - 9*m3^2*mU3^10 - 
        3*mQ3^2*mU3^10)*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
        2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*
       (-mQ3^2 + mU3^2)) + (16*(m3^4 - 34*m3^2*mU3^2 + 17*mU3^4)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]])^2)/
      (m3^2 - mU3^2)^2) + (32*a3*at*k*mt^2*(3*m3^4 - 2*m3^2*mU3^2 + mU3^4 - 
      2*m3^2*mU3^2*Log[MR^2] + mU3^4*Log[MR^2] + 4*m3^2*mU3^2*Log[mt] - 
      2*mU3^4*Log[mt] + 4*m3^2*mU3^2*Log[Csc[beta]] - 
      2*mU3^4*Log[Csc[beta]] + 4*m3^2*mU3^2*Log[Sin[beta]] - 
      2*mU3^4*Log[Sin[beta]]))/(m3^2 - mU3^2)^2 + 
   Log[mQ3^2/MR^2]*((16*a3^2*at*k^(3/2)*mt^2*(90*m3^16 - 330*m3^14*mQ3^2 + 
        355*m3^12*mQ3^4 - 128*m3^10*mQ3^6 + 17*m3^8*mQ3^8 - 330*m3^14*mU3^2 + 
        1196*m3^12*mQ3^2*mU3^2 - 1260*m3^10*mQ3^4*mU3^2 + 
        444*m3^8*mQ3^6*mU3^2 - 66*m3^6*mQ3^8*mU3^2 + 355*m3^12*mU3^4 - 
        1260*m3^10*mQ3^2*mU3^4 + 1160*m3^8*mQ3^4*mU3^4 - 
        248*m3^6*mQ3^6*mU3^4 + 17*m3^4*mQ3^8*mU3^4 - 128*m3^10*mU3^6 + 
        444*m3^8*mQ3^2*mU3^6 - 248*m3^6*mQ3^4*mU3^6 - 132*m3^4*mQ3^6*mU3^6 + 
        48*m3^2*mQ3^8*mU3^6 + 17*m3^8*mU3^8 - 66*m3^6*mQ3^2*mU3^8 + 
        17*m3^4*mQ3^4*mU3^8 + 48*m3^2*mQ3^6*mU3^8 - 12*mQ3^8*mU3^8)*
       Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^4) - 
     (8*a3^2*at*k^(3/2)*mt^2*(348*m3^12 - 264*m3^10*mQ3^2 - 201*m3^8*mQ3^4 + 
        192*m3^6*mQ3^6 - 20*m3^4*mQ3^8 + 9*m3^2*mQ3^10 - 
        90*m3^8*mQ3^2*msq^2 - 30*m3^6*mQ3^4*msq^2 - 276*m3^10*mU3^2 - 
        704*m3^8*mQ3^2*mU3^2 + 1498*m3^6*mQ3^4*mU3^2 - 693*m3^4*mQ3^6*mU3^2 - 
        32*m3^2*mQ3^8*mU3^2 + 3*mQ3^10*mU3^2 - 90*m3^8*msq^2*mU3^2 + 
        540*m3^6*mQ3^2*msq^2*mU3^2 - 180*m3^4*mQ3^4*msq^2*mU3^2 + 
        90*m3^2*mQ3^6*msq^2*mU3^2 - 187*m3^8*mU3^4 + 1608*m3^6*mQ3^2*mU3^4 - 
        2017*m3^4*mQ3^4*mU3^4 + 820*m3^2*mQ3^6*mU3^4 + 4*mQ3^8*mU3^4 - 
        30*m3^6*msq^2*mU3^4 - 180*m3^4*mQ3^2*msq^2*mU3^4 - 
        180*m3^2*mQ3^4*msq^2*mU3^4 + 30*mQ3^6*msq^2*mU3^4 + 158*m3^6*mU3^6 - 
        787*m3^4*mQ3^2*mU3^6 + 828*m3^2*mQ3^4*mU3^6 - 299*mQ3^6*mU3^6 + 
        90*m3^2*mQ3^2*msq^2*mU3^6 + 30*mQ3^4*msq^2*mU3^6 + 9*m3^4*mU3^8 + 
        3*m3^2*mQ3^2*mU3^8 - 24*m3^10*mQ3^2*Log[MR^2] + 
        28*m3^8*mQ3^4*Log[MR^2] - 68*m3^6*mQ3^6*Log[MR^2] + 
        58*m3^4*mQ3^8*Log[MR^2] - 18*m3^2*mQ3^10*Log[MR^2] - 
        128*m3^8*mQ3^2*mU3^2*Log[MR^2] + 376*m3^6*mQ3^4*mU3^2*Log[MR^2] - 
        240*m3^4*mQ3^6*mU3^2*Log[MR^2] + 70*m3^2*mQ3^8*mU3^2*Log[MR^2] - 
        6*mQ3^10*mU3^2*Log[MR^2] + 156*m3^6*mQ3^2*mU3^4*Log[MR^2] - 
        350*m3^4*mQ3^4*mU3^4*Log[MR^2] + 130*m3^2*mQ3^6*mU3^4*Log[MR^2] - 
        8*mQ3^8*mU3^4*Log[MR^2] - 52*m3^4*mQ3^2*mU3^6*Log[MR^2] + 
        114*m3^2*mQ3^4*mU3^6*Log[MR^2] - 38*mQ3^6*mU3^6*Log[MR^2] + 
        48*m3^10*mQ3^2*Log[mt] - 56*m3^8*mQ3^4*Log[mt] + 
        136*m3^6*mQ3^6*Log[mt] - 116*m3^4*mQ3^8*Log[mt] + 
        36*m3^2*mQ3^10*Log[mt] + 256*m3^8*mQ3^2*mU3^2*Log[mt] - 
        752*m3^6*mQ3^4*mU3^2*Log[mt] + 480*m3^4*mQ3^6*mU3^2*Log[mt] - 
        140*m3^2*mQ3^8*mU3^2*Log[mt] + 12*mQ3^10*mU3^2*Log[mt] - 
        312*m3^6*mQ3^2*mU3^4*Log[mt] + 700*m3^4*mQ3^4*mU3^4*Log[mt] - 
        260*m3^2*mQ3^6*mU3^4*Log[mt] + 16*mQ3^8*mU3^4*Log[mt] + 
        104*m3^4*mQ3^2*mU3^6*Log[mt] - 228*m3^2*mQ3^4*mU3^6*Log[mt] + 
        76*mQ3^6*mU3^6*Log[mt] + 48*m3^10*mQ3^2*Log[Csc[beta]] - 
        56*m3^8*mQ3^4*Log[Csc[beta]] + 136*m3^6*mQ3^6*Log[Csc[beta]] - 
        116*m3^4*mQ3^8*Log[Csc[beta]] + 36*m3^2*mQ3^10*Log[Csc[beta]] + 
        256*m3^8*mQ3^2*mU3^2*Log[Csc[beta]] - 752*m3^6*mQ3^4*mU3^2*
         Log[Csc[beta]] + 480*m3^4*mQ3^6*mU3^2*Log[Csc[beta]] - 
        140*m3^2*mQ3^8*mU3^2*Log[Csc[beta]] + 12*mQ3^10*mU3^2*
         Log[Csc[beta]] - 312*m3^6*mQ3^2*mU3^4*Log[Csc[beta]] + 
        700*m3^4*mQ3^4*mU3^4*Log[Csc[beta]] - 260*m3^2*mQ3^6*mU3^4*
         Log[Csc[beta]] + 16*mQ3^8*mU3^4*Log[Csc[beta]] + 
        104*m3^4*mQ3^2*mU3^6*Log[Csc[beta]] - 228*m3^2*mQ3^4*mU3^6*
         Log[Csc[beta]] + 76*mQ3^6*mU3^6*Log[Csc[beta]] + 
        48*m3^10*mQ3^2*Log[Sin[beta]] - 56*m3^8*mQ3^4*Log[Sin[beta]] + 
        136*m3^6*mQ3^6*Log[Sin[beta]] - 116*m3^4*mQ3^8*Log[Sin[beta]] + 
        36*m3^2*mQ3^10*Log[Sin[beta]] + 256*m3^8*mQ3^2*mU3^2*Log[Sin[beta]] - 
        752*m3^6*mQ3^4*mU3^2*Log[Sin[beta]] + 480*m3^4*mQ3^6*mU3^2*
         Log[Sin[beta]] - 140*m3^2*mQ3^8*mU3^2*Log[Sin[beta]] + 
        12*mQ3^10*mU3^2*Log[Sin[beta]] - 312*m3^6*mQ3^2*mU3^4*
         Log[Sin[beta]] + 700*m3^4*mQ3^4*mU3^4*Log[Sin[beta]] - 
        260*m3^2*mQ3^6*mU3^4*Log[Sin[beta]] + 16*mQ3^8*mU3^4*Log[Sin[beta]] + 
        104*m3^4*mQ3^2*mU3^6*Log[Sin[beta]] - 228*m3^2*mQ3^4*mU3^6*
         Log[Sin[beta]] + 76*mQ3^6*mU3^6*Log[Sin[beta]]))/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3))) + 
 ((-64*a3*at*k*m3^4*mt^2)/(-m3^2 + mQ3^2)^2 - 
   (64*a3^2*at*k^(3/2)*m3^4*mt^2*
     (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
         m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*Log[m3^2/MR^2])/
    (-m3^2 + mQ3^2)^2 + a3^2*at*k^(3/2)*mt^2*
    ((-32*m3^4*(m3^4 - 34*m3^2*mQ3^2 + 17*mQ3^4))/(3*(m3^2 - mQ3^2)^4) + 
     (8*(-128*m3^12 + 294*m3^10*mQ3^2 - 98*m3^8*mQ3^4 - 9*m3^6*mQ3^6 - 
        20*m3^4*mQ3^8 + 9*m3^2*mQ3^10 + 346*m3^10*mU3^2 - 
        902*m3^8*mQ3^2*mU3^2 + 251*m3^6*mQ3^4*mU3^2 + 151*m3^4*mQ3^6*mU3^2 - 
        41*m3^2*mQ3^8*mU3^2 + 3*mQ3^10*mU3^2 - 280*m3^8*mU3^4 + 
        1025*m3^6*mQ3^2*mU3^4 - 417*m3^4*mQ3^4*mU3^4 - 41*m3^2*mQ3^6*mU3^4 + 
        mQ3^8*mU3^4 + 13*m3^6*mU3^6 - 391*m3^4*mQ3^2*mU3^6 + 
        167*m3^2*mQ3^4*mU3^6 + 19*mQ3^6*mU3^6 + 37*m3^4*mU3^8 + 
        34*m3^2*mQ3^2*mU3^8 - 23*mQ3^4*mU3^8))/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)))*Log[mQ3^2/MR^2] - 
   (320*a3^2*at*k^(3/2)*m3^4*mt^2*Log[msq^2/MR^2])/(3*(-m3^2 + mQ3^2)^2) - 
   (8*a3^2*at*k^(3/2)*mt^2*(128*m3^12 - 290*m3^10*mQ3^2 + 98*m3^8*mQ3^4 + 
      9*m3^6*mQ3^6 + 20*m3^4*mQ3^8 - 9*m3^2*mQ3^10 - 350*m3^10*mU3^2 + 
      762*m3^8*mQ3^2*mU3^2 - 251*m3^6*mQ3^4*mU3^2 - 151*m3^4*mQ3^6*mU3^2 + 
      41*m3^2*mQ3^8*mU3^2 - 3*mQ3^10*mU3^2 + 420*m3^8*mU3^4 - 
      821*m3^6*mQ3^2*mU3^4 + 417*m3^4*mQ3^4*mU3^4 + 41*m3^2*mQ3^6*mU3^4 - 
      mQ3^8*mU3^4 - 217*m3^6*mU3^6 + 323*m3^4*mQ3^2*mU3^6 - 
      167*m3^2*mQ3^4*mU3^6 - 19*mQ3^6*mU3^6 + 31*m3^4*mU3^8 - 
      34*m3^2*mQ3^2*mU3^8 + 23*mQ3^4*mU3^8)*Log[mU3^2/MR^2])/
    (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) + 
   a3^2*at*k^(3/2)*mt^2*((64*m3^4*(11*m3^4 - 3*m3^2*mQ3^2 - 3*m3^2*mU3^2 - 
        5*mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)) - 
     (8*(-128*m3^12 + 294*m3^10*mQ3^2 - 98*m3^8*mQ3^4 - 9*m3^6*mQ3^6 - 
        20*m3^4*mQ3^8 + 9*m3^2*mQ3^10 + 346*m3^10*mU3^2 - 
        902*m3^8*mQ3^2*mU3^2 + 251*m3^6*mQ3^4*mU3^2 + 151*m3^4*mQ3^6*mU3^2 - 
        41*m3^2*mQ3^8*mU3^2 + 3*mQ3^10*mU3^2 - 280*m3^8*mU3^4 + 
        1025*m3^6*mQ3^2*mU3^4 - 417*m3^4*mQ3^4*mU3^4 - 41*m3^2*mQ3^6*mU3^4 + 
        mQ3^8*mU3^4 + 13*m3^6*mU3^6 - 391*m3^4*mQ3^2*mU3^6 + 
        167*m3^2*mQ3^4*mU3^6 + 19*mQ3^6*mU3^6 + 37*m3^4*mU3^8 + 
        34*m3^2*mQ3^2*mU3^8 - 23*mQ3^4*mU3^8))/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) + 
     (16*(-128*m3^12 + 294*m3^10*mQ3^2 - 98*m3^8*mQ3^4 - 9*m3^6*mQ3^6 - 
        20*m3^4*mQ3^8 + 9*m3^2*mQ3^10 + 346*m3^10*mU3^2 - 
        902*m3^8*mQ3^2*mU3^2 + 251*m3^6*mQ3^4*mU3^2 + 151*m3^4*mQ3^6*mU3^2 - 
        41*m3^2*mQ3^8*mU3^2 + 3*mQ3^10*mU3^2 - 280*m3^8*mU3^4 + 
        1025*m3^6*mQ3^2*mU3^4 - 417*m3^4*mQ3^4*mU3^4 - 41*m3^2*mQ3^6*mU3^4 + 
        mQ3^8*mU3^4 + 13*m3^6*mU3^6 - 391*m3^4*mQ3^2*mU3^6 + 
        167*m3^2*mQ3^4*mU3^6 + 19*mQ3^6*mU3^6 + 37*m3^4*mU3^8 + 
        34*m3^2*mQ3^2*mU3^8 - 23*mQ3^4*mU3^8)*(Log[MR^2] - 2*Log[mt] - 
        2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2))))*PolyLog[2, 1 - m3^2/mQ3^2] + 
 (-12*a3^2*at*k^(3/2)*(2*m3^2 - 2*msq^2)*mt^2*
    ((40*mQ3^4)/(3*(-m3^2 + mQ3^2)^3) - (10*mQ3^2)/(-m3^2 + mQ3^2)^2 - 
     10/(3*(-m3^2 + mQ3^2)) - (40*mQ3^2*msq^2)/(3*(-m3^2 + mQ3^2)^3) + 
     (10*msq^2)/(-m3^2 + mQ3^2)^2 - (40*msq^2*mU3^2)/(3*(-m3^2 + mU3^2)^3) + 
     (40*mU3^4)/(3*(-m3^2 + mU3^2)^3) + (10*msq^2)/(-m3^2 + mU3^2)^2 - 
     (10*mU3^2)/(-m3^2 + mU3^2)^2 - 10/(3*(-m3^2 + mU3^2)))*Log[mQ3^2/MR^2] - 
   12*a3^2*at*k^(3/2)*(2*m3^2 - 2*msq^2)*mt^2*
    ((40*mQ3^4)/(3*(-m3^2 + mQ3^2)^3) - (10*mQ3^2)/(-m3^2 + mQ3^2)^2 - 
     10/(3*(-m3^2 + mQ3^2)) - (40*mQ3^2*msq^2)/(3*(-m3^2 + mQ3^2)^3) + 
     (10*msq^2)/(-m3^2 + mQ3^2)^2 - (40*msq^2*mU3^2)/(3*(-m3^2 + mU3^2)^3) + 
     (40*mU3^4)/(3*(-m3^2 + mU3^2)^3) + (10*msq^2)/(-m3^2 + mU3^2)^2 - 
     (10*mU3^2)/(-m3^2 + mU3^2)^2 - 10/(3*(-m3^2 + mU3^2)))*Log[mU3^2/MR^2] - 
   (80*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*(2*m3^10 - 8*m3^8*mQ3^2 + 
      3*m3^6*mQ3^4 - m3^4*mQ3^6 + 6*m3^8*msq^2 - 8*m3^6*mQ3^2*msq^2 + 
      9*m3^4*mQ3^4*msq^2 - 3*m3^2*mQ3^6*msq^2 - 8*m3^8*mU3^2 + 
      30*m3^6*mQ3^2*mU3^2 - 15*m3^4*mQ3^4*mU3^2 + 5*m3^2*mQ3^6*mU3^2 - 
      8*m3^6*msq^2*mU3^2 - 6*m3^4*mQ3^2*msq^2*mU3^2 + 
      3*m3^2*mQ3^4*msq^2*mU3^2 - mQ3^6*msq^2*mU3^2 + 3*m3^6*mU3^4 - 
      15*m3^4*mQ3^2*mU3^4 + 9*m3^4*msq^2*mU3^4 + 3*m3^2*mQ3^2*msq^2*mU3^4 - 
      m3^4*mU3^6 + 5*m3^2*mQ3^2*mU3^6 - 3*m3^2*msq^2*mU3^6 - 
      mQ3^2*msq^2*mU3^6)*(-1 + 2*Log[MR^2] - 4*Log[mt] - 4*Log[Csc[beta]] - 
      4*Log[Sin[beta]]))/((m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3))*
  PolyLog[2, 1 - m3^2/msq^2] + 
 ((40*a3^2*at*k^(3/2)*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 
      3*m3^2*msq^2 + mQ3^2*msq^2)*mt^2*Log[mQ3^2/MR^2])/(-m3^2 + mQ3^2)^3 + 
   (40*a3^2*at*k^(3/2)*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 
      3*m3^2*msq^2 + mQ3^2*msq^2)*mt^2*Log[mU3^2/MR^2])/(-m3^2 + mQ3^2)^3 + 
   (80*a3^2*at*k^(3/2)*(mQ3^2 - msq^2)*(2*m3^4 + 3*m3^2*mQ3^2 - mQ3^4 - 
      3*m3^2*msq^2 - mQ3^2*msq^2)*mt^2*(-1 + 2*Log[MR^2] - 4*Log[mt] - 
      4*Log[Csc[beta]] - 4*Log[Sin[beta]]))/(m3^2 - mQ3^2)^3)*
  PolyLog[2, 1 - msq^2/mQ3^2] + ((-64*a3*at*k*m3^4*mt^2)/(m3^2 - mU3^2)^2 - 
   (64*a3^2*at*k^(3/2)*m3^4*mt^2*
     (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
         m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*Log[m3^2/MR^2])/
    (m3^2 - mU3^2)^2 + (8*a3^2*at*k^(3/2)*mt^2*(128*m3^12 - 350*m3^10*mQ3^2 + 
      420*m3^8*mQ3^4 - 217*m3^6*mQ3^6 + 31*m3^4*mQ3^8 - 290*m3^10*mU3^2 + 
      762*m3^8*mQ3^2*mU3^2 - 821*m3^6*mQ3^4*mU3^2 + 323*m3^4*mQ3^6*mU3^2 - 
      34*m3^2*mQ3^8*mU3^2 + 98*m3^8*mU3^4 - 251*m3^6*mQ3^2*mU3^4 + 
      417*m3^4*mQ3^4*mU3^4 - 167*m3^2*mQ3^6*mU3^4 + 23*mQ3^8*mU3^4 + 
      9*m3^6*mU3^6 - 151*m3^4*mQ3^2*mU3^6 + 41*m3^2*mQ3^4*mU3^6 - 
      19*mQ3^6*mU3^6 + 20*m3^4*mU3^8 + 41*m3^2*mQ3^2*mU3^8 - mQ3^4*mU3^8 - 
      9*m3^2*mU3^10 - 3*mQ3^2*mU3^10)*Log[mQ3^2/MR^2])/
    (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) - 
   (320*a3^2*at*k^(3/2)*m3^4*mt^2*Log[msq^2/MR^2])/(3*(m3^2 - mU3^2)^2) + 
   a3^2*at*k^(3/2)*mt^2*((-32*m3^4*(m3^4 - 34*m3^2*mU3^2 + 17*mU3^4))/
      (3*(m3^2 - mU3^2)^4) - (8*(-128*m3^12 + 346*m3^10*mQ3^2 - 
        280*m3^8*mQ3^4 + 13*m3^6*mQ3^6 + 37*m3^4*mQ3^8 + 294*m3^10*mU3^2 - 
        902*m3^8*mQ3^2*mU3^2 + 1025*m3^6*mQ3^4*mU3^2 - 391*m3^4*mQ3^6*mU3^2 + 
        34*m3^2*mQ3^8*mU3^2 - 98*m3^8*mU3^4 + 251*m3^6*mQ3^2*mU3^4 - 
        417*m3^4*mQ3^4*mU3^4 + 167*m3^2*mQ3^6*mU3^4 - 23*mQ3^8*mU3^4 - 
        9*m3^6*mU3^6 + 151*m3^4*mQ3^2*mU3^6 - 41*m3^2*mQ3^4*mU3^6 + 
        19*mQ3^6*mU3^6 - 20*m3^4*mU3^8 - 41*m3^2*mQ3^2*mU3^8 + mQ3^4*mU3^8 + 
        9*m3^2*mU3^10 + 3*mQ3^2*mU3^10))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*
       (mQ3^2 - mU3^2)))*Log[mU3^2/MR^2] + a3^2*at*k^(3/2)*mt^2*
    ((-64*m3^4*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2))/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^3) + 
     (8*(-128*m3^12 + 346*m3^10*mQ3^2 - 280*m3^8*mQ3^4 + 13*m3^6*mQ3^6 + 
        37*m3^4*mQ3^8 + 294*m3^10*mU3^2 - 902*m3^8*mQ3^2*mU3^2 + 
        1025*m3^6*mQ3^4*mU3^2 - 391*m3^4*mQ3^6*mU3^2 + 34*m3^2*mQ3^8*mU3^2 - 
        98*m3^8*mU3^4 + 251*m3^6*mQ3^2*mU3^4 - 417*m3^4*mQ3^4*mU3^4 + 
        167*m3^2*mQ3^6*mU3^4 - 23*mQ3^8*mU3^4 - 9*m3^6*mU3^6 + 
        151*m3^4*mQ3^2*mU3^6 - 41*m3^2*mQ3^4*mU3^6 + 19*mQ3^6*mU3^6 - 
        20*m3^4*mU3^8 - 41*m3^2*mQ3^2*mU3^8 + mQ3^4*mU3^8 + 9*m3^2*mU3^10 + 
        3*mQ3^2*mU3^10))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*
       (mQ3^2 - mU3^2)) - (16*(-128*m3^12 + 346*m3^10*mQ3^2 - 
        280*m3^8*mQ3^4 + 13*m3^6*mQ3^6 + 37*m3^4*mQ3^8 + 294*m3^10*mU3^2 - 
        902*m3^8*mQ3^2*mU3^2 + 1025*m3^6*mQ3^4*mU3^2 - 391*m3^4*mQ3^6*mU3^2 + 
        34*m3^2*mQ3^8*mU3^2 - 98*m3^8*mU3^4 + 251*m3^6*mQ3^2*mU3^4 - 
        417*m3^4*mQ3^4*mU3^4 + 167*m3^2*mQ3^6*mU3^4 - 23*mQ3^8*mU3^4 - 
        9*m3^6*mU3^6 + 151*m3^4*mQ3^2*mU3^6 - 41*m3^2*mQ3^4*mU3^6 + 
        19*mQ3^6*mU3^6 - 20*m3^4*mU3^8 - 41*m3^2*mQ3^2*mU3^8 + mQ3^4*mU3^8 + 
        9*m3^2*mU3^10 + 3*mQ3^2*mU3^10)*(Log[MR^2] - 2*Log[mt] - 
        2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^2*(mQ3^2 - mU3^2))))*PolyLog[2, 1 - m3^2/mU3^2] + 
 ((-4*a3^2*at*k^(3/2)*mt^2*(-24*m3^10*mQ3^2 + 28*m3^8*mQ3^4 - 68*m3^6*mQ3^6 + 
      58*m3^4*mQ3^8 - 18*m3^2*mQ3^10 + 24*m3^10*mU3^2 + 
      220*m3^6*mQ3^4*mU3^2 - 188*m3^4*mQ3^6*mU3^2 + 70*m3^2*mQ3^8*mU3^2 - 
      6*mQ3^10*mU3^2 - 28*m3^8*mU3^4 - 220*m3^6*mQ3^2*mU3^4 + 
      16*m3^2*mQ3^6*mU3^4 - 8*mQ3^8*mU3^4 + 68*m3^6*mU3^6 + 
      188*m3^4*mQ3^2*mU3^6 - 16*m3^2*mQ3^4*mU3^6 - 58*m3^4*mU3^8 - 
      70*m3^2*mQ3^2*mU3^8 + 8*mQ3^4*mU3^8 + 18*m3^2*mU3^10 + 6*mQ3^2*mU3^10)*
     Log[mQ3^2/MR^2])/(3*(-m3^2 + mQ3^2)^3*(m3^2 - mU3^2)^3) - 
   (4*a3^2*at*k^(3/2)*mt^2*(-24*m3^10*mQ3^2 + 28*m3^8*mQ3^4 - 68*m3^6*mQ3^6 + 
      58*m3^4*mQ3^8 - 18*m3^2*mQ3^10 + 24*m3^10*mU3^2 + 
      220*m3^6*mQ3^4*mU3^2 - 188*m3^4*mQ3^6*mU3^2 + 70*m3^2*mQ3^8*mU3^2 - 
      6*mQ3^10*mU3^2 - 28*m3^8*mU3^4 - 220*m3^6*mQ3^2*mU3^4 + 
      16*m3^2*mQ3^6*mU3^4 - 8*mQ3^8*mU3^4 + 68*m3^6*mU3^6 + 
      188*m3^4*mQ3^2*mU3^6 - 16*m3^2*mQ3^4*mU3^6 - 58*m3^4*mU3^8 - 
      70*m3^2*mQ3^2*mU3^8 + 8*mQ3^4*mU3^8 + 18*m3^2*mU3^10 + 6*mQ3^2*mU3^10)*
     Log[mU3^2/MR^2])/(3*(-m3^2 + mQ3^2)^3*(m3^2 - mU3^2)^3) + 
   a3^2*at*k^(3/2)*mt^2*((4*(-24*m3^10*mQ3^2 + 28*m3^8*mQ3^4 - 
        68*m3^6*mQ3^6 + 58*m3^4*mQ3^8 - 18*m3^2*mQ3^10 + 24*m3^10*mU3^2 + 
        220*m3^6*mQ3^4*mU3^2 - 188*m3^4*mQ3^6*mU3^2 + 70*m3^2*mQ3^8*mU3^2 - 
        6*mQ3^10*mU3^2 - 28*m3^8*mU3^4 - 220*m3^6*mQ3^2*mU3^4 + 
        16*m3^2*mQ3^6*mU3^4 - 8*mQ3^8*mU3^4 + 68*m3^6*mU3^6 + 
        188*m3^4*mQ3^2*mU3^6 - 16*m3^2*mQ3^4*mU3^6 - 58*m3^4*mU3^8 - 
        70*m3^2*mQ3^2*mU3^8 + 8*mQ3^4*mU3^8 + 18*m3^2*mU3^10 + 
        6*mQ3^2*mU3^10))/(3*(-m3^2 + mQ3^2)^3*(m3^2 - mU3^2)^3) - 
     (8*(-24*m3^10*mQ3^2 + 28*m3^8*mQ3^4 - 68*m3^6*mQ3^6 + 58*m3^4*mQ3^8 - 
        18*m3^2*mQ3^10 + 24*m3^10*mU3^2 + 220*m3^6*mQ3^4*mU3^2 - 
        188*m3^4*mQ3^6*mU3^2 + 70*m3^2*mQ3^8*mU3^2 - 6*mQ3^10*mU3^2 - 
        28*m3^8*mU3^4 - 220*m3^6*mQ3^2*mU3^4 + 16*m3^2*mQ3^6*mU3^4 - 
        8*mQ3^8*mU3^4 + 68*m3^6*mU3^6 + 188*m3^4*mQ3^2*mU3^6 - 
        16*m3^2*mQ3^4*mU3^6 - 58*m3^4*mU3^8 - 70*m3^2*mQ3^2*mU3^8 + 
        8*mQ3^4*mU3^8 + 18*m3^2*mU3^10 + 6*mQ3^2*mU3^10)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      (3*(-m3^2 + mQ3^2)^3*(m3^2 - mU3^2)^3)))*PolyLog[2, 1 - mQ3^2/mU3^2] + 
 ((40*a3^2*at*k^(3/2)*mt^2*(-2*msq^2 + 2*mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 
      3*m3^2*mU3^2 + msq^2*mU3^2 + mU3^4)*Log[mQ3^2/MR^2])/
    (-m3^2 + mU3^2)^3 + (40*a3^2*at*k^(3/2)*mt^2*(-2*msq^2 + 2*mU3^2)*
     (-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + msq^2*mU3^2 + mU3^4)*
     Log[mU3^2/MR^2])/(-m3^2 + mU3^2)^3 + 
   (80*a3^2*at*k^(3/2)*mt^2*(-msq^2 + mU3^2)*(2*m3^4 - 3*m3^2*msq^2 + 
      3*m3^2*mU3^2 - msq^2*mU3^2 - mU3^4)*(-1 + 2*Log[MR^2] - 4*Log[mt] - 
      4*Log[Csc[beta]] - 4*Log[Sin[beta]]))/(m3^2 - mU3^2)^3)*
  PolyLog[2, 1 - msq^2/mU3^2] + 
 Xt^3*((-512*a3*at*k*m3*mt^2)/(mQ3^2 - mU3^2)^2 - 
   (512*a3^2*at*k^(3/2)*m3*mt^2*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 
      5*mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^2) + 
   a3^2*at*k^(3/2)*mt^2*((2048*m3^5)/(3*mQ3^2*(m3^2 - mQ3^2)*mU3^2*
       (m3^2 - mU3^2)) - 
     (512*m3*(2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
           m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)))/
      (mQ3^2 - mU3^2)^2)*Log[m3^2/MR^2] - 
   (2048*a3^2*at*k^(3/2)*m3^7*mt^2*(m3^2 - mQ3^2 - mU3^2)*Log[m3^2/MR^2]^2)/
    (3*mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*(m3^2 - mU3^2)^2) + 
   (128*a3*at*k*m3*mt^2*(3*mQ3^2 + mU3^2)*Log[mQ3^2/MR^2])/
    (mQ3^2 - mU3^2)^3 + a3^2*at*k^(3/2)*mt^2*
    ((128*(14*m3^3 - 3*m3*mQ3^2 - 30*m3*msq^2 - 3*m3*mU3^2))/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) + 
     (128*m3*(3*mQ3^2 + mU3^2)*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 
        5*mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^3) - 
     (256*(8*m3^5*mQ3^2 - 8*m3^3*mQ3^4 - 7*m3^5*mU3^2 - 26*m3^3*mQ3^2*mU3^2 + 
        17*m3*mQ3^4*mU3^2))/(3*(m3^2 - mQ3^2)^2*mU3^2*(mQ3^2 - mU3^2)^2))*
    Log[mQ3^2/MR^2] + (128*a3^2*at*k^(3/2)*mt^2*(16*m3^9*mQ3^4 - 
      16*m3^7*mQ3^6 + 22*m3^11*mU3^2 + 6*m3^9*mQ3^2*mU3^2 - 
      217*m3^7*mQ3^4*mU3^2 + 165*m3^5*mQ3^6*mU3^2 - 3*m3^3*mQ3^8*mU3^2 + 
      60*m3^5*mQ3^4*msq^2*mU3^2 - 30*m3^3*mQ3^6*msq^2*mU3^2 + 10*m3^9*mU3^4 + 
      238*m3^7*mQ3^2*mU3^4 - 17*m3^5*mQ3^4*mU3^4 - 115*m3^3*mQ3^6*mU3^4 - 
      120*m3^5*mQ3^2*msq^2*mU3^4 + 30*m3^3*mQ3^4*msq^2*mU3^4 - 
      193*m3^7*mU3^6 - 85*m3^5*mQ3^2*mU3^6 + 106*m3^3*mQ3^4*mU3^6 + 
      18*m3*mQ3^6*mU3^6 + 60*m3^5*msq^2*mU3^6 + 30*m3^3*mQ3^2*msq^2*mU3^6 + 
      153*m3^5*mU3^8 - 91*m3^3*mQ3^2*mU3^8 + 6*m3*mQ3^4*mU3^8 - 
      30*m3^3*msq^2*mU3^8 - 3*m3^3*mU3^10)*Log[m3^2/MR^2]*Log[mQ3^2/MR^2])/
    (3*(m3^2 - mQ3^2)^2*mU3^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) + 
   ((-256*a3*at*k*m3^3*mt^2*Log[m3^2/MR^2])/(mQ3^2 - mU3^2)^3 - 
     (128*a3^2*at*k^(3/2)*mt^2*(44*m3^15 - 100*m3^13*mQ3^2 + 45*m3^11*mQ3^4 + 
        66*m3^9*mQ3^6 - 63*m3^7*mQ3^8 - 100*m3^13*mU3^2 + 
        282*m3^11*mQ3^2*mU3^2 - 250*m3^9*mQ3^4*mU3^2 - 42*m3^7*mQ3^6*mU3^2 + 
        142*m3^5*mQ3^8*mU3^2 + 45*m3^11*mU3^4 - 250*m3^9*mQ3^2*mU3^4 + 
        422*m3^7*mQ3^4*mU3^4 - 178*m3^5*mQ3^6*mU3^4 - 87*m3^3*mQ3^8*mU3^4 + 
        66*m3^9*mU3^6 - 42*m3^7*mQ3^2*mU3^6 - 178*m3^5*mQ3^4*mU3^6 + 
        186*m3^3*mQ3^6*mU3^6 - 63*m3^7*mU3^8 + 142*m3^5*mQ3^2*mU3^8 - 
        87*m3^3*mQ3^4*mU3^8)*Log[m3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^3))*Log[mQ3^2/MR^2] - 
   (128*a3^2*at*k^(3/2)*mt^2*(11*m3^9 - 37*m3^7*mQ3^2 - 48*m3^5*mQ3^4 + 
      66*m3^3*mQ3^6 - 3*m3*mQ3^8 + 30*m3^3*mQ3^4*msq^2 + 7*m3^7*mU3^2 + 
      150*m3^5*mQ3^2*mU3^2 - 47*m3^3*mQ3^4*mU3^2 - 53*m3*mQ3^6*mU3^2 - 
      30*m3^3*mQ3^2*msq^2*mU3^2 - 30*m3*mQ3^4*msq^2*mU3^2 - 14*m3^5*mU3^4 - 
      117*m3^3*mQ3^2*mU3^4 + 82*m3*mQ3^4*mU3^4 + 30*m3*mQ3^2*msq^2*mU3^4 + 
      3*m3*mQ3^2*mU3^6)*Log[mQ3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*
     (mQ3^2 - mU3^2)^3) + ((64*a3*at*k*m3*mt^2*(2*m3^2 - mQ3^2 - mU3^2))/
      (mQ3^2 - mU3^2)^3 + (64*a3^2*at*k^(3/2)*mt^2*
       (42*m3^13 - 120*m3^11*mQ3^2 + 239*m3^9*mQ3^4 - 252*m3^7*mQ3^6 + 
        59*m3^5*mQ3^8 - 6*m3^11*mU3^2 - 44*m3^9*mQ3^2*mU3^2 - 
        67*m3^7*mQ3^4*mU3^2 + 268*m3^5*mQ3^6*mU3^2 - 55*m3^3*mQ3^8*mU3^2 - 
        67*m3^9*mU3^4 + 214*m3^7*mQ3^2*mU3^4 - 195*m3^5*mQ3^4*mU3^4 - 
        42*m3^3*mQ3^6*mU3^4 - 6*m3*mQ3^8*mU3^4 + 45*m3^7*mU3^6 - 
        90*m3^5*mQ3^2*mU3^6 + 59*m3^3*mQ3^4*mU3^6 + 18*m3*mQ3^6*mU3^6)*
       Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*
       (mQ3^2 - mU3^2)^3))*Log[mQ3^2/MR^2]^2 + 
   (32*a3^2*at*k^(3/2)*mt^2*(2*m3^13 - 35*m3^11*mQ3^2 - 131*m3^9*mQ3^4 + 
      363*m3^7*mQ3^6 - 177*m3^5*mQ3^8 + 16*m3^3*mQ3^10 - 6*m3*mQ3^12 + 
      120*m3^7*mQ3^4*msq^2 - 120*m3^5*mQ3^6*msq^2 - 60*m3^7*mQ3^2*msq^4 + 
      60*m3^5*mQ3^4*msq^4 - 41*m3^11*mU3^2 + 275*m3^9*mQ3^2*mU3^2 - 
      49*m3^7*mQ3^4*mU3^2 - 575*m3^5*mQ3^6*mU3^2 + 274*m3^3*mQ3^8*mU3^2 + 
      20*m3*mQ3^10*mU3^2 - 120*m3^7*mQ3^2*msq^2*mU3^2 - 
      120*m3^5*mQ3^4*msq^2*mU3^2 + 240*m3^3*mQ3^6*msq^2*mU3^2 + 
      60*m3^7*msq^4*mU3^2 + 60*m3^5*mQ3^2*msq^4*mU3^2 - 
      120*m3^3*mQ3^4*msq^4*mU3^2 + 78*m3^9*mU3^4 - 437*m3^7*mQ3^2*mU3^4 + 
      421*m3^5*mQ3^4*mU3^4 + 221*m3^3*mQ3^6*mU3^4 - 187*m3*mQ3^8*mU3^4 + 
      240*m3^5*mQ3^2*msq^2*mU3^4 - 120*m3^3*mQ3^4*msq^2*mU3^4 - 
      120*m3*mQ3^6*msq^2*mU3^4 - 120*m3^5*msq^4*mU3^4 + 
      60*m3^3*mQ3^2*msq^4*mU3^4 + 60*m3*mQ3^4*msq^4*mU3^4 - 45*m3^7*mU3^6 + 
      233*m3^5*mQ3^2*mU3^6 - 331*m3^3*mQ3^4*mU3^6 + 111*m3*mQ3^6*mU3^6 - 
      120*m3^3*mQ3^2*msq^2*mU3^6 + 120*m3*mQ3^4*msq^2*mU3^6 + 
      60*m3^3*msq^4*mU3^6 - 60*m3*mQ3^2*msq^4*mU3^6)*Log[mQ3^2/MR^2]^3)/
    (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) - 
   (2560*a3^2*at*k^(3/2)*m3*mt^2*Log[msq^2/MR^2])/(3*(mQ3^2 - mU3^2)^2) + 
   a3^2*at*k^(3/2)*mt^2*((-128*(14*m3^3 - 3*m3*mQ3^2 - 30*m3*msq^2 - 
        3*m3*mU3^2))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) - 
     (128*m3*(mQ3^2 + 3*mU3^2)*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 
        5*mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^3) + 
     (256*(7*m3^5*mQ3^2 - 8*m3^5*mU3^2 + 26*m3^3*mQ3^2*mU3^2 + 8*m3^3*mU3^4 - 
        17*m3*mQ3^2*mU3^4))/(3*mQ3^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2))*
    Log[mU3^2/MR^2] + 
   ((-128*a3^2*at*k^(3/2)*mt^2*(22*m3^11*mQ3^2 + 10*m3^9*mQ3^4 - 
        193*m3^7*mQ3^6 + 153*m3^5*mQ3^8 - 3*m3^3*mQ3^10 + 
        60*m3^5*mQ3^6*msq^2 - 30*m3^3*mQ3^8*msq^2 + 6*m3^9*mQ3^2*mU3^2 + 
        238*m3^7*mQ3^4*mU3^2 - 85*m3^5*mQ3^6*mU3^2 - 91*m3^3*mQ3^8*mU3^2 - 
        120*m3^5*mQ3^4*msq^2*mU3^2 + 30*m3^3*mQ3^6*msq^2*mU3^2 + 
        16*m3^9*mU3^4 - 217*m3^7*mQ3^2*mU3^4 - 17*m3^5*mQ3^4*mU3^4 + 
        106*m3^3*mQ3^6*mU3^4 + 6*m3*mQ3^8*mU3^4 + 60*m3^5*mQ3^2*msq^2*mU3^4 + 
        30*m3^3*mQ3^4*msq^2*mU3^4 - 16*m3^7*mU3^6 + 165*m3^5*mQ3^2*mU3^6 - 
        115*m3^3*mQ3^4*mU3^6 + 18*m3*mQ3^6*mU3^6 - 30*m3^3*mQ3^2*msq^2*
         mU3^6 - 3*m3^3*mQ3^2*mU3^8)*Log[m3^2/MR^2])/
      (3*mQ3^2*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) - 
     (128*a3^2*at*k^(3/2)*mt^2*(31*m3^9 + 63*m3^7*mQ3^2 - 86*m3^5*mQ3^4 + 
        3*m3^3*mQ3^6 - 30*m3^5*mQ3^2*msq^2 + 63*m3^7*mU3^2 - 
        424*m3^5*mQ3^2*mU3^2 + 310*m3^3*mQ3^4*mU3^2 - 6*m3*mQ3^6*mU3^2 - 
        30*m3^5*msq^2*mU3^2 + 120*m3^3*mQ3^2*msq^2*mU3^2 - 
        30*m3*mQ3^4*msq^2*mU3^2 - 86*m3^5*mU3^4 + 310*m3^3*mQ3^2*mU3^4 - 
        175*m3*mQ3^4*mU3^4 - 30*m3*mQ3^2*msq^2*mU3^4 + 3*m3^3*mU3^6 - 
        6*m3*mQ3^2*mU3^6)*Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2))*Log[mU3^2/MR^2] + 
   ((-128*a3*at*k*m3*mt^2*(mQ3^2 + 3*mU3^2))/(mQ3^2 - mU3^2)^3 - 
     (256*a3*at*k*m3^3*mt^2*Log[m3^2/MR^2])/(-mQ3^2 + mU3^2)^3 + 
     (128*a3^2*at*k^(3/2)*mt^2*(44*m3^15 - 100*m3^13*mQ3^2 + 45*m3^11*mQ3^4 + 
        66*m3^9*mQ3^6 - 63*m3^7*mQ3^8 - 100*m3^13*mU3^2 + 
        282*m3^11*mQ3^2*mU3^2 - 250*m3^9*mQ3^4*mU3^2 - 42*m3^7*mQ3^6*mU3^2 + 
        142*m3^5*mQ3^8*mU3^2 + 45*m3^11*mU3^4 - 250*m3^9*mQ3^2*mU3^4 + 
        422*m3^7*mQ3^4*mU3^4 - 178*m3^5*mQ3^6*mU3^4 - 87*m3^3*mQ3^8*mU3^4 + 
        66*m3^9*mU3^6 - 42*m3^7*mQ3^2*mU3^6 - 178*m3^5*mQ3^4*mU3^6 + 
        186*m3^3*mQ3^6*mU3^6 - 63*m3^7*mU3^8 + 142*m3^5*mQ3^2*mU3^8 - 
        87*m3^3*mQ3^4*mU3^8)*Log[m3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^3) + 
     (256*a3^2*at*k^(3/2)*mt^2*(18*m3^13 - 75*m3^11*mQ3^2 + 91*m3^9*mQ3^4 - 
        26*m3^7*mQ3^6 - 75*m3^11*mU3^2 + 232*m3^9*mQ3^2*mU3^2 - 
        232*m3^7*mQ3^4*mU3^2 + 51*m3^5*mQ3^6*mU3^2 + 91*m3^9*mU3^4 - 
        232*m3^7*mQ3^2*mU3^4 + 192*m3^5*mQ3^4*mU3^4 - 27*m3^3*mQ3^6*mU3^4 - 
        26*m3^7*mU3^6 + 51*m3^5*mQ3^2*mU3^6 - 27*m3^3*mQ3^4*mU3^6 - 
        6*m3*mQ3^6*mU3^6)*Log[m3^2/MR^2]*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^2) + 
     (32*a3^2*at*k^(3/2)*mt^2*(2*m3^13 - 43*m3^11*mQ3^2 + 243*m3^9*mQ3^4 - 
        405*m3^7*mQ3^6 + 201*m3^5*mQ3^8 - 48*m3^3*mQ3^10 + 18*m3*mQ3^12 - 
        120*m3^7*mQ3^4*msq^2 + 120*m3^5*mQ3^6*msq^2 + 60*m3^7*mQ3^2*msq^4 - 
        60*m3^5*mQ3^4*msq^4 - 33*m3^11*mU3^2 + 403*m3^9*mQ3^2*mU3^2 - 
        1127*m3^7*mQ3^4*mU3^2 + 1201*m3^5*mQ3^6*mU3^2 - 
        288*m3^3*mQ3^8*mU3^2 - 60*m3*mQ3^10*mU3^2 + 120*m3^7*mQ3^2*msq^2*
         mU3^2 + 120*m3^5*mQ3^4*msq^2*mU3^2 - 240*m3^3*mQ3^6*msq^2*mU3^2 - 
        60*m3^7*msq^4*mU3^2 - 60*m3^5*mQ3^2*msq^4*mU3^2 + 
        120*m3^3*mQ3^4*msq^4*mU3^2 - 40*m3^9*mU3^4 - 233*m3^7*mQ3^2*mU3^4 + 
        959*m3^5*mQ3^4*mU3^4 - 1143*m3^3*mQ3^6*mU3^4 + 361*m3*mQ3^8*mU3^4 - 
        240*m3^5*mQ3^2*msq^2*mU3^4 + 120*m3^3*mQ3^4*msq^2*mU3^4 + 
        120*m3*mQ3^6*msq^2*mU3^4 + 120*m3^5*msq^4*mU3^4 - 
        60*m3^3*mQ3^2*msq^4*mU3^4 - 60*m3*mQ3^4*msq^4*mU3^4 + 61*m3^7*mU3^6 - 
        155*m3^5*mQ3^2*mU3^6 + 123*m3^3*mQ3^4*mU3^6 + 3*m3*mQ3^6*mU3^6 + 
        120*m3^3*mQ3^2*msq^2*mU3^6 - 120*m3*mQ3^4*msq^2*mU3^6 - 
        60*m3^3*msq^4*mU3^6 + 60*m3*mQ3^2*msq^4*mU3^6)*Log[mQ3^2/MR^2]^2)/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3))*
    Log[mU3^2/MR^2] + (128*a3^2*at*k^(3/2)*mt^2*(11*m3^9 + 7*m3^7*mQ3^2 - 
      14*m3^5*mQ3^4 - 37*m3^7*mU3^2 + 150*m3^5*mQ3^2*mU3^2 - 
      117*m3^3*mQ3^4*mU3^2 + 3*m3*mQ3^6*mU3^2 - 30*m3^3*mQ3^2*msq^2*mU3^2 + 
      30*m3*mQ3^4*msq^2*mU3^2 - 48*m3^5*mU3^4 - 47*m3^3*mQ3^2*mU3^4 + 
      82*m3*mQ3^4*mU3^4 + 30*m3^3*msq^2*mU3^4 - 30*m3*mQ3^2*msq^2*mU3^4 + 
      66*m3^3*mU3^6 - 53*m3*mQ3^2*mU3^6 - 3*m3*mU3^8)*Log[mU3^2/MR^2]^2)/
    (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) + 
   ((64*a3*at*k*m3*mt^2*(2*m3^2 - mQ3^2 - mU3^2))/(-mQ3^2 + mU3^2)^3 - 
     (64*a3^2*at*k^(3/2)*mt^2*(42*m3^13 - 6*m3^11*mQ3^2 - 67*m3^9*mQ3^4 + 
        45*m3^7*mQ3^6 - 120*m3^11*mU3^2 - 44*m3^9*mQ3^2*mU3^2 + 
        214*m3^7*mQ3^4*mU3^2 - 90*m3^5*mQ3^6*mU3^2 + 239*m3^9*mU3^4 - 
        67*m3^7*mQ3^2*mU3^4 - 195*m3^5*mQ3^4*mU3^4 + 59*m3^3*mQ3^6*mU3^4 - 
        252*m3^7*mU3^6 + 268*m3^5*mQ3^2*mU3^6 - 42*m3^3*mQ3^4*mU3^6 + 
        18*m3*mQ3^6*mU3^6 + 59*m3^5*mU3^8 - 55*m3^3*mQ3^2*mU3^8 - 
        6*m3*mQ3^4*mU3^8)*Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^3) - 
     (32*a3^2*at*k^(3/2)*mt^2*(2*m3^13 - 33*m3^11*mQ3^2 - 40*m3^9*mQ3^4 + 
        31*m3^7*mQ3^6 + 30*m3^5*mQ3^8 - 18*m3^3*mQ3^10 - 
        60*m3^7*mQ3^2*msq^4 + 120*m3^5*mQ3^4*msq^4 - 60*m3^3*mQ3^6*msq^4 - 
        43*m3^11*mU3^2 + 403*m3^9*mQ3^2*mU3^2 - 143*m3^7*mQ3^4*mU3^2 - 
        185*m3^5*mQ3^6*mU3^2 + 30*m3^3*mQ3^8*mU3^2 + 18*m3*mQ3^10*mU3^2 + 
        120*m3^7*mQ3^2*msq^2*mU3^2 - 240*m3^5*mQ3^4*msq^2*mU3^2 + 
        120*m3^3*mQ3^6*msq^2*mU3^2 + 60*m3^7*msq^4*mU3^2 - 
        60*m3^5*mQ3^2*msq^4*mU3^2 - 60*m3^3*mQ3^4*msq^4*mU3^2 + 
        60*m3*mQ3^6*msq^4*mU3^2 + 243*m3^9*mU3^4 - 1217*m3^7*mQ3^2*mU3^4 + 
        869*m3^5*mQ3^4*mU3^4 + 93*m3^3*mQ3^6*mU3^4 - 60*m3*mQ3^8*mU3^4 - 
        120*m3^7*msq^2*mU3^4 + 120*m3^5*mQ3^2*msq^2*mU3^4 + 
        120*m3^3*mQ3^4*msq^2*mU3^4 - 120*m3*mQ3^6*msq^2*mU3^4 - 
        60*m3^5*msq^4*mU3^4 + 120*m3^3*mQ3^2*msq^4*mU3^4 - 
        60*m3*mQ3^4*msq^4*mU3^4 - 375*m3^7*mU3^6 + 1351*m3^5*mQ3^2*mU3^6 - 
        1053*m3^3*mQ3^4*mU3^6 + 93*m3*mQ3^6*mU3^6 + 120*m3^5*msq^2*mU3^6 - 
        240*m3^3*mQ3^2*msq^2*mU3^6 + 120*m3*mQ3^4*msq^2*mU3^6 + 
        141*m3^5*mU3^8 - 408*m3^3*mQ3^2*mU3^8 + 271*m3*mQ3^4*mU3^8)*
       Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*
       (mQ3^2 - mU3^2)^3))*Log[mU3^2/MR^2]^2 + 
   (32*a3^2*at*k^(3/2)*mt^2*(2*m3^9 - 37*m3^7*mQ3^2 + 2*m3^5*mQ3^4 + 
      6*m3^3*mQ3^6 + 60*m3^3*mQ3^2*msq^4 - 35*m3^7*mU3^2 + 
      205*m3^5*mQ3^2*mU3^2 - 22*m3^3*mQ3^4*mU3^2 - 6*m3*mQ3^6*mU3^2 - 
      120*m3^3*mQ3^2*msq^2*mU3^2 - 60*m3^3*msq^4*mU3^2 - 
      60*m3*mQ3^2*msq^4*mU3^2 - 131*m3^5*mU3^4 - 281*m3^3*mQ3^2*mU3^4 + 
      20*m3*mQ3^4*mU3^4 + 120*m3^3*msq^2*mU3^4 + 120*m3*mQ3^2*msq^2*mU3^4 + 
      60*m3*msq^4*mU3^4 + 353*m3^3*mU3^6 + 81*m3*mQ3^2*mU3^6 - 
      120*m3*msq^2*mU3^6 - 157*m3*mU3^8)*Log[mU3^2/MR^2]^3)/
    (3*(m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)^3) + 
   Log[msq^2/MR^2]^2*((-48*a3^2*at*k^(3/2)*mt^2*
       ((40*m3*(mQ3^2 - msq^2)^2)/(3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) + 
        (40*m3*(-msq^2 + mU3^2)^2)/(3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)))*
       Log[mQ3^2/MR^2])/(mQ3^2 - mU3^2) + 
     (48*a3^2*at*k^(3/2)*mt^2*((40*m3*(mQ3^2 - msq^2)^2)/
         (3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) + (40*m3*(-msq^2 + mU3^2)^2)/
         (3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)))*Log[mU3^2/MR^2])/
      (mQ3^2 - mU3^2)) + Log[msq^2/MR^2]*
    ((a3^2*at*k^(3/2)*mt^2*((-1280*m3*msq^2)/((m3^2 - mQ3^2)*(m3^2 - mU3^2)*
           (mQ3^2 - mU3^2)) + (640*m3*(3*mQ3^2 + mU3^2))/
          (3*(mQ3^2 - mU3^2)^3)) + a3^2*at*k^(3/2)*mt^2*
        ((1280*(m3^5*mQ3^2 + m3^5*mU3^2 - 2*m3^3*mQ3^2*mU3^2))/
          ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) - 
         (640*m3^3*(2*m3^4 - mQ3^4 + 4*mQ3^2*mU3^2 - mU3^4 - 
            2*m3^2*(mQ3^2 + mU3^2)))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*
           (mQ3^2 - mU3^2)^3))*Log[m3^2/MR^2])*Log[mQ3^2/MR^2] + 
     a3^2*at*k^(3/2)*mt^2*((-1280*(2*m3*mQ3^2*msq^2 - m3*msq^4))/
        ((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^2) + 
       (320*m3*(2*m3^4 - mQ3^4 + 3*mQ3^2*mU3^2 - m3^2*(3*mQ3^2 + mU3^2)))/
        (3*(m3^2 - mQ3^2)*(mQ3^2 - mU3^2)^3))*Log[mQ3^2/MR^2]^2 + 
     (a3^2*at*k^(3/2)*mt^2*((1280*m3*msq^2)/((m3^2 - mQ3^2)*(m3^2 - mU3^2)*
           (mQ3^2 - mU3^2)) - (640*m3*(mQ3^2 + 3*mU3^2))/
          (3*(mQ3^2 - mU3^2)^3)) + a3^2*at*k^(3/2)*mt^2*
        ((-1280*(m3^5*mQ3^2 + m3^5*mU3^2 - 2*m3^3*mQ3^2*mU3^2))/
          ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) + 
         (640*m3^3*(2*m3^4 - mQ3^4 + 4*mQ3^2*mU3^2 - mU3^4 - 
            2*m3^2*(mQ3^2 + mU3^2)))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*
           (mQ3^2 - mU3^2)^3))*Log[m3^2/MR^2] + a3^2*at*k^(3/2)*mt^2*
        ((1280*(2*m3*mQ3^2*msq^2 - m3*msq^4))/((m3^2 - mQ3^2)^2*
           (mQ3^2 - mU3^2)^2) + (1280*(-(m3*msq^4) + 2*m3*msq^2*mU3^2))/
          ((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^2) + 
         (640*(-2*m3*mQ3^2*mU3^2 + m3^3*(mQ3^2 + mU3^2)))/
          (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^2))*
        Log[mQ3^2/MR^2])*Log[mU3^2/MR^2] + a3^2*at*k^(3/2)*mt^2*
      ((-1280*(-(m3*msq^4) + 2*m3*msq^2*mU3^2))/((m3^2 - mU3^2)^2*
         (-mQ3^2 + mU3^2)^2) + (320*m3*(2*m3^4 + 3*mQ3^2*mU3^2 - mU3^4 - 
          m3^2*(mQ3^2 + 3*mU3^2)))/(3*(m3^2 - mU3^2)*(-mQ3^2 + mU3^2)^3))*
      Log[mU3^2/MR^2]^2) + ((128*a3*at*k*m3*mt^2*(2*m3^2 - mQ3^2 - mU3^2))/
      (mQ3^2 - mU3^2)^3 + (128*a3^2*at*k^(3/2)*m3*mt^2*
       (2*m3^2 - mQ3^2 - mU3^2)*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 
        5*mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^3) + 
     (128*a3^2*at*k^(3/2)*m3*mt^2*(2*m3^2 - mQ3^2 - mU3^2)*
       (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
           m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*
       Log[m3^2/MR^2])/(mQ3^2 - mU3^2)^3 + a3^2*at*k^(3/2)*mt^2*
      ((64*m3*(m3^4 - 34*m3^2*mQ3^2 + 17*mQ3^4)*(2*m3^2 - mQ3^2 - mU3^2))/
        (3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^3) + 
       (128*(18*m3^5 + m3^3*mQ3^2 + 3*m3*mQ3^4 - 37*m3^3*mU3^2 - 
          7*m3*mQ3^2*mU3^2 + 22*m3*mU3^4))/(3*(m3^2 - mU3^2)^2*
         (-mQ3^2 + mU3^2)^2))*Log[mQ3^2/MR^2] + 
     (640*a3^2*at*k^(3/2)*m3*mt^2*(2*m3^2 - mQ3^2 - mU3^2)*Log[msq^2/MR^2])/
      (3*(mQ3^2 - mU3^2)^3) - (64*a3^2*at*k^(3/2)*mt^2*
       (2*m3^7 - 37*m3^5*mQ3^2 - 2*m3^3*mQ3^4 - 6*m3*mQ3^6 - 33*m3^5*mU3^2 + 
        110*m3^3*mQ3^2*mU3^2 + 20*m3*mQ3^4*mU3^2 - 6*m3^3*mU3^4 - 
        75*m3*mQ3^2*mU3^4 + 27*m3*mU3^6)*Log[mU3^2/MR^2])/
      (3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^3))*PolyLog[2, 1 - m3^2/mQ3^2] + 
   ((-1280*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*(m3^3*mQ3^2 - 2*m3^3*msq^2 + 
        m3*mQ3^2*msq^2 + m3^3*mU3^2 - 2*m3*mQ3^2*mU3^2 + m3*msq^2*mU3^2)*
       Log[mQ3^2/MR^2])/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) + 
     (1280*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*(m3^3*mQ3^2 - 2*m3^3*msq^2 + 
        m3*mQ3^2*msq^2 + m3^3*mU3^2 - 2*m3*mQ3^2*mU3^2 + m3*msq^2*mU3^2)*
       Log[mU3^2/MR^2])/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)))*
    PolyLog[2, 1 - m3^2/msq^2] + 
   ((-1280*a3^2*at*k^(3/2)*m3*(mQ3^2 - msq^2)^2*mt^2*Log[mQ3^2/MR^2])/
      ((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^2) + 
     (1280*a3^2*at*k^(3/2)*m3*(mQ3^2 - msq^2)^2*mt^2*Log[mU3^2/MR^2])/
      ((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^2))*PolyLog[2, 1 - msq^2/mQ3^2] + 
   ((128*a3*at*k*m3*mt^2*(-2*m3^2 + mQ3^2 + mU3^2))/(mQ3^2 - mU3^2)^3 + 
     (128*a3^2*at*k^(3/2)*m3*mt^2*(-2*m3^2 + mQ3^2 + mU3^2)*
       (-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2))/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^3) + 
     (128*a3^2*at*k^(3/2)*m3*mt^2*(-2*m3^2 + mQ3^2 + mU3^2)*
       (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
           m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*
       Log[m3^2/MR^2])/(mQ3^2 - mU3^2)^3 - 
     (64*a3^2*at*k^(3/2)*mt^2*(2*m3^7 - 33*m3^5*mQ3^2 - 6*m3^3*mQ3^4 + 
        27*m3*mQ3^6 - 37*m3^5*mU3^2 + 110*m3^3*mQ3^2*mU3^2 - 
        75*m3*mQ3^4*mU3^2 - 2*m3^3*mU3^4 + 20*m3*mQ3^2*mU3^4 - 6*m3*mU3^6)*
       Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^3) + 
     (640*a3^2*at*k^(3/2)*m3*mt^2*(-2*m3^2 + mQ3^2 + mU3^2)*Log[msq^2/MR^2])/
      (3*(mQ3^2 - mU3^2)^3) + a3^2*at*k^(3/2)*mt^2*
      ((64*m3*(-2*m3^2 + mQ3^2 + mU3^2)*(m3^4 - 34*m3^2*mU3^2 + 17*mU3^4))/
        (3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) + 
       (128*(18*m3^5 - 37*m3^3*mQ3^2 + 22*m3*mQ3^4 + m3^3*mU3^2 - 
          7*m3*mQ3^2*mU3^2 + 3*m3*mU3^4))/(3*(m3^2 - mQ3^2)^2*
         (mQ3^2 - mU3^2)^2))*Log[mU3^2/MR^2])*PolyLog[2, 1 - m3^2/mU3^2] + 
   ((128*a3^2*at*k^(3/2)*mt^2*(5*m3^5 - 5*m3^3*mQ3^2 + 3*m3*mQ3^4 - 
        5*m3^3*mU3^2 - m3*mQ3^2*mU3^2 + 3*m3*mU3^4)*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) - 
     (128*a3^2*at*k^(3/2)*mt^2*(5*m3^5 - 5*m3^3*mQ3^2 + 3*m3*mQ3^4 - 
        5*m3^3*mU3^2 - m3*mQ3^2*mU3^2 + 3*m3*mU3^4)*Log[mU3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*PolyLog[2, 1 - mQ3^2/mU3^2] + 
   ((1280*a3^2*at*k^(3/2)*m3*mt^2*(-msq^2 + mU3^2)^2*Log[mQ3^2/MR^2])/
      ((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^2) - 
     (1280*a3^2*at*k^(3/2)*m3*mt^2*(-msq^2 + mU3^2)^2*Log[mU3^2/MR^2])/
      ((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^2))*PolyLog[2, 1 - msq^2/mU3^2]) + 
 Xt^5*((128*a3^2*at*k^(3/2)*mt^2*(14*m3^3 - 3*m3*mQ3^2 - 30*m3*msq^2 - 
      3*m3*mU3^2))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^2) - 
   (128*a3^2*at*k^(3/2)*mt^2*(8*m3^9*mQ3^2 - 8*m3^7*mQ3^4 + 8*m3^9*mU3^2 + 
      123*m3^7*mQ3^2*mU3^2 - 115*m3^5*mQ3^4*mU3^2 + 3*m3^3*mQ3^6*mU3^2 - 
      60*m3^5*mQ3^2*msq^2*mU3^2 + 30*m3^3*mQ3^4*msq^2*mU3^2 - 8*m3^7*mU3^4 - 
      115*m3^5*mQ3^2*mU3^4 + 101*m3^3*mQ3^4*mU3^4 + 30*m3^3*mQ3^2*msq^2*
       mU3^4 + 3*m3^3*mQ3^2*mU3^6)*Log[m3^2/MR^2])/
    (3*mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2) + 
   (128*a3^2*at*k^(3/2)*mt^2*(8*m3^13*mQ3^2 - 16*m3^11*mQ3^4 + 8*m3^9*mQ3^6 + 
      8*m3^13*mU3^2 + 7*m3^11*mQ3^2*mU3^2 - 54*m3^9*mQ3^4*mU3^2 + 
      47*m3^7*mQ3^6*mU3^2 - 16*m3^11*mU3^4 - 54*m3^9*mQ3^2*mU3^4 + 
      188*m3^7*mQ3^4*mU3^4 - 134*m3^5*mQ3^6*mU3^4 + 8*m3^9*mU3^6 + 
      47*m3^7*mQ3^2*mU3^6 - 134*m3^5*mQ3^4*mU3^6 + 87*m3^3*mQ3^6*mU3^6)*
     Log[m3^2/MR^2]^2)/(3*mQ3^2*(m3^2 - mQ3^2)^3*mU3^2*(m3^2 - mU3^2)^3*
     (mQ3^2 - mU3^2)^2) + (64*a3^2*at*k^(3/2)*mt^2*
     (16*m3^7*mQ3^2 - 16*m3^5*mQ3^4 + 48*m3^7*mU3^2 + 172*m3^5*mQ3^2*mU3^2 - 
      185*m3^3*mQ3^4*mU3^2 + 3*m3*mQ3^6*mU3^2 - 30*m3^3*mQ3^2*msq^2*mU3^2 - 
      30*m3*mQ3^4*msq^2*mU3^2 - 62*m3^5*mU3^4 - 166*m3^3*mQ3^2*mU3^4 + 
      184*m3*mQ3^4*mU3^4 + 30*m3^3*msq^2*mU3^4 + 30*m3*mQ3^2*msq^2*mU3^4 + 
      3*m3^3*mU3^6 + 3*m3*mQ3^2*mU3^6)*Log[mQ3^2/MR^2])/
    (3*(m3^2 - mQ3^2)^2*mU3^2*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^3) - 
   (64*a3^2*at*k^(3/2)*mt^2*(16*m3^11*mQ3^2 - 32*m3^9*mQ3^4 + 16*m3^7*mQ3^6 + 
      52*m3^11*mU3^2 - 362*m3^9*mQ3^2*mU3^2 + 558*m3^7*mQ3^4*mU3^2 - 
      219*m3^5*mQ3^6*mU3^2 + 3*m3^3*mQ3^8*mU3^2 + 60*m3^7*mQ3^2*msq^2*mU3^2 - 
      90*m3^5*mQ3^4*msq^2*mU3^2 + 30*m3^3*mQ3^6*msq^2*mU3^2 - 
      276*m3^9*mU3^4 + 770*m3^7*mQ3^2*mU3^4 - 737*m3^5*mQ3^4*mU3^4 + 
      179*m3^3*mQ3^6*mU3^4 + 60*m3^7*msq^2*mU3^4 - 120*m3^5*mQ3^2*msq^2*
       mU3^4 + 60*m3^3*mQ3^4*msq^2*mU3^4 + 214*m3^7*mU3^6 - 
      407*m3^5*mQ3^2*mU3^6 + 213*m3^3*mQ3^4*mU3^6 + 12*m3*mQ3^6*mU3^6 - 
      30*m3^5*msq^2*mU3^6 + 30*m3^3*mQ3^2*msq^2*mU3^6 - 3*m3^5*mU3^8 + 
      3*m3^3*mQ3^2*mU3^8)*Log[m3^2/MR^2]*Log[mQ3^2/MR^2])/
    (3*(m3^2 - mQ3^2)^3*mU3^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) - 
   (64*a3^2*at*k^(3/2)*mt^2*(39*m3^11*mQ3^2 - 94*m3^9*mQ3^4 + 63*m3^7*mQ3^6 + 
      39*m3^11*mU3^2 - 188*m3^9*mQ3^2*mU3^2 + 283*m3^7*mQ3^4*mU3^2 - 
      142*m3^5*mQ3^6*mU3^2 - 94*m3^9*mU3^4 + 283*m3^7*mQ3^2*mU3^4 - 
      284*m3^5*mQ3^4*mU3^4 + 87*m3^3*mQ3^6*mU3^4 + 63*m3^7*mU3^6 - 
      142*m3^5*mQ3^2*mU3^6 + 87*m3^3*mQ3^4*mU3^6)*Log[m3^2/MR^2]^2*
     Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
     (mQ3^2 - mU3^2)^3) - (64*a3^2*at*k^(3/2)*mt^2*
     (30*m3^11*mQ3^2 + 179*m3^9*mQ3^4 - 442*m3^7*mQ3^6 + 225*m3^5*mQ3^8 - 
      11*m3^3*mQ3^10 + 3*m3*mQ3^12 - 90*m3^7*mQ3^4*msq^2 + 
      90*m3^5*mQ3^6*msq^2 + 30*m3^7*mQ3^2*msq^4 - 30*m3^5*mQ3^4*msq^4 + 
      34*m3^11*mU3^2 - 36*m3^9*mQ3^2*mU3^2 - 470*m3^7*mQ3^4*mU3^2 + 
      923*m3^5*mQ3^6*mU3^2 - 396*m3^3*mQ3^8*mU3^2 - 7*m3*mQ3^10*mU3^2 + 
      30*m3^7*mQ3^2*msq^2*mU3^2 + 150*m3^5*mQ3^4*msq^2*mU3^2 - 
      180*m3^3*mQ3^6*msq^2*mU3^2 - 30*m3^7*msq^4*mU3^2 - 
      30*m3^5*mQ3^2*msq^4*mU3^2 + 60*m3^3*mQ3^4*msq^4*mU3^2 - 69*m3^9*mU3^4 - 
      14*m3^7*mQ3^2*mU3^4 + 427*m3^5*mQ3^4*mU3^4 - 608*m3^3*mQ3^6*mU3^4 + 
      216*m3*mQ3^8*mU3^4 - 60*m3^5*mQ3^2*msq^2*mU3^4 - 
      30*m3^3*mQ3^4*msq^2*mU3^4 + 90*m3*mQ3^6*msq^2*mU3^4 + 
      60*m3^5*msq^4*mU3^4 - 30*m3^3*mQ3^2*msq^4*mU3^4 - 
      30*m3*mQ3^4*msq^4*mU3^4 + 38*m3^7*mU3^6 + 5*m3^5*mQ3^2*mU3^6 - 
      94*m3^3*mQ3^4*mU3^6 + 67*m3*mQ3^6*mU3^6 + 30*m3^3*mQ3^2*msq^2*mU3^6 - 
      30*m3*mQ3^4*msq^2*mU3^6 - 30*m3^3*msq^4*mU3^6 + 
      30*m3*mQ3^2*msq^4*mU3^6 - 3*m3^3*mQ3^2*mU3^8 + 3*m3*mQ3^4*mU3^8)*
     Log[mQ3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*
     (mQ3^2 - mU3^2)^4) + (32*a3^2*at*k^(3/2)*mt^2*
     (36*m3^11*mQ3^2 - 159*m3^9*mQ3^4 + 216*m3^7*mQ3^6 - 61*m3^5*mQ3^8 + 
      100*m3^11*mU3^2 - 392*m3^9*mQ3^2*mU3^2 + 585*m3^7*mQ3^4*mU3^2 - 
      384*m3^5*mQ3^6*mU3^2 + 59*m3^3*mQ3^8*mU3^2 - 169*m3^9*mU3^4 + 
      508*m3^7*mQ3^2*mU3^4 - 535*m3^5*mQ3^4*mU3^4 + 152*m3^3*mQ3^6*mU3^4 + 
      12*m3*mQ3^8*mU3^4 + 75*m3^7*mU3^6 - 148*m3^5*mQ3^2*mU3^6 + 
      93*m3^3*mQ3^4*mU3^6 + 12*m3*mQ3^6*mU3^6)*Log[m3^2/MR^2]*
     Log[mQ3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*
     (mQ3^2 - mU3^2)^4) + (32*a3^2*at*k^(3/2)*mt^2*
     (14*m3^11*mQ3^4 + 86*m3^9*mQ3^6 - 224*m3^7*mQ3^8 + 113*m3^5*mQ3^10 - 
      8*m3^3*mQ3^12 + 3*m3*mQ3^14 - 60*m3^7*mQ3^6*msq^2 + 
      60*m3^5*mQ3^8*msq^2 + 30*m3^7*mQ3^4*msq^4 - 30*m3^5*mQ3^6*msq^4 - 
      32*m3^11*mQ3^2*mU3^2 + 117*m3^9*mQ3^4*mU3^2 - 362*m3^7*mQ3^6*mU3^2 + 
      510*m3^5*mQ3^8*mU3^2 - 194*m3^3*mQ3^10*mU3^2 - 7*m3*mQ3^12*mU3^2 + 
      120*m3^5*mQ3^6*msq^2*mU3^2 - 120*m3^3*mQ3^8*msq^2*mU3^2 - 
      60*m3^5*mQ3^4*msq^4*mU3^2 + 60*m3^3*mQ3^6*msq^4*mU3^2 + 
      18*m3^11*mU3^4 - 6*m3^9*mQ3^2*mU3^4 - 182*m3^7*mQ3^4*mU3^4 + 
      450*m3^5*mQ3^6*mU3^4 - 388*m3^3*mQ3^8*mU3^4 + 108*m3*mQ3^10*mU3^4 + 
      60*m3^7*mQ3^2*msq^2*mU3^4 - 60*m3^5*mQ3^4*msq^2*mU3^4 - 
      60*m3^3*mQ3^6*msq^2*mU3^4 + 60*m3*mQ3^8*msq^2*mU3^4 - 
      30*m3^7*msq^4*mU3^4 + 30*m3^5*mQ3^2*msq^4*mU3^4 + 
      30*m3^3*mQ3^4*msq^4*mU3^4 - 30*m3*mQ3^6*msq^4*mU3^4 - 37*m3^9*mU3^6 + 
      106*m3^7*mQ3^2*mU3^6 - 30*m3^5*mQ3^4*mU3^6 - 158*m3^3*mQ3^6*mU3^6 + 
      87*m3*mQ3^8*mU3^6 - 120*m3^5*mQ3^2*msq^2*mU3^6 + 
      120*m3^3*mQ3^4*msq^2*mU3^6 + 60*m3^5*msq^4*mU3^6 - 
      60*m3^3*mQ3^2*msq^4*mU3^6 + 22*m3^7*mU3^8 - 83*m3^5*mQ3^2*mU3^8 + 
      108*m3^3*mQ3^4*mU3^8 - 31*m3*mQ3^6*mU3^8 + 60*m3^3*mQ3^2*msq^2*mU3^8 - 
      60*m3*mQ3^4*msq^2*mU3^8 - 30*m3^3*msq^4*mU3^8 + 
      30*m3*mQ3^2*msq^4*mU3^8)*Log[mQ3^2/MR^2]^3)/
    (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^5) - 
   (1280*a3^2*at*k^(3/2)*m3*msq^2*mt^2*Log[msq^2/MR^2])/
    ((m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^2) - 
   (64*a3^2*at*k^(3/2)*mt^2*(48*m3^7*mQ3^2 - 62*m3^5*mQ3^4 + 3*m3^3*mQ3^6 + 
      30*m3^3*mQ3^4*msq^2 + 16*m3^7*mU3^2 + 172*m3^5*mQ3^2*mU3^2 - 
      166*m3^3*mQ3^4*mU3^2 + 3*m3*mQ3^6*mU3^2 - 30*m3^3*mQ3^2*msq^2*mU3^2 + 
      30*m3*mQ3^4*msq^2*mU3^2 - 16*m3^5*mU3^4 - 185*m3^3*mQ3^2*mU3^4 + 
      184*m3*mQ3^4*mU3^4 - 30*m3*mQ3^2*msq^2*mU3^4 + 3*m3*mQ3^2*mU3^6)*
     Log[mU3^2/MR^2])/(3*mQ3^2*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*
     (mQ3^2 - mU3^2)^3) + 
   ((64*a3^2*at*k^(3/2)*mt^2*(52*m3^11*mQ3^2 - 276*m3^9*mQ3^4 + 
        214*m3^7*mQ3^6 - 3*m3^5*mQ3^8 + 60*m3^7*mQ3^4*msq^2 - 
        30*m3^5*mQ3^6*msq^2 + 16*m3^11*mU3^2 - 362*m3^9*mQ3^2*mU3^2 + 
        770*m3^7*mQ3^4*mU3^2 - 407*m3^5*mQ3^6*mU3^2 + 3*m3^3*mQ3^8*mU3^2 + 
        60*m3^7*mQ3^2*msq^2*mU3^2 - 120*m3^5*mQ3^4*msq^2*mU3^2 + 
        30*m3^3*mQ3^6*msq^2*mU3^2 - 32*m3^9*mU3^4 + 558*m3^7*mQ3^2*mU3^4 - 
        737*m3^5*mQ3^4*mU3^4 + 213*m3^3*mQ3^6*mU3^4 - 90*m3^5*mQ3^2*msq^2*
         mU3^4 + 60*m3^3*mQ3^4*msq^2*mU3^4 + 16*m3^7*mU3^6 - 
        219*m3^5*mQ3^2*mU3^6 + 179*m3^3*mQ3^4*mU3^6 + 12*m3*mQ3^6*mU3^6 + 
        30*m3^3*mQ3^2*msq^2*mU3^6 + 3*m3^3*mQ3^2*mU3^8)*Log[m3^2/MR^2])/
      (3*mQ3^2*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^3) + 
     (64*a3^2*at*k^(3/2)*mt^2*(64*m3^9*mQ3^2 + 44*m3^7*mQ3^4 - 
        102*m3^5*mQ3^6 + 13*m3^3*mQ3^8 - 6*m3*mQ3^10 - 30*m3^5*mQ3^4*msq^2 + 
        64*m3^9*mU3^2 + 188*m3^7*mQ3^2*mU3^2 - 638*m3^5*mQ3^4*mU3^2 + 
        352*m3^3*mQ3^6*mU3^2 + 14*m3*mQ3^8*mU3^2 - 60*m3^5*mQ3^2*msq^2*
         mU3^2 + 120*m3^3*mQ3^4*msq^2*mU3^2 - 30*m3*mQ3^6*msq^2*mU3^2 + 
        44*m3^7*mU3^4 - 668*m3^5*mQ3^2*mU3^4 + 920*m3^3*mQ3^4*mU3^4 - 
        298*m3*mQ3^6*mU3^4 - 30*m3^5*msq^2*mU3^4 + 120*m3^3*mQ3^2*msq^2*
         mU3^4 - 60*m3*mQ3^4*msq^2*mU3^4 - 92*m3^5*mU3^6 + 
        372*m3^3*mQ3^2*mU3^6 - 268*m3*mQ3^4*mU3^6 - 30*m3*mQ3^2*msq^2*mU3^6 + 
        3*m3^3*mU3^8 - 6*m3*mQ3^2*mU3^8)*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^4))*
    Log[mU3^2/MR^2] + 
   ((64*a3^2*at*k^(3/2)*mt^2*(39*m3^11*mQ3^2 - 94*m3^9*mQ3^4 + 
        63*m3^7*mQ3^6 + 39*m3^11*mU3^2 - 188*m3^9*mQ3^2*mU3^2 + 
        283*m3^7*mQ3^4*mU3^2 - 142*m3^5*mQ3^6*mU3^2 - 94*m3^9*mU3^4 + 
        283*m3^7*mQ3^2*mU3^4 - 284*m3^5*mQ3^4*mU3^4 + 87*m3^3*mQ3^6*mU3^4 + 
        63*m3^7*mU3^6 - 142*m3^5*mQ3^2*mU3^6 + 87*m3^3*mQ3^4*mU3^6)*
       Log[m3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
       (mQ3^2 - mU3^2)^3) - (128*a3^2*at*k^(3/2)*mt^2*
       (34*m3^13*mQ3^2 - 107*m3^11*mQ3^4 + 115*m3^9*mQ3^6 - 34*m3^7*mQ3^8 + 
        34*m3^13*mU3^2 - 214*m3^11*mQ3^2*mU3^2 + 411*m3^9*mQ3^4*mU3^2 - 
        314*m3^7*mQ3^6*mU3^2 + 67*m3^5*mQ3^8*mU3^2 - 107*m3^11*mU3^4 + 
        411*m3^9*mQ3^2*mU3^4 - 560*m3^7*mQ3^4*mU3^4 + 291*m3^5*mQ3^6*mU3^4 - 
        35*m3^3*mQ3^8*mU3^4 + 115*m3^9*mU3^6 - 314*m3^7*mQ3^2*mU3^6 + 
        291*m3^5*mQ3^4*mU3^6 - 70*m3^3*mQ3^6*mU3^6 - 6*m3*mQ3^8*mU3^6 - 
        34*m3^7*mU3^8 + 67*m3^5*mQ3^2*mU3^8 - 35*m3^3*mQ3^4*mU3^8 - 
        6*m3*mQ3^6*mU3^8)*Log[m3^2/MR^2]*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^4) + 
     (32*a3^2*at*k^(3/2)*mt^2*(18*m3^11*mQ3^4 - 149*m3^9*mQ3^6 + 
        264*m3^7*mQ3^8 - 132*m3^5*mQ3^10 + 24*m3^3*mQ3^12 - 9*m3*mQ3^14 + 
        60*m3^7*mQ3^6*msq^2 - 60*m3^5*mQ3^8*msq^2 - 30*m3^7*mQ3^4*msq^4 + 
        30*m3^5*mQ3^6*msq^4 - 32*m3^11*mQ3^2*mU3^2 - 182*m3^9*mQ3^4*mU3^2 + 
        753*m3^7*mQ3^6*mU3^2 - 823*m3^5*mQ3^8*mU3^2 + 231*m3^3*mQ3^10*mU3^2 + 
        21*m3*mQ3^12*mU3^2 - 120*m3^5*mQ3^6*msq^2*mU3^2 + 
        120*m3^3*mQ3^8*msq^2*mU3^2 + 60*m3^5*mQ3^4*msq^4*mU3^2 - 
        60*m3^3*mQ3^6*msq^4*mU3^2 + 14*m3^11*mU3^4 - 155*m3^9*mQ3^2*mU3^4 + 
        719*m3^7*mQ3^4*mU3^4 - 1267*m3^5*mQ3^6*mU3^4 + 879*m3^3*mQ3^8*mU3^4 - 
        190*m3*mQ3^10*mU3^4 - 60*m3^7*mQ3^2*msq^2*mU3^4 + 
        60*m3^5*mQ3^4*msq^2*mU3^4 + 60*m3^3*mQ3^6*msq^2*mU3^4 - 
        60*m3*mQ3^8*msq^2*mU3^4 + 30*m3^7*msq^4*mU3^4 - 
        30*m3^5*mQ3^2*msq^4*mU3^4 - 30*m3^3*mQ3^4*msq^4*mU3^4 + 
        30*m3*mQ3^6*msq^4*mU3^4 + 6*m3^9*mU3^6 + 207*m3^7*mQ3^2*mU3^6 - 
        681*m3^5*mQ3^4*mU3^6 + 761*m3^3*mQ3^6*mU3^6 - 261*m3*mQ3^8*mU3^6 + 
        120*m3^5*mQ3^2*msq^2*mU3^6 - 120*m3^3*mQ3^4*msq^2*mU3^6 - 
        60*m3^5*msq^4*mU3^6 + 60*m3^3*mQ3^2*msq^4*mU3^6 - 23*m3^7*mU3^8 + 
        23*m3^5*mQ3^2*mU3^8 + 25*m3^3*mQ3^4*mU3^8 - 41*m3*mQ3^6*mU3^8 - 
        60*m3^3*mQ3^2*msq^2*mU3^8 + 60*m3*mQ3^4*msq^2*mU3^8 + 
        30*m3^3*msq^4*mU3^8 - 30*m3*mQ3^2*msq^4*mU3^8)*Log[mQ3^2/MR^2]^2)/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^5))*
    Log[mU3^2/MR^2] - (64*a3^2*at*k^(3/2)*mt^2*(34*m3^9*mQ3^2 - 
      35*m3^7*mQ3^4 - 2*m3^5*mQ3^6 + 3*m3^3*mQ3^8 - 30*m3^5*mQ3^2*msq^4 + 
      30*m3^3*mQ3^4*msq^4 + 30*m3^9*mU3^2 - 6*m3^7*mQ3^2*mU3^2 - 
      5*m3^5*mQ3^4*mU3^2 - 5*m3^3*mQ3^6*mU3^2 - 3*m3*mQ3^8*mU3^2 + 
      30*m3^5*mQ3^2*msq^2*mU3^2 - 30*m3^3*mQ3^4*msq^2*mU3^2 + 
      30*m3^5*msq^4*mU3^2 - 30*m3*mQ3^4*msq^4*mU3^2 + 179*m3^7*mU3^4 - 
      306*m3^5*mQ3^2*mU3^4 + 106*m3^3*mQ3^4*mU3^4 + 7*m3*mQ3^6*mU3^4 - 
      90*m3^5*msq^2*mU3^4 + 60*m3^3*mQ3^2*msq^2*mU3^4 + 
      30*m3*mQ3^4*msq^2*mU3^4 - 30*m3^3*msq^4*mU3^4 + 
      30*m3*mQ3^2*msq^4*mU3^4 - 437*m3^5*mU3^6 + 511*m3^3*mQ3^2*mU3^6 - 
      82*m3*mQ3^4*mU3^6 + 90*m3^3*msq^2*mU3^6 - 90*m3*mQ3^2*msq^2*mU3^6 + 
      215*m3^3*mU3^8 - 201*m3*mQ3^2*mU3^8 - 3*m3*mU3^10)*Log[mU3^2/MR^2]^2)/
    (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^4) + 
   ((32*a3^2*at*k^(3/2)*mt^2*(100*m3^11*mQ3^2 - 169*m3^9*mQ3^4 + 
        75*m3^7*mQ3^6 + 36*m3^11*mU3^2 - 392*m3^9*mQ3^2*mU3^2 + 
        508*m3^7*mQ3^4*mU3^2 - 148*m3^5*mQ3^6*mU3^2 - 159*m3^9*mU3^4 + 
        585*m3^7*mQ3^2*mU3^4 - 535*m3^5*mQ3^4*mU3^4 + 93*m3^3*mQ3^6*mU3^4 + 
        216*m3^7*mU3^6 - 384*m3^5*mQ3^2*mU3^6 + 152*m3^3*mQ3^4*mU3^6 + 
        12*m3*mQ3^6*mU3^6 - 61*m3^5*mU3^8 + 59*m3^3*mQ3^2*mU3^8 + 
        12*m3*mQ3^4*mU3^8)*Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^4) - 
     (32*a3^2*at*k^(3/2)*mt^2*(14*m3^11*mQ3^4 + 6*m3^9*mQ3^6 - 8*m3^7*mQ3^8 - 
        15*m3^5*mQ3^10 + 9*m3^3*mQ3^12 + 30*m3^7*mQ3^4*msq^4 - 
        60*m3^5*mQ3^6*msq^4 + 30*m3^3*mQ3^8*msq^4 - 32*m3^11*mQ3^2*mU3^2 - 
        155*m3^9*mQ3^4*mU3^2 + 177*m3^7*mQ3^6*mU3^2 + 23*m3^5*mQ3^8*mU3^2 - 
        6*m3^3*mQ3^10*mU3^2 - 9*m3*mQ3^12*mU3^2 - 60*m3^7*mQ3^4*msq^2*mU3^2 + 
        120*m3^5*mQ3^6*msq^2*mU3^2 - 60*m3^3*mQ3^8*msq^2*mU3^2 - 
        30*m3^5*mQ3^4*msq^4*mU3^2 + 60*m3^3*mQ3^6*msq^4*mU3^2 - 
        30*m3*mQ3^8*msq^4*mU3^2 + 18*m3^11*mU3^4 - 182*m3^9*mQ3^2*mU3^4 + 
        719*m3^7*mQ3^4*mU3^4 - 621*m3^5*mQ3^6*mU3^4 + 25*m3^3*mQ3^8*mU3^4 + 
        21*m3*mQ3^10*mU3^4 + 60*m3^5*mQ3^4*msq^2*mU3^4 - 
        120*m3^3*mQ3^6*msq^2*mU3^4 + 60*m3*mQ3^8*msq^2*mU3^4 - 
        30*m3^7*msq^4*mU3^4 + 60*m3^5*mQ3^2*msq^4*mU3^4 - 
        30*m3^3*mQ3^4*msq^4*mU3^4 - 149*m3^9*mU3^6 + 783*m3^7*mQ3^2*mU3^6 - 
        1297*m3^5*mQ3^4*mU3^6 + 731*m3^3*mQ3^6*mU3^6 - 56*m3*mQ3^8*mU3^6 + 
        60*m3^7*msq^2*mU3^6 - 120*m3^5*mQ3^2*msq^2*mU3^6 + 
        60*m3^3*mQ3^4*msq^2*mU3^6 + 30*m3^5*msq^4*mU3^6 - 
        60*m3^3*mQ3^2*msq^4*mU3^6 + 30*m3*mQ3^4*msq^4*mU3^6 + 
        249*m3^7*mU3^8 - 868*m3^5*mQ3^2*mU3^8 + 894*m3^3*mQ3^4*mU3^8 - 
        261*m3*mQ3^6*mU3^8 - 60*m3^5*msq^2*mU3^8 + 120*m3^3*mQ3^2*msq^2*
         mU3^8 - 60*m3*mQ3^4*msq^2*mU3^8 - 102*m3^5*mU3^10 + 
        267*m3^3*mQ3^2*mU3^10 - 175*m3*mQ3^4*mU3^10)*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^5))*
    Log[mU3^2/MR^2]^2 + (32*a3^2*at*k^(3/2)*mt^2*(18*m3^7*mQ3^4 - 
      m3^5*mQ3^6 - 3*m3^3*mQ3^8 - 30*m3^3*mQ3^4*msq^4 - 32*m3^7*mQ3^2*mU3^2 - 
      70*m3^5*mQ3^4*mU3^2 + 8*m3^3*mQ3^6*mU3^2 + 3*m3*mQ3^8*mU3^2 + 
      60*m3^3*mQ3^4*msq^2*mU3^2 + 30*m3*mQ3^4*msq^4*mU3^2 + 14*m3^7*mU3^4 + 
      145*m3^5*mQ3^2*mU3^4 + 94*m3^3*mQ3^4*mU3^4 - 7*m3*mQ3^6*mU3^4 - 
      60*m3*mQ3^4*msq^2*mU3^4 + 30*m3^3*msq^4*mU3^4 + 86*m3^5*mU3^6 - 
      200*m3^3*mQ3^2*mU3^6 - 26*m3*mQ3^4*mU3^6 - 60*m3^3*msq^2*mU3^6 - 
      30*m3*msq^4*mU3^6 - 219*m3^3*mU3^8 + 87*m3*mQ3^2*mU3^8 + 
      60*m3*msq^2*mU3^8 + 103*m3*mU3^10)*Log[mU3^2/MR^2]^3)/
    (3*(m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)^5) + 
   Log[msq^2/MR^2]^2*((-48*a3^2*at*k^(3/2)*mt^2*
       ((40*m3*(mQ3^2 - msq^2)^2)/(3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) + 
        (40*m3*(-msq^2 + mU3^2)^2)/(3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2))))/
      (mQ3^2 - mU3^2)^2 + (24*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*
       ((40*m3*(mQ3^2 - msq^2)^2)/(3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) + 
        (40*m3*(-msq^2 + mU3^2)^2)/(3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)))*
       Log[mQ3^2/MR^2])/(mQ3^2 - mU3^2)^3 - 
     (24*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*
       ((40*m3*(mQ3^2 - msq^2)^2)/(3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) + 
        (40*m3*(-msq^2 + mU3^2)^2)/(3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)))*
       Log[mU3^2/MR^2])/(mQ3^2 - mU3^2)^3) + 
   Log[msq^2/MR^2]*((640*a3^2*at*k^(3/2)*mt^2*(m3^7 + 5*m3^5*mQ3^2 + 
        5*m3^5*mU3^2 - 11*m3^3*mQ3^2*mU3^2)*Log[m3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2) + 
     (a3^2*at*k^(3/2)*mt^2*((-640*m3*mQ3^2)/(3*(m3^2 - mQ3^2)*
           (mQ3^2 - mU3^2)^3) - (1280*(2*m3*mQ3^2*msq^2 - m3*msq^4))/
          ((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^3) + 
         (640*m3*msq^2*(mQ3^2 + mU3^2))/((m3^2 - mQ3^2)*(m3^2 - mU3^2)*
           (mQ3^2 - mU3^2)^3)) - (320*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*
         (m3^7 + 5*m3^5*mQ3^2 + 5*m3^5*mU3^2 - 11*m3^3*mQ3^2*mU3^2)*
         Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*
         (mQ3^2 - mU3^2)^3))*Log[mQ3^2/MR^2] + 
     (320*a3^2*at*k^(3/2)*(m3^3*mQ3^2 - m3*mQ3^4 + 12*m3*mQ3^2*msq^2 - 
        6*m3*msq^4)*mt^2*(mQ3^2 + mU3^2)*Log[mQ3^2/MR^2]^2)/
      (3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^4) + 
     (a3^2*at*k^(3/2)*mt^2*((-640*m3*mU3^2)/(3*(m3^2 - mU3^2)*
           (-mQ3^2 + mU3^2)^3) - (640*m3*msq^2*(mQ3^2 + mU3^2))/
          ((m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^3) - 
         (1280*(-(m3*msq^4) + 2*m3*msq^2*mU3^2))/((m3^2 - mU3^2)^2*
           (-mQ3^2 + mU3^2)^3)) + (320*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*
         (m3^7 + 5*m3^5*mQ3^2 + 5*m3^5*mU3^2 - 11*m3^3*mQ3^2*mU3^2)*
         Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*
         (mQ3^2 - mU3^2)^3) + a3^2*at*k^(3/2)*mt^2*
        ((-640*(2*m3*mQ3^2*msq^2 - m3*msq^4)*(mQ3^2 + mU3^2))/
          ((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^4) - 
         (640*(mQ3^2 + mU3^2)*(-(m3*msq^4) + 2*m3*msq^2*mU3^2))/
          ((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^4) - 
         (320*m3*(mQ3^2 + mU3^2)*(-2*mQ3^2*mU3^2 + m3^2*(mQ3^2 + mU3^2)))/
          (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^4))*
        Log[mQ3^2/MR^2])*Log[mU3^2/MR^2] + 
     (320*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*(-6*m3*msq^4 + m3^3*mU3^2 + 
        12*m3*msq^2*mU3^2 - m3*mU3^4)*Log[mU3^2/MR^2]^2)/
      (3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^4)) + 
   ((-128*a3^2*at*k^(3/2)*mt^2*(18*m3^5 + m3^3*mQ3^2 + 3*m3*mQ3^4 - 
        37*m3^3*mU3^2 - 7*m3*mQ3^2*mU3^2 + 22*m3*mU3^4))/
      (3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^3) + 
     (1024*a3^2*at*k^(3/2)*m3^3*mt^2*(-2*m3^2 + mQ3^2 + mU3^2)*
       Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^3) + 
     a3^2*at*k^(3/2)*mt^2*((-1024*m3*mQ3^2*(-2*m3^2 + mQ3^2 + mU3^2))/
        (3*(m3^2 - mQ3^2)*(mQ3^2 - mU3^2)^4) - 
       (64*(mQ3^2 + mU3^2)*(18*m3^5 + m3^3*mQ3^2 + 3*m3*mQ3^4 - 
          37*m3^3*mU3^2 - 7*m3*mQ3^2*mU3^2 + 22*m3*mU3^4))/
        (3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^4))*Log[mQ3^2/MR^2] - 
     (64*a3^2*at*k^(3/2)*mt^2*(-18*m3^5*mQ3^2 - m3^3*mQ3^4 - 3*m3*mQ3^6 + 
        14*m3^5*mU3^2 + 20*m3^3*mQ3^2*mU3^2 + 4*m3*mQ3^4*mU3^2 - 
        11*m3^3*mU3^4 + m3*mQ3^2*mU3^4 - 6*m3*mU3^6)*Log[mU3^2/MR^2])/
      (3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^4))*PolyLog[2, 1 - m3^2/mQ3^2] + 
   ((-1280*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*(m3^3*mQ3^2 - 2*m3^3*msq^2 + 
        m3*mQ3^2*msq^2 + m3^3*mU3^2 - 2*m3*mQ3^2*mU3^2 + m3*msq^2*mU3^2))/
      ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2) + 
     (640*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*(mQ3^2 + mU3^2)*
       (m3^3*mQ3^2 - 2*m3^3*msq^2 + m3*mQ3^2*msq^2 + m3^3*mU3^2 - 
        2*m3*mQ3^2*mU3^2 + m3*msq^2*mU3^2)*Log[mQ3^2/MR^2])/
      ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) - 
     (640*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*(mQ3^2 + mU3^2)*
       (m3^3*mQ3^2 - 2*m3^3*msq^2 + m3*mQ3^2*msq^2 + m3^3*mU3^2 - 
        2*m3*mQ3^2*mU3^2 + m3*msq^2*mU3^2)*Log[mU3^2/MR^2])/
      ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3))*
    PolyLog[2, 1 - m3^2/msq^2] + 
   ((-1280*a3^2*at*k^(3/2)*m3*(mQ3^2 - msq^2)^2*mt^2)/
      ((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^3) + 
     (640*a3^2*at*k^(3/2)*m3*(mQ3^2 - msq^2)^2*mt^2*(mQ3^2 + mU3^2)*
       Log[mQ3^2/MR^2])/((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^4) - 
     (640*a3^2*at*k^(3/2)*m3*(mQ3^2 - msq^2)^2*mt^2*(mQ3^2 + mU3^2)*
       Log[mU3^2/MR^2])/((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^4))*
    PolyLog[2, 1 - msq^2/mQ3^2] + 
   ((-128*a3^2*at*k^(3/2)*mt^2*(18*m3^5 - 37*m3^3*mQ3^2 + 22*m3*mQ3^4 + 
        m3^3*mU3^2 - 7*m3*mQ3^2*mU3^2 + 3*m3*mU3^4))/
      (3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^3) - 
     (1024*a3^2*at*k^(3/2)*m3^3*mt^2*(-2*m3^2 + mQ3^2 + mU3^2)*
       Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^3) - 
     (64*a3^2*at*k^(3/2)*mt^2*(14*m3^5*mQ3^2 - 11*m3^3*mQ3^4 - 6*m3*mQ3^6 - 
        18*m3^5*mU3^2 + 20*m3^3*mQ3^2*mU3^2 + m3*mQ3^4*mU3^2 - m3^3*mU3^4 + 
        4*m3*mQ3^2*mU3^4 - 3*m3*mU3^6)*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^4) + a3^2*at*k^(3/2)*mt^2*
      ((1024*m3*mU3^2*(2*m3^2 - mQ3^2 - mU3^2))/(3*(m3^2 - mU3^2)*
         (-mQ3^2 + mU3^2)^4) - (64*(mQ3^2 + mU3^2)*(18*m3^5 - 37*m3^3*mQ3^2 + 
          22*m3*mQ3^4 + m3^3*mU3^2 - 7*m3*mQ3^2*mU3^2 + 3*m3*mU3^4))/
        (3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^4))*Log[mU3^2/MR^2])*
    PolyLog[2, 1 - m3^2/mU3^2] + 
   ((128*a3^2*at*k^(3/2)*mt^2*(5*m3^5 - 5*m3^3*mQ3^2 + 3*m3*mQ3^4 - 
        5*m3^3*mU3^2 - m3*mQ3^2*mU3^2 + 3*m3*mU3^4))/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) - 
     (64*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*(5*m3^5 - 5*m3^3*mQ3^2 + 
        3*m3*mQ3^4 - 5*m3^3*mU3^2 - m3*mQ3^2*mU3^2 + 3*m3*mU3^4)*
       Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*
       (mQ3^2 - mU3^2)^2) + (64*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*
       (5*m3^5 - 5*m3^3*mQ3^2 + 3*m3*mQ3^4 - 5*m3^3*mU3^2 - m3*mQ3^2*mU3^2 + 
        3*m3*mU3^4)*Log[mU3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*
       (mQ3^2 - mU3^2)^2))*PolyLog[2, 1 - mQ3^2/mU3^2] + 
   ((-1280*a3^2*at*k^(3/2)*m3*mt^2*(-msq^2 + mU3^2)^2)/
      ((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^3) - 
     (640*a3^2*at*k^(3/2)*m3*mt^2*(mQ3^2 + mU3^2)*(-msq^2 + mU3^2)^2*
       Log[mQ3^2/MR^2])/((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^4) + 
     (640*a3^2*at*k^(3/2)*m3*mt^2*(mQ3^2 + mU3^2)*(-msq^2 + mU3^2)^2*
       Log[mU3^2/MR^2])/((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^4))*
    PolyLog[2, 1 - msq^2/mU3^2]) + 
 Xt^2*((-64*a3*at*k*m3^2*mt^2)/(mQ3^2*mU3^2) - 
   (64*a3^2*at*k^(3/2)*m3^2*mt^2*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 
      5*mQ3^2*mU3^2))/(3*mQ3^2*(m3^2 - mQ3^2)*mU3^2*(m3^2 - mU3^2)) + 
   (64*a3*at*k*m3^4*mt^2*(-m3^2 + mQ3^2 + mU3^2)*Log[m3^2/MR^2])/
    (mQ3^2*(-m3^2 + mQ3^2)*mU3^2*(m3^2 - mU3^2)) - 
   (64*a3^2*at*k^(3/2)*mt^2*(33*m3^10 - 42*m3^8*mQ3^2 + 17*m3^6*mQ3^4 - 
      42*m3^8*mU3^2 + 25*m3^6*mQ3^2*mU3^2 - 7*m3^4*mQ3^4*mU3^2 + 
      17*m3^6*mU3^4 - 7*m3^4*mQ3^2*mU3^4 + 6*m3^2*mQ3^4*mU3^4)*
     Log[m3^2/MR^2])/(3*mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*(m3^2 - mU3^2)^2) + 
   (64*a3*at*k*(3*m3^2 - 2*mQ3^2)*mt^2*Log[mQ3^2/MR^2])/
    ((m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) - 
   (8*a3^2*at*k^(3/2)*mt^2*(4*m3^10*mQ3^2 - 136*m3^8*mQ3^4 + 68*m3^6*mQ3^6 - 
      4*m3^10*mU3^2 + 695*m3^8*mQ3^2*mU3^2 - 326*m3^6*mQ3^4*mU3^2 + 
      5*m3^4*mQ3^6*mU3^2 - 18*m3^2*mQ3^8*mU3^2 - 360*m3^6*mQ3^2*msq^2*mU3^2 + 
      300*m3^4*mQ3^4*msq^2*mU3^2 - 180*m3^2*mQ3^6*msq^2*mU3^2 + 
      8*m3^8*mU3^4 - 886*m3^6*mQ3^2*mU3^4 + 40*m3^4*mQ3^4*mU3^4 + 
      308*m3^2*mQ3^6*mU3^4 - 6*mQ3^8*mU3^4 + 300*m3^4*mQ3^2*msq^2*mU3^4 + 
      240*m3^2*mQ3^4*msq^2*mU3^4 - 60*mQ3^6*msq^2*mU3^4 - 4*m3^6*mU3^6 + 
      301*m3^4*mQ3^2*mU3^6 + 212*m3^2*mQ3^4*mU3^6 - 237*mQ3^6*mU3^6 - 
      180*m3^2*mQ3^2*msq^2*mU3^6 - 60*mQ3^4*msq^2*mU3^6 - 
      18*m3^2*mQ3^2*mU3^8 - 6*mQ3^4*mU3^8)*Log[mQ3^2/MR^2])/
    (3*mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) + 
   ((12*at*Sqrt[k]*mt^2)/(mQ3^2 - mU3^2) - 
     (16*a3^2*at*k^(3/2)*mt^2*(30*m3^16 - 334*m3^14*mQ3^2 + 517*m3^12*mQ3^4 - 
        264*m3^10*mQ3^6 + 55*m3^8*mQ3^8 - 238*m3^14*mU3^2 + 
        1484*m3^12*mQ3^2*mU3^2 - 1948*m3^10*mQ3^4*mU3^2 + 
        844*m3^8*mQ3^6*mU3^2 - 158*m3^6*mQ3^8*mU3^2 + 325*m3^12*mU3^4 - 
        1660*m3^10*mQ3^2*mU3^4 + 1648*m3^8*mQ3^4*mU3^4 - 
        288*m3^6*mQ3^6*mU3^4 - m3^4*mQ3^8*mU3^4 - 168*m3^10*mU3^6 + 
        652*m3^8*mQ3^2*mU3^6 - 192*m3^6*mQ3^4*mU3^6 - 452*m3^4*mQ3^6*mU3^6 + 
        144*m3^2*mQ3^8*mU3^6 + 55*m3^8*mU3^8 - 158*m3^6*mQ3^2*mU3^8 - 
        m3^4*mQ3^4*mU3^8 + 144*m3^2*mQ3^6*mU3^8 - 36*mQ3^8*mU3^8)*
       Log[m3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^4*
       (mQ3^2 - mU3^2)))*Log[mQ3^2/MR^2] + 
   ((-32*a3*at*k*mt^2*(3*mQ3^2 - mU3^2))/(mQ3^2 - mU3^2)^2 + 
     (32*a3^2*at*k^(3/2)*mt^2*(64*m3^16 - 502*m3^14*mQ3^2 + 954*m3^12*mQ3^4 - 
        759*m3^10*mQ3^6 + 302*m3^8*mQ3^8 - 55*m3^6*mQ3^10 - 54*m3^14*mU3^2 + 
        962*m3^12*mQ3^2*mU3^2 - 1897*m3^10*mQ3^4*mU3^2 + 
        1401*m3^8*mQ3^6*mU3^2 - 525*m3^6*mQ3^8*mU3^2 + 97*m3^4*mQ3^10*mU3^2 - 
        72*m3^12*mU3^4 - 545*m3^10*mQ3^2*mU3^4 + 1199*m3^8*mQ3^4*mU3^4 - 
        811*m3^6*mQ3^6*mU3^4 + 325*m3^4*mQ3^8*mU3^4 - 72*m3^2*mQ3^10*mU3^4 + 
        69*m3^10*mU3^6 + 79*m3^8*mQ3^2*mU3^6 - 189*m3^6*mQ3^4*mU3^6 + 
        61*m3^4*mQ3^6*mU3^6 - 60*m3^2*mQ3^8*mU3^6 + 24*mQ3^10*mU3^6 - 
        m3^8*mU3^8 - 24*m3^6*mQ3^2*mU3^8 - 7*m3^4*mQ3^4*mU3^8 + 
        48*m3^2*mQ3^6*mU3^8 - 12*mQ3^8*mU3^8)*Log[m3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^2))*
    Log[mQ3^2/MR^2]^2 - (8*a3^2*at*k^(3/2)*mt^2*
     (128*m3^16 - 1062*m3^14*mQ3^2 + 1682*m3^12*mQ3^4 - 641*m3^10*mQ3^6 - 
      200*m3^8*mQ3^8 + 72*m3^6*mQ3^10 + 38*m3^4*mQ3^12 - 9*m3^2*mQ3^14 + 
      60*m3^12*mQ3^2*msq^2 + 120*m3^10*mQ3^4*msq^2 - 180*m3^8*mQ3^6*msq^2 - 
      90*m3^10*mQ3^2*msq^4 + 60*m3^8*mQ3^4*msq^4 + 30*m3^6*mQ3^6*msq^4 - 
      342*m3^14*mU3^2 + 3302*m3^12*mQ3^2*mU3^2 - 5333*m3^10*mQ3^4*mU3^2 + 
      1921*m3^8*mQ3^6*mU3^2 + 948*m3^6*mQ3^8*mU3^2 - 572*m3^4*mQ3^10*mU3^2 + 
      47*m3^2*mQ3^12*mU3^2 - 3*mQ3^14*mU3^2 - 60*m3^12*msq^2*mU3^2 - 
      300*m3^10*mQ3^2*msq^2*mU3^2 - 180*m3^8*mQ3^4*msq^2*mU3^2 + 
      540*m3^6*mQ3^6*msq^2*mU3^2 + 90*m3^10*msq^4*mU3^2 + 
      210*m3^8*mQ3^2*msq^4*mU3^2 - 210*m3^6*mQ3^4*msq^4*mU3^2 - 
      90*m3^4*mQ3^6*msq^4*mU3^2 + 268*m3^12*mU3^4 - 3637*m3^10*mQ3^2*mU3^4 + 
      6335*m3^8*mQ3^4*mU3^4 - 2862*m3^6*mQ3^6*mU3^4 - 434*m3^4*mQ3^8*mU3^4 + 
      379*m3^2*mQ3^10*mU3^4 - mQ3^12*mU3^4 + 180*m3^10*msq^2*mU3^4 + 
      540*m3^8*mQ3^2*msq^2*mU3^4 - 180*m3^6*mQ3^4*msq^2*mU3^4 - 
      540*m3^4*mQ3^6*msq^2*mU3^4 - 270*m3^8*msq^4*mU3^4 - 
      90*m3^6*mQ3^2*msq^4*mU3^4 + 270*m3^4*mQ3^4*msq^4*mU3^4 + 
      90*m3^2*mQ3^6*msq^4*mU3^4 - m3^10*mU3^6 + 1445*m3^8*mQ3^2*mU3^6 - 
      2814*m3^6*mQ3^4*mU3^6 + 1430*m3^4*mQ3^6*mU3^6 + 39*m3^2*mQ3^8*mU3^6 - 
      131*mQ3^10*mU3^6 - 180*m3^8*msq^2*mU3^6 - 420*m3^6*mQ3^2*msq^2*mU3^6 + 
      420*m3^4*mQ3^4*msq^2*mU3^6 + 180*m3^2*mQ3^6*msq^2*mU3^6 + 
      270*m3^6*msq^4*mU3^6 - 90*m3^4*mQ3^2*msq^4*mU3^6 - 
      150*m3^2*mQ3^4*msq^4*mU3^6 - 30*mQ3^6*msq^4*mU3^6 - 41*m3^8*mU3^8 - 
      132*m3^6*mQ3^2*mU3^8 + 382*m3^4*mQ3^4*mU3^8 - 268*m3^2*mQ3^6*mU3^8 + 
      67*mQ3^8*mU3^8 + 60*m3^6*msq^2*mU3^8 + 120*m3^4*mQ3^2*msq^2*mU3^8 - 
      180*m3^2*mQ3^4*msq^2*mU3^8 - 90*m3^4*msq^4*mU3^8 + 
      60*m3^2*mQ3^2*msq^4*mU3^8 + 30*mQ3^4*msq^4*mU3^8)*Log[mQ3^2/MR^2]^3)/
    (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^2) - 
   (320*a3^2*at*k^(3/2)*m3^2*mt^2*Log[msq^2/MR^2])/(3*mQ3^2*mU3^2) - 
   (8*a3^2*at*k^(3/2)*mt^2*(4*m3^10*mQ3^2 - 8*m3^8*mQ3^4 + 4*m3^6*mQ3^6 - 
      4*m3^10*mU3^2 - 695*m3^8*mQ3^2*mU3^2 + 886*m3^6*mQ3^4*mU3^2 - 
      301*m3^4*mQ3^6*mU3^2 + 18*m3^2*mQ3^8*mU3^2 + 360*m3^6*mQ3^2*msq^2*
       mU3^2 - 300*m3^4*mQ3^4*msq^2*mU3^2 + 180*m3^2*mQ3^6*msq^2*mU3^2 + 
      136*m3^8*mU3^4 + 326*m3^6*mQ3^2*mU3^4 - 40*m3^4*mQ3^4*mU3^4 - 
      212*m3^2*mQ3^6*mU3^4 + 6*mQ3^8*mU3^4 - 300*m3^4*mQ3^2*msq^2*mU3^4 - 
      240*m3^2*mQ3^4*msq^2*mU3^4 + 60*mQ3^6*msq^2*mU3^4 - 68*m3^6*mU3^6 - 
      5*m3^4*mQ3^2*mU3^6 - 308*m3^2*mQ3^4*mU3^6 + 237*mQ3^6*mU3^6 + 
      180*m3^2*mQ3^2*msq^2*mU3^6 + 60*mQ3^4*msq^2*mU3^6 + 
      18*m3^2*mQ3^2*mU3^8 + 6*mQ3^4*mU3^8)*Log[mU3^2/MR^2])/
    (3*mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) + 
   ((-12*at*Sqrt[k]*mt^2)/(mQ3^2 - mU3^2) + 
     (64*a3*at*k*mt^2*(3*m3^2 - 2*mU3^2))/((m3^2 - mU3^2)*(-mQ3^2 + mU3^2)) + 
     (16*a3^2*at*k^(3/2)*mt^2*(30*m3^16 - 238*m3^14*mQ3^2 + 325*m3^12*mQ3^4 - 
        168*m3^10*mQ3^6 + 55*m3^8*mQ3^8 - 334*m3^14*mU3^2 + 
        1484*m3^12*mQ3^2*mU3^2 - 1660*m3^10*mQ3^4*mU3^2 + 
        652*m3^8*mQ3^6*mU3^2 - 158*m3^6*mQ3^8*mU3^2 + 517*m3^12*mU3^4 - 
        1948*m3^10*mQ3^2*mU3^4 + 1648*m3^8*mQ3^4*mU3^4 - 
        192*m3^6*mQ3^6*mU3^4 - m3^4*mQ3^8*mU3^4 - 264*m3^10*mU3^6 + 
        844*m3^8*mQ3^2*mU3^6 - 288*m3^6*mQ3^4*mU3^6 - 452*m3^4*mQ3^6*mU3^6 + 
        144*m3^2*mQ3^8*mU3^6 + 55*m3^8*mU3^8 - 158*m3^6*mQ3^2*mU3^8 - 
        m3^4*mQ3^4*mU3^8 + 144*m3^2*mQ3^6*mU3^8 - 36*mQ3^8*mU3^8)*
       Log[m3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^4*
       (mQ3^2 - mU3^2)) + ((64*a3*at*k*mt^2*(mQ3^2 + mU3^2))/
        (mQ3^2 - mU3^2)^2 - (64*a3^2*at*k^(3/2)*mt^2*
         (64*m3^18 - 310*m3^16*mQ3^2 + 516*m3^14*mQ3^4 - 405*m3^12*mQ3^6 + 
          164*m3^10*mQ3^8 - 27*m3^8*mQ3^10 - 310*m3^16*mU3^2 + 
          1368*m3^14*mQ3^2*mU3^2 - 2083*m3^12*mQ3^4*mU3^2 + 
          1488*m3^10*mQ3^6*mU3^2 - 561*m3^8*mQ3^8*mU3^2 + 
          88*m3^6*mQ3^10*mU3^2 + 516*m3^14*mU3^4 - 2083*m3^12*mQ3^2*mU3^4 + 
          2808*m3^10*mQ3^4*mU3^4 - 1704*m3^8*mQ3^6*mU3^4 + 
          564*m3^6*mQ3^8*mU3^4 - 81*m3^4*mQ3^10*mU3^4 - 405*m3^12*mU3^6 + 
          1488*m3^10*mQ3^2*mU3^6 - 1704*m3^8*mQ3^4*mU3^6 + 
          776*m3^6*mQ3^6*mU3^6 - 199*m3^4*mQ3^8*mU3^6 + 24*m3^2*mQ3^10*
           mU3^6 + 164*m3^10*mU3^8 - 561*m3^8*mQ3^2*mU3^8 + 
          564*m3^6*mQ3^4*mU3^8 - 199*m3^4*mQ3^6*mU3^8 + 48*m3^2*mQ3^8*mU3^8 - 
          6*mQ3^10*mU3^8 - 27*m3^8*mU3^10 + 88*m3^6*mQ3^2*mU3^10 - 
          81*m3^4*mQ3^4*mU3^10 + 24*m3^2*mQ3^6*mU3^10 - 6*mQ3^8*mU3^10)*
         Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^4*
         (mQ3^2 - mU3^2)^2))*Log[mQ3^2/MR^2] + 
     (8*a3^2*at*k^(3/2)*mt^2*(128*m3^16 - 558*m3^14*mQ3^2 + 462*m3^12*mQ3^4 + 
        241*m3^10*mQ3^6 - 256*m3^8*mQ3^8 - 96*m3^6*mQ3^10 + 114*m3^4*mQ3^12 - 
        27*m3^2*mQ3^14 + 60*m3^12*mQ3^2*msq^2 + 120*m3^10*mQ3^4*msq^2 - 
        180*m3^8*mQ3^6*msq^2 - 90*m3^10*mQ3^2*msq^4 + 60*m3^8*mQ3^4*msq^4 + 
        30*m3^6*mQ3^6*msq^4 - 846*m3^14*mU3^2 + 3590*m3^12*mQ3^2*mU3^2 - 
        4419*m3^10*mQ3^4*mU3^2 + 1383*m3^8*mQ3^6*mU3^2 + 
        636*m3^6*mQ3^8*mU3^2 - 508*m3^4*mQ3^10*mU3^2 + 
        141*m3^2*mQ3^12*mU3^2 - 9*mQ3^14*mU3^2 - 60*m3^12*msq^2*mU3^2 - 
        300*m3^10*mQ3^2*msq^2*mU3^2 - 180*m3^8*mQ3^4*msq^2*mU3^2 + 
        540*m3^6*mQ3^6*msq^2*mU3^2 + 90*m3^10*msq^4*mU3^2 + 
        210*m3^8*mQ3^2*msq^4*mU3^2 - 210*m3^6*mQ3^4*msq^4*mU3^2 - 
        90*m3^4*mQ3^6*msq^4*mU3^2 + 1200*m3^12*mU3^4 - 
        5047*m3^10*mQ3^2*mU3^4 + 6841*m3^8*mQ3^4*mU3^4 - 
        3782*m3^6*mQ3^6*mU3^4 + 1010*m3^4*mQ3^8*mU3^4 - 
        171*m3^2*mQ3^10*mU3^4 - 3*mQ3^12*mU3^4 + 180*m3^10*msq^2*mU3^4 + 
        540*m3^8*mQ3^2*msq^2*mU3^4 - 180*m3^6*mQ3^4*msq^2*mU3^4 - 
        540*m3^4*mQ3^6*msq^2*mU3^4 - 270*m3^8*msq^4*mU3^4 - 
        90*m3^6*mQ3^2*msq^4*mU3^4 + 270*m3^4*mQ3^4*msq^4*mU3^4 + 
        90*m3^2*mQ3^6*msq^4*mU3^4 - 387*m3^10*mU3^6 + 1575*m3^8*mQ3^2*mU3^6 - 
        1854*m3^6*mQ3^4*mU3^6 + 742*m3^4*mQ3^6*mU3^6 - 151*m3^2*mQ3^8*mU3^6 + 
        43*mQ3^10*mU3^6 - 180*m3^8*msq^2*mU3^6 - 420*m3^6*mQ3^2*msq^2*mU3^6 + 
        420*m3^4*mQ3^4*msq^2*mU3^6 + 180*m3^2*mQ3^6*msq^2*mU3^6 + 
        270*m3^6*msq^4*mU3^6 - 90*m3^4*mQ3^2*msq^4*mU3^6 - 
        150*m3^2*mQ3^4*msq^4*mU3^6 - 30*mQ3^6*msq^4*mU3^6 - 83*m3^8*mU3^8 + 
        308*m3^6*mQ3^2*mU3^8 - 514*m3^4*mQ3^4*mU3^8 + 396*m3^2*mQ3^6*mU3^8 - 
        99*mQ3^8*mU3^8 + 60*m3^6*msq^2*mU3^8 + 120*m3^4*mQ3^2*msq^2*mU3^8 - 
        180*m3^2*mQ3^4*msq^2*mU3^8 - 90*m3^4*msq^4*mU3^8 + 
        60*m3^2*mQ3^2*msq^4*mU3^8 + 30*mQ3^4*msq^4*mU3^8)*Log[mQ3^2/MR^2]^2)/
      (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^2))*
    Log[mU3^2/MR^2] + ((-32*a3*at*k*mt^2*(-mQ3^2 + 3*mU3^2))/
      (-mQ3^2 + mU3^2)^2 + (32*a3^2*at*k^(3/2)*mt^2*
       (64*m3^16 - 54*m3^14*mQ3^2 - 72*m3^12*mQ3^4 + 69*m3^10*mQ3^6 - 
        m3^8*mQ3^8 - 502*m3^14*mU3^2 + 962*m3^12*mQ3^2*mU3^2 - 
        545*m3^10*mQ3^4*mU3^2 + 79*m3^8*mQ3^6*mU3^2 - 24*m3^6*mQ3^8*mU3^2 + 
        954*m3^12*mU3^4 - 1897*m3^10*mQ3^2*mU3^4 + 1199*m3^8*mQ3^4*mU3^4 - 
        189*m3^6*mQ3^6*mU3^4 - 7*m3^4*mQ3^8*mU3^4 - 759*m3^10*mU3^6 + 
        1401*m3^8*mQ3^2*mU3^6 - 811*m3^6*mQ3^4*mU3^6 + 61*m3^4*mQ3^6*mU3^6 + 
        48*m3^2*mQ3^8*mU3^6 + 302*m3^8*mU3^8 - 525*m3^6*mQ3^2*mU3^8 + 
        325*m3^4*mQ3^4*mU3^8 - 60*m3^2*mQ3^6*mU3^8 - 12*mQ3^8*mU3^8 - 
        55*m3^6*mU3^10 + 97*m3^4*mQ3^2*mU3^10 - 72*m3^2*mQ3^4*mU3^10 + 
        24*mQ3^6*mU3^10)*Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^4*
       (mQ3^2 - mU3^2)^2) + (8*a3^2*at*k^(3/2)*mt^2*
       (128*m3^16 - 846*m3^14*mQ3^2 + 1164*m3^12*mQ3^4 - 345*m3^10*mQ3^6 - 
        185*m3^8*mQ3^8 + 87*m3^6*mQ3^10 - 27*m3^4*mQ3^12 - 
        60*m3^12*mQ3^2*msq^2 + 180*m3^10*mQ3^4*msq^2 - 180*m3^8*mQ3^6*msq^2 + 
        60*m3^6*mQ3^8*msq^2 + 90*m3^10*mQ3^2*msq^4 - 270*m3^8*mQ3^4*msq^4 + 
        270*m3^6*mQ3^6*msq^4 - 90*m3^4*mQ3^8*msq^4 - 558*m3^14*mU3^2 + 
        3662*m3^12*mQ3^2*mU3^2 - 5053*m3^10*mQ3^4*mU3^2 + 
        1965*m3^8*mQ3^6*mU3^2 + 41*m3^6*mQ3^8*mU3^2 + 45*m3^4*mQ3^10*mU3^2 + 
        18*m3^2*mQ3^12*mU3^2 + 60*m3^12*msq^2*mU3^2 - 300*m3^10*mQ3^2*msq^2*
         mU3^2 + 540*m3^8*mQ3^4*msq^2*mU3^2 - 420*m3^6*mQ3^6*msq^2*mU3^2 + 
        120*m3^4*mQ3^8*msq^2*mU3^2 - 90*m3^10*msq^4*mU3^2 + 
        210*m3^8*mQ3^2*msq^4*mU3^2 - 90*m3^6*mQ3^4*msq^4*mU3^2 - 
        90*m3^4*mQ3^6*msq^4*mU3^2 + 60*m3^2*mQ3^8*msq^4*mU3^2 + 
        426*m3^12*mU3^4 - 4533*m3^10*mQ3^2*mU3^4 + 6223*m3^8*mQ3^4*mU3^4 - 
        2004*m3^6*mQ3^6*mU3^4 - 226*m3^4*mQ3^8*mU3^4 - 
        135*m3^2*mQ3^10*mU3^4 + 9*mQ3^12*mU3^4 + 120*m3^10*msq^2*mU3^4 - 
        180*m3^8*mQ3^2*msq^2*mU3^4 - 180*m3^6*mQ3^4*msq^2*mU3^4 + 
        420*m3^4*mQ3^6*msq^2*mU3^4 - 180*m3^2*mQ3^8*msq^2*mU3^4 + 
        60*m3^8*msq^4*mU3^4 - 210*m3^6*mQ3^2*msq^4*mU3^4 + 
        270*m3^4*mQ3^4*msq^4*mU3^4 - 150*m3^2*mQ3^6*msq^4*mU3^4 + 
        30*mQ3^8*msq^4*mU3^4 + 319*m3^10*mU3^6 + 1857*m3^8*mQ3^2*mU3^6 - 
        2840*m3^6*mQ3^4*mU3^6 + 412*m3^4*mQ3^6*mU3^6 + 489*m3^2*mQ3^8*mU3^6 + 
        3*mQ3^10*mU3^6 - 180*m3^8*msq^2*mU3^6 + 540*m3^6*mQ3^2*msq^2*mU3^6 - 
        540*m3^4*mQ3^4*msq^2*mU3^6 + 180*m3^2*mQ3^6*msq^2*mU3^6 + 
        30*m3^6*msq^4*mU3^6 - 90*m3^4*mQ3^2*msq^4*mU3^6 + 
        90*m3^2*mQ3^4*msq^4*mU3^6 - 30*mQ3^6*msq^4*mU3^6 - 400*m3^8*mU3^8 - 
        165*m3^6*mQ3^2*mU3^8 + 647*m3^4*mQ3^4*mU3^8 - 91*m3^2*mQ3^6*mU3^8 - 
        111*mQ3^8*mU3^8 + 93*m3^6*mU3^10 - 7*m3^4*mQ3^2*mU3^10 - 
        93*m3^2*mQ3^4*mU3^10 + 31*mQ3^6*mU3^10)*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^4*(mQ3^2 - mU3^2)^2))*
    Log[mU3^2/MR^2]^2 - (8*a3^2*at*k^(3/2)*mt^2*(128*m3^10 + 42*m3^8*mQ3^2 - 
      2*m3^6*mQ3^4 + 9*m3^4*mQ3^6 - 60*m3^6*mQ3^2*msq^2 + 
      90*m3^4*mQ3^2*msq^4 - 1062*m3^8*mU3^2 + 140*m3^6*mQ3^2*mU3^2 - 
      33*m3^4*mQ3^4*mU3^2 - 6*m3^2*mQ3^6*mU3^2 + 60*m3^6*msq^2*mU3^2 - 
      120*m3^4*mQ3^2*msq^2*mU3^2 - 90*m3^4*msq^4*mU3^2 - 
      60*m3^2*mQ3^2*msq^4*mU3^2 + 1670*m3^6*mU3^4 - 361*m3^4*mQ3^2*mU3^4 + 
      36*m3^2*mQ3^4*mU3^4 - 3*mQ3^6*mU3^4 + 120*m3^4*msq^2*mU3^4 + 
      180*m3^2*mQ3^2*msq^2*mU3^4 + 60*m3^2*msq^4*mU3^4 - 
      30*mQ3^2*msq^4*mU3^4 - 615*m3^4*mU3^6 + 234*m3^2*mQ3^2*mU3^6 - 
      mQ3^4*mU3^6 - 180*m3^2*msq^2*mU3^6 + 30*msq^4*mU3^6 - 248*m3^2*mU3^8 - 
      63*mQ3^2*mU3^8 + 135*mU3^10)*Log[mU3^2/MR^2]^3)/
    (3*(m3^2 - mU3^2)^4*(-mQ3^2 + mU3^2)^2) + 
   Log[msq^2/MR^2]^2*((-48*a3^2*at*k^(3/2)*mt^2*
       ((-5*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 
           3*m3^2*msq^2 + mQ3^2*msq^2))/(6*(-m3^2 + mQ3^2)^3) - 
        (5*(-2*msq^2 + 2*mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
           msq^2*mU3^2 + mU3^4))/(6*(-m3^2 + mU3^2)^3))*Log[mQ3^2/MR^2])/
      (mQ3^2 - mU3^2) + (48*a3^2*at*k^(3/2)*mt^2*
       ((-5*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 
           3*m3^2*msq^2 + mQ3^2*msq^2))/(6*(-m3^2 + mQ3^2)^3) - 
        (5*(-2*msq^2 + 2*mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
           msq^2*mU3^2 + mU3^4))/(6*(-m3^2 + mU3^2)^3))*Log[mU3^2/MR^2])/
      (mQ3^2 - mU3^2)) + Log[msq^2/MR^2]*
    ((320*a3^2*at*k^(3/2)*m3^4*mt^2*(-m3^2 + mQ3^2 + mU3^2)*Log[m3^2/MR^2])/
      (3*mQ3^2*(-m3^2 + mQ3^2)*mU3^2*(m3^2 - mU3^2)) + 
     ((160*a3^2*at*k^(3/2)*mt^2*(18*m3^8 - 33*m3^6*mQ3^2 + 15*m3^4*mQ3^4 + 
          18*m3^6*msq^2 - 15*m3^4*mQ3^2*msq^2 + 9*m3^2*mQ3^4*msq^2 - 
          35*m3^6*mU3^2 + 64*m3^4*mQ3^2*mU3^2 - 29*m3^2*mQ3^4*mU3^2 - 
          15*m3^4*msq^2*mU3^2 - 12*m3^2*mQ3^2*msq^2*mU3^2 + 
          3*mQ3^4*msq^2*mU3^2 + 17*m3^4*mU3^4 - 31*m3^2*mQ3^2*mU3^4 + 
          14*mQ3^4*mU3^4 + 9*m3^2*msq^2*mU3^4 + 3*mQ3^2*msq^2*mU3^4))/
        (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) + 
       (320*a3^2*at*k^(3/2)*mt^2*(2*m3^12 - 10*m3^10*mQ3^2 + 3*m3^8*mQ3^4 - 
          m3^6*mQ3^6 - 10*m3^10*mU3^2 + 42*m3^8*mQ3^2*mU3^2 - 
          21*m3^6*mQ3^4*mU3^2 + 7*m3^4*mQ3^6*mU3^2 + 3*m3^8*mU3^4 - 
          21*m3^6*mQ3^2*mU3^4 - m3^6*mU3^6 + 7*m3^4*mQ3^2*mU3^6)*
         Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
         (mQ3^2 - mU3^2)))*Log[mQ3^2/MR^2] + a3^2*at*k^(3/2)*mt^2*
      ((160*(2*m3^4*msq^2 + 6*m3^2*mQ3^2*msq^2 - 3*m3^2*msq^4 - mQ3^2*msq^4))/
        ((m3^2 - mQ3^2)^3*(mQ3^2 - mU3^2)) - 
       (160*(4*mQ3^6 - 2*mQ3^4*mU3^2 + m3^4*(3*mQ3^2 - mU3^2) + 
          m3^2*(-8*mQ3^4 + 4*mQ3^2*mU3^2)))/(3*(m3^2 - mQ3^2)^2*
         (mQ3^2 - mU3^2)^2))*Log[mQ3^2/MR^2]^2 + 
     ((-160*a3^2*at*k^(3/2)*mt^2*(18*m3^8 - 35*m3^6*mQ3^2 + 17*m3^4*mQ3^4 + 
          18*m3^6*msq^2 - 15*m3^4*mQ3^2*msq^2 + 9*m3^2*mQ3^4*msq^2 - 
          33*m3^6*mU3^2 + 64*m3^4*mQ3^2*mU3^2 - 31*m3^2*mQ3^4*mU3^2 - 
          15*m3^4*msq^2*mU3^2 - 12*m3^2*mQ3^2*msq^2*mU3^2 + 
          3*mQ3^4*msq^2*mU3^2 + 15*m3^4*mU3^4 - 29*m3^2*mQ3^2*mU3^4 + 
          14*mQ3^4*mU3^4 + 9*m3^2*msq^2*mU3^4 + 3*mQ3^2*msq^2*mU3^4))/
        (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) - 
       (320*a3^2*at*k^(3/2)*mt^2*(2*m3^12 - 10*m3^10*mQ3^2 + 3*m3^8*mQ3^4 - 
          m3^6*mQ3^6 - 10*m3^10*mU3^2 + 42*m3^8*mQ3^2*mU3^2 - 
          21*m3^6*mQ3^4*mU3^2 + 7*m3^4*mQ3^6*mU3^2 + 3*m3^8*mU3^4 - 
          21*m3^6*mQ3^2*mU3^4 - m3^6*mU3^6 + 7*m3^4*mQ3^2*mU3^6)*
         Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
         (mQ3^2 - mU3^2)) + a3^2*at*k^(3/2)*mt^2*
        ((-160*(2*m3^4*msq^2 + 6*m3^2*mQ3^2*msq^2 - 3*m3^2*msq^4 - 
            mQ3^2*msq^4))/((m3^2 - mQ3^2)^3*(mQ3^2 - mU3^2)) - 
         (160*(2*m3^4*msq^2 - 3*m3^2*msq^4 + 6*m3^2*msq^2*mU3^2 - 
            msq^4*mU3^2))/((m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)) + 
         (160*(2*m3^8*(mQ3^2 + mU3^2) + 2*mQ3^4*mU3^4*(mQ3^2 + mU3^2) - 
            4*m3^2*mQ3^2*mU3^2*(mQ3^2 + mU3^2)^2 + 3*m3^4*(mQ3^2 + mU3^2)^3 - 
            2*m3^6*(3*mQ3^4 + 2*mQ3^2*mU3^2 + 3*mU3^4)))/(3*(m3^2 - mQ3^2)^2*
           (m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2))*Log[mQ3^2/MR^2])*
      Log[mU3^2/MR^2] + a3^2*at*k^(3/2)*mt^2*
      ((160*(2*m3^4*msq^2 - 3*m3^2*msq^4 + 6*m3^2*msq^2*mU3^2 - msq^4*mU3^2))/
        ((m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)) + 
       (160*(m3^4*(mQ3^2 - 3*mU3^2) + 2*mU3^4*(mQ3^2 - 2*mU3^2) + 
          m3^2*(-4*mQ3^2*mU3^2 + 8*mU3^4)))/(3*(m3^2 - mU3^2)^2*
         (mQ3^2 - mU3^2)^2))*Log[mU3^2/MR^2]^2) + 
   a3^2*at*k^(3/2)*mt^2*Log[m3^2/MR^2]^2*
    ((-1792*m3^6)/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
     (64*m3^4*(-m3^2 + mQ3^2 + mU3^2)*
       (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
           m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)))/
      (mQ3^2*(-m3^2 + mQ3^2)*mU3^2*(m3^2 - mU3^2)) + 
     (512*m3^6*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)) + a3^2*at*k^(3/2)*mt^2*
    Log[mQ3^2/MR^2]^2*((-1792*m3^2*mQ3^4)/(3*(m3^2 - mQ3^2)^2*
       (mQ3^2 - mU3^2)^2) - (4096*m3^2*mQ3^2)/(3*(m3^2 - mQ3^2)*
       (mQ3^2 - mU3^2)^2) + (16*(-309*m3^10*mQ3^2 + 419*m3^8*mQ3^4 - 
        21*m3^6*mQ3^6 - 100*m3^4*mQ3^8 - 9*m3^2*mQ3^10 + 
        90*m3^6*mQ3^4*msq^2 + 30*m3^4*mQ3^6*msq^2 + 53*m3^10*mU3^2 + 
        711*m3^8*mQ3^2*mU3^2 - 1011*m3^6*mQ3^4*mU3^2 + 34*m3^4*mQ3^6*mU3^2 + 
        288*m3^2*mQ3^8*mU3^2 - 3*mQ3^10*mU3^2 - 90*m3^6*mQ3^2*msq^2*mU3^2 - 
        210*m3^4*mQ3^4*msq^2*mU3^2 - 60*m3^2*mQ3^6*msq^2*mU3^2 - 
        106*m3^8*mU3^4 - 557*m3^6*mQ3^2*mU3^4 + 938*m3^4*mQ3^4*mU3^4 - 
        231*m3^2*mQ3^6*mU3^4 - 140*mQ3^8*mU3^4 + 180*m3^4*mQ3^2*msq^2*mU3^4 + 
        150*m3^2*mQ3^4*msq^2*mU3^4 + 30*mQ3^6*msq^2*mU3^4 + 53*m3^6*mU3^6 + 
        152*m3^4*mQ3^2*mU3^6 - 295*m3^2*mQ3^4*mU3^6 + 146*mQ3^6*mU3^6 - 
        90*m3^2*mQ3^2*msq^2*mU3^6 - 30*mQ3^4*msq^2*mU3^6 - 
        9*m3^2*mQ3^2*mU3^8 - 3*mQ3^4*mU3^8))/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2) + 
     (16*(m3^4 - 34*m3^2*mQ3^2 + 17*mQ3^4)*(7*m3^4 + 3*mQ3^2*mU3^2 - 
        2*m3^2*(2*mQ3^2 + 3*mU3^2)))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)*
       (mQ3^2 - mU3^2)) - (32*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 
        5*mQ3^2*mU3^2)*(4*mQ3^6 - 2*mQ3^4*mU3^2 + m3^4*(3*mQ3^2 - mU3^2) + 
        m3^2*(-8*mQ3^4 + 4*mQ3^2*mU3^2)))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)*
       (mQ3^2 - mU3^2)^2) + (512*m3^2*mQ3^4*(Log[MR^2] - 2*Log[mt] - 
        2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/((m3^2 - mQ3^2)^2*
       (mQ3^2 - mU3^2)^2)) + a3^2*at*k^(3/2)*mt^2*Log[m3^2/MR^2]*
    Log[mQ3^2/MR^2]*((32*m3^4*(m3^4 - 34*m3^2*mQ3^2 + 17*mQ3^4)*
       (m3^2 - mQ3^2 - mU3^2))/(3*mQ3^2*(m3^2 - mQ3^2)^3*mU3^2*
       (m3^2 - mU3^2)) + (3584*m3^4*mQ3^2)/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*
       (mQ3^2 - mU3^2)) + (4096*m3^4)/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*
       (mQ3^2 - mU3^2)) + (16*(800*m3^12 - 1854*m3^10*mQ3^2 + 
        1410*m3^8*mQ3^4 - 345*m3^6*mQ3^6 + 9*m3^4*mQ3^8 - 180*m3^10*msq^2 + 
        240*m3^8*mQ3^2*msq^2 - 270*m3^6*mQ3^4*msq^2 + 90*m3^4*mQ3^6*msq^2 - 
        1854*m3^10*mU3^2 + 4056*m3^8*mQ3^2*mU3^2 - 2875*m3^6*mQ3^4*mU3^2 + 
        598*m3^4*mQ3^6*mU3^2 + 3*m3^2*mQ3^8*mU3^2 + 240*m3^8*msq^2*mU3^2 + 
        180*m3^6*mQ3^2*msq^2*mU3^2 - 90*m3^4*mQ3^4*msq^2*mU3^2 + 
        30*m3^2*mQ3^6*msq^2*mU3^2 + 1410*m3^8*mU3^4 - 2875*m3^6*mQ3^2*mU3^4 + 
        1930*m3^4*mQ3^4*mU3^4 - 369*m3^2*mQ3^6*mU3^4 - 270*m3^6*msq^2*mU3^4 - 
        90*m3^4*mQ3^2*msq^2*mU3^4 - 345*m3^6*mU3^6 + 598*m3^4*mQ3^2*mU3^6 - 
        369*m3^2*mQ3^4*mU3^6 + 60*mQ3^6*mU3^6 + 90*m3^4*msq^2*mU3^6 + 
        30*m3^2*mQ3^2*msq^2*mU3^6 + 9*m3^4*mU3^8 + 3*m3^2*mQ3^2*mU3^8))/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) - 
     (32*m3^4*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2)*
       (2*m3^4 + mQ3^4 + mU3^4 - 2*m3^2*(mQ3^2 + mU3^2)))/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) + 
     (32*(7*m3^4 + 3*mQ3^2*mU3^2 - 2*m3^2*(2*mQ3^2 + 3*mU3^2))*
       (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
           m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)))/
      ((m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) - 
     (1024*m3^4*mQ3^2*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
        2*Log[Sin[beta]]))/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*
       (mQ3^2 - mU3^2))) - (16*a3^2*at*k^(3/2)*mt^2*Log[mU3^2/MR^2]^2*
     (-24*m3^10*mQ3^2 + 65*m3^8*mQ3^4 - 41*m3^6*mQ3^6 + 492*m3^10*mU3^2 - 
      1462*m3^8*mQ3^2*mU3^2 + 1346*m3^6*mQ3^4*mU3^2 - 373*m3^4*mQ3^6*mU3^2 + 
      9*m3^2*mQ3^8*mU3^2 + 90*m3^6*mQ3^2*msq^2*mU3^2 - 
      180*m3^4*mQ3^4*msq^2*mU3^2 + 90*m3^2*mQ3^6*msq^2*mU3^2 - 
      383*m3^8*mU3^4 + 1473*m3^6*mQ3^2*mU3^4 - 1612*m3^4*mQ3^4*mU3^4 + 
      471*m3^2*mQ3^6*mU3^4 + 3*mQ3^8*mU3^4 - 90*m3^6*msq^2*mU3^4 + 
      210*m3^4*mQ3^2*msq^2*mU3^4 - 150*m3^2*mQ3^4*msq^2*mU3^4 + 
      30*mQ3^6*msq^2*mU3^4 - 226*m3^6*mU3^6 + 137*m3^4*mQ3^2*mU3^6 + 
      338*m3^2*mQ3^4*mU3^6 - 177*mQ3^6*mU3^6 - 30*m3^4*msq^2*mU3^6 + 
      60*m3^2*mQ3^2*msq^2*mU3^6 - 30*mQ3^4*msq^2*mU3^6 + 192*m3^4*mU3^8 - 
      391*m3^2*mQ3^2*mU3^8 + 151*mQ3^4*mU3^8 + 9*m3^2*mU3^10 + 
      3*mQ3^2*mU3^10 - 96*m3^8*mU3^4*Log[MR^2] + 192*m3^6*mQ3^2*mU3^4*
       Log[MR^2] - 96*m3^4*mQ3^4*mU3^4*Log[MR^2] + 96*m3^6*mU3^6*Log[MR^2] - 
      192*m3^4*mQ3^2*mU3^6*Log[MR^2] + 96*m3^2*mQ3^4*mU3^6*Log[MR^2] + 
      192*m3^8*mU3^4*Log[mt] - 384*m3^6*mQ3^2*mU3^4*Log[mt] + 
      192*m3^4*mQ3^4*mU3^4*Log[mt] - 192*m3^6*mU3^6*Log[mt] + 
      384*m3^4*mQ3^2*mU3^6*Log[mt] - 192*m3^2*mQ3^4*mU3^6*Log[mt] + 
      192*m3^8*mU3^4*Log[Csc[beta]] - 384*m3^6*mQ3^2*mU3^4*Log[Csc[beta]] + 
      192*m3^4*mQ3^4*mU3^4*Log[Csc[beta]] - 192*m3^6*mU3^6*Log[Csc[beta]] + 
      384*m3^4*mQ3^2*mU3^6*Log[Csc[beta]] - 192*m3^2*mQ3^4*mU3^6*
       Log[Csc[beta]] + 192*m3^8*mU3^4*Log[Sin[beta]] - 
      384*m3^6*mQ3^2*mU3^4*Log[Sin[beta]] + 192*m3^4*mQ3^4*mU3^4*
       Log[Sin[beta]] - 192*m3^6*mU3^6*Log[Sin[beta]] + 
      384*m3^4*mQ3^2*mU3^6*Log[Sin[beta]] - 192*m3^2*mQ3^4*mU3^6*
       Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*
     (-mQ3^2 + mU3^2)^2) + Log[mU3^2/MR^2]*
    (a3^2*at*k^(3/2)*mt^2*Log[m3^2/MR^2]*
      ((-3584*m3^4*mU3^2)/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*
         (mQ3^2 - mU3^2)) - (4096*m3^4)/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*
         (mQ3^2 - mU3^2)) + (32*m3^4*(-m3^2 + mQ3^2 + mU3^2)*
         (m3^4 - 34*m3^2*mU3^2 + 17*mU3^4))/(3*mQ3^2*(-m3^2 + mQ3^2)*mU3^2*
         (m3^2 - mU3^2)^3) - (16*(800*m3^12 - 1854*m3^10*mQ3^2 + 
          1410*m3^8*mQ3^4 - 345*m3^6*mQ3^6 + 9*m3^4*mQ3^8 - 180*m3^10*msq^2 + 
          240*m3^8*mQ3^2*msq^2 - 270*m3^6*mQ3^4*msq^2 + 90*m3^4*mQ3^6*msq^2 - 
          1854*m3^10*mU3^2 + 4056*m3^8*mQ3^2*mU3^2 - 2875*m3^6*mQ3^4*mU3^2 + 
          598*m3^4*mQ3^6*mU3^2 + 3*m3^2*mQ3^8*mU3^2 + 240*m3^8*msq^2*mU3^2 + 
          180*m3^6*mQ3^2*msq^2*mU3^2 - 90*m3^4*mQ3^4*msq^2*mU3^2 + 
          30*m3^2*mQ3^6*msq^2*mU3^2 + 1410*m3^8*mU3^4 - 2875*m3^6*mQ3^2*
           mU3^4 + 1930*m3^4*mQ3^4*mU3^4 - 369*m3^2*mQ3^6*mU3^4 - 
          270*m3^6*msq^2*mU3^4 - 90*m3^4*mQ3^2*msq^2*mU3^4 - 345*m3^6*mU3^6 + 
          598*m3^4*mQ3^2*mU3^6 - 369*m3^2*mQ3^4*mU3^6 + 60*mQ3^6*mU3^6 + 
          90*m3^4*msq^2*mU3^6 + 30*m3^2*mQ3^2*msq^2*mU3^6 + 9*m3^4*mU3^8 + 
          3*m3^2*mQ3^2*mU3^8))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
         (mQ3^2 - mU3^2)) + (32*m3^4*(-11*m3^4 + 3*m3^2*mQ3^2 + 
          3*m3^2*mU3^2 + 5*mQ3^2*mU3^2)*(2*m3^4 + mQ3^4 + mU3^4 - 
          2*m3^2*(mQ3^2 + mU3^2)))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
         (mQ3^2 - mU3^2)) - (32*(7*m3^4 + 3*mQ3^2*mU3^2 - 
          2*m3^2*(3*mQ3^2 + 2*mU3^2))*
         (2 + (8*(2*m3^8 - 2*m3^6*mQ3^2 + m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
             m3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)))/
        ((m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) + 
       (1024*m3^4*mU3^2*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
          2*Log[Sin[beta]]))/((m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*
         (mQ3^2 - mU3^2))) + (16*a3^2*at*k^(3/2)*mt^2*Log[mQ3^2/MR^2]*
       (468*m3^12*mQ3^2 - 402*m3^10*mQ3^4 - 292*m3^8*mQ3^6 + 269*m3^6*mQ3^8 + 
        9*m3^4*mQ3^10 - 90*m3^8*mQ3^4*msq^2 - 30*m3^6*mQ3^6*msq^2 + 
        468*m3^12*mU3^2 - 3692*m3^10*mQ3^2*mU3^2 + 4624*m3^8*mQ3^4*mU3^2 - 
        1268*m3^6*mQ3^6*mU3^2 - 337*m3^4*mQ3^8*mU3^2 - 15*m3^2*mQ3^10*mU3^2 + 
        180*m3^8*mQ3^2*msq^2*mU3^2 + 30*m3^6*mQ3^4*msq^2*mU3^2 + 
        360*m3^4*mQ3^6*msq^2*mU3^2 - 90*m3^2*mQ3^8*msq^2*mU3^2 - 
        402*m3^10*mU3^4 + 4624*m3^8*mQ3^2*mU3^4 - 6418*m3^6*mQ3^4*mU3^4 + 
        2420*m3^4*mQ3^6*mU3^4 + 142*m3^2*mQ3^8*mU3^4 - 6*mQ3^10*mU3^4 - 
        90*m3^8*msq^2*mU3^4 + 30*m3^6*mQ3^2*msq^2*mU3^4 - 
        720*m3^4*mQ3^4*msq^2*mU3^4 + 90*m3^2*mQ3^6*msq^2*mU3^4 - 
        30*mQ3^8*msq^2*mU3^4 - 292*m3^8*mU3^6 - 1268*m3^6*mQ3^2*mU3^6 + 
        2420*m3^4*mQ3^4*mU3^6 - 1166*m3^2*mQ3^6*mU3^6 + 26*mQ3^8*mU3^6 - 
        30*m3^6*msq^2*mU3^6 + 360*m3^4*mQ3^2*msq^2*mU3^6 + 
        90*m3^2*mQ3^4*msq^2*mU3^6 + 60*mQ3^6*msq^2*mU3^6 + 269*m3^6*mU3^8 - 
        337*m3^4*mQ3^2*mU3^8 + 142*m3^2*mQ3^4*mU3^8 + 26*mQ3^6*mU3^8 - 
        90*m3^2*mQ3^2*msq^2*mU3^8 - 30*mQ3^4*msq^2*mU3^8 + 9*m3^4*mU3^10 - 
        15*m3^2*mQ3^2*mU3^10 - 6*mQ3^4*mU3^10 - 192*m3^10*mQ3^2*mU3^2*
         Log[MR^2] + 384*m3^8*mQ3^4*mU3^2*Log[MR^2] - 192*m3^6*mQ3^6*mU3^2*
         Log[MR^2] + 384*m3^8*mQ3^2*mU3^4*Log[MR^2] - 768*m3^6*mQ3^4*mU3^4*
         Log[MR^2] + 384*m3^4*mQ3^6*mU3^4*Log[MR^2] - 192*m3^6*mQ3^2*mU3^6*
         Log[MR^2] + 384*m3^4*mQ3^4*mU3^6*Log[MR^2] - 192*m3^2*mQ3^6*mU3^6*
         Log[MR^2] + 384*m3^10*mQ3^2*mU3^2*Log[mt] - 768*m3^8*mQ3^4*mU3^2*
         Log[mt] + 384*m3^6*mQ3^6*mU3^2*Log[mt] - 768*m3^8*mQ3^2*mU3^4*
         Log[mt] + 1536*m3^6*mQ3^4*mU3^4*Log[mt] - 768*m3^4*mQ3^6*mU3^4*
         Log[mt] + 384*m3^6*mQ3^2*mU3^6*Log[mt] - 768*m3^4*mQ3^4*mU3^6*
         Log[mt] + 384*m3^2*mQ3^6*mU3^6*Log[mt] + 384*m3^10*mQ3^2*mU3^2*
         Log[Csc[beta]] - 768*m3^8*mQ3^4*mU3^2*Log[Csc[beta]] + 
        384*m3^6*mQ3^6*mU3^2*Log[Csc[beta]] - 768*m3^8*mQ3^2*mU3^4*
         Log[Csc[beta]] + 1536*m3^6*mQ3^4*mU3^4*Log[Csc[beta]] - 
        768*m3^4*mQ3^6*mU3^4*Log[Csc[beta]] + 384*m3^6*mQ3^2*mU3^6*
         Log[Csc[beta]] - 768*m3^4*mQ3^4*mU3^6*Log[Csc[beta]] + 
        384*m3^2*mQ3^6*mU3^6*Log[Csc[beta]] + 384*m3^10*mQ3^2*mU3^2*
         Log[Sin[beta]] - 768*m3^8*mQ3^4*mU3^2*Log[Sin[beta]] + 
        384*m3^6*mQ3^6*mU3^2*Log[Sin[beta]] - 768*m3^8*mQ3^2*mU3^4*
         Log[Sin[beta]] + 1536*m3^6*mQ3^4*mU3^4*Log[Sin[beta]] - 
        768*m3^4*mQ3^6*mU3^4*Log[Sin[beta]] + 384*m3^6*mQ3^2*mU3^6*
         Log[Sin[beta]] - 768*m3^4*mQ3^4*mU3^6*Log[Sin[beta]] + 
        384*m3^2*mQ3^6*mU3^6*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^2)) + 
   ((-8192*a3^2*at*k^(3/2)*m3^6*mt^2*Log[m3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) - 
     (16*a3^2*at*k^(3/2)*mt^2*(128*m3^12 - 806*m3^10*mQ3^2 + 98*m3^8*mQ3^4 + 
        9*m3^6*mQ3^6 + 20*m3^4*mQ3^8 - 9*m3^2*mQ3^10 - 346*m3^10*mU3^2 + 
        2438*m3^8*mQ3^2*mU3^2 - 251*m3^6*mQ3^4*mU3^2 - 151*m3^4*mQ3^6*mU3^2 + 
        41*m3^2*mQ3^8*mU3^2 - 3*mQ3^10*mU3^2 + 280*m3^8*mU3^4 - 
        2561*m3^6*mQ3^2*mU3^4 + 417*m3^4*mQ3^4*mU3^4 + 41*m3^2*mQ3^6*mU3^4 - 
        mQ3^8*mU3^4 - 13*m3^6*mU3^6 + 903*m3^4*mQ3^2*mU3^6 - 
        167*m3^2*mQ3^4*mU3^6 - 19*mQ3^6*mU3^6 - 37*m3^4*mU3^8 - 
        34*m3^2*mQ3^2*mU3^8 + 23*mQ3^4*mU3^8)*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^2) + 
     (16*a3^2*at*k^(3/2)*mt^2*(128*m3^12 - 294*m3^10*mQ3^2 + 98*m3^8*mQ3^4 + 
        9*m3^6*mQ3^6 + 20*m3^4*mQ3^8 - 9*m3^2*mQ3^10 - 858*m3^10*mU3^2 + 
        1414*m3^8*mQ3^2*mU3^2 - 251*m3^6*mQ3^4*mU3^2 - 151*m3^4*mQ3^6*mU3^2 + 
        41*m3^2*mQ3^8*mU3^2 - 3*mQ3^10*mU3^2 + 1304*m3^8*mU3^4 - 
        2049*m3^6*mQ3^2*mU3^4 + 417*m3^4*mQ3^4*mU3^4 + 41*m3^2*mQ3^6*mU3^4 - 
        mQ3^8*mU3^4 - 525*m3^6*mU3^6 + 903*m3^4*mQ3^2*mU3^6 - 
        167*m3^2*mQ3^4*mU3^6 - 19*mQ3^6*mU3^6 - 37*m3^4*mU3^8 - 
        34*m3^2*mQ3^2*mU3^8 + 23*mQ3^4*mU3^8)*Log[mU3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^2))*
    PolyLog[2, 1 - m3^2/mQ3^2] - (160*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*
     (2*m3^10 - 8*m3^8*mQ3^2 + 3*m3^6*mQ3^4 - m3^4*mQ3^6 + 6*m3^8*msq^2 - 
      8*m3^6*mQ3^2*msq^2 + 9*m3^4*mQ3^4*msq^2 - 3*m3^2*mQ3^6*msq^2 - 
      8*m3^8*mU3^2 + 30*m3^6*mQ3^2*mU3^2 - 15*m3^4*mQ3^4*mU3^2 + 
      5*m3^2*mQ3^6*mU3^2 - 8*m3^6*msq^2*mU3^2 - 6*m3^4*mQ3^2*msq^2*mU3^2 + 
      3*m3^2*mQ3^4*msq^2*mU3^2 - mQ3^6*msq^2*mU3^2 + 3*m3^6*mU3^4 - 
      15*m3^4*mQ3^2*mU3^4 + 9*m3^4*msq^2*mU3^4 + 3*m3^2*mQ3^2*msq^2*mU3^4 - 
      m3^4*mU3^6 + 5*m3^2*mQ3^2*mU3^6 - 3*m3^2*msq^2*mU3^6 - 
      mQ3^2*msq^2*mU3^6)*(Log[mQ3^2/MR^2] - Log[mU3^2/MR^2])*
     PolyLog[2, 1 - m3^2/msq^2])/((m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
     (mQ3^2 - mU3^2)) + 
   ((80*a3^2*at*k^(3/2)*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 
        3*m3^2*msq^2 + mQ3^2*msq^2)*mt^2*Log[mQ3^2/MR^2])/
      ((-m3^2 + mQ3^2)^3*(mQ3^2 - mU3^2)) - 
     (80*a3^2*at*k^(3/2)*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + 
        mQ3^4 + 3*m3^2*msq^2 + mQ3^2*msq^2)*mt^2*Log[mU3^2/MR^2])/
      ((-m3^2 + mQ3^2)^3*(mQ3^2 - mU3^2)))*PolyLog[2, 1 - msq^2/mQ3^2] + 
   ((-8192*a3^2*at*k^(3/2)*m3^6*mt^2*Log[m3^2/MR^2])/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)) + 
     (16*a3^2*at*k^(3/2)*mt^2*(128*m3^12 - 858*m3^10*mQ3^2 + 
        1304*m3^8*mQ3^4 - 525*m3^6*mQ3^6 - 37*m3^4*mQ3^8 - 294*m3^10*mU3^2 + 
        1414*m3^8*mQ3^2*mU3^2 - 2049*m3^6*mQ3^4*mU3^2 + 
        903*m3^4*mQ3^6*mU3^2 - 34*m3^2*mQ3^8*mU3^2 + 98*m3^8*mU3^4 - 
        251*m3^6*mQ3^2*mU3^4 + 417*m3^4*mQ3^4*mU3^4 - 167*m3^2*mQ3^6*mU3^4 + 
        23*mQ3^8*mU3^4 + 9*m3^6*mU3^6 - 151*m3^4*mQ3^2*mU3^6 + 
        41*m3^2*mQ3^4*mU3^6 - 19*mQ3^6*mU3^6 + 20*m3^4*mU3^8 + 
        41*m3^2*mQ3^2*mU3^8 - mQ3^4*mU3^8 - 9*m3^2*mU3^10 - 3*mQ3^2*mU3^10)*
       Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*
       (mQ3^2 - mU3^2)^2) - (16*a3^2*at*k^(3/2)*mt^2*
       (128*m3^12 - 346*m3^10*mQ3^2 + 280*m3^8*mQ3^4 - 13*m3^6*mQ3^6 - 
        37*m3^4*mQ3^8 - 806*m3^10*mU3^2 + 2438*m3^8*mQ3^2*mU3^2 - 
        2561*m3^6*mQ3^4*mU3^2 + 903*m3^4*mQ3^6*mU3^2 - 34*m3^2*mQ3^8*mU3^2 + 
        98*m3^8*mU3^4 - 251*m3^6*mQ3^2*mU3^4 + 417*m3^4*mQ3^4*mU3^4 - 
        167*m3^2*mQ3^6*mU3^4 + 23*mQ3^8*mU3^4 + 9*m3^6*mU3^6 - 
        151*m3^4*mQ3^2*mU3^6 + 41*m3^2*mQ3^4*mU3^6 - 19*mQ3^6*mU3^6 + 
        20*m3^4*mU3^8 + 41*m3^2*mQ3^2*mU3^8 - mQ3^4*mU3^8 - 9*m3^2*mU3^10 - 
        3*mQ3^2*mU3^10)*Log[mU3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*
       (-mQ3^2 + mU3^2)^2))*PolyLog[2, 1 - m3^2/mU3^2] + 
   ((-16*a3^2*at*k^(3/2)*mt^2*(12*m3^10 - 14*m3^8*mQ3^2 + 34*m3^6*mQ3^4 - 
        29*m3^4*mQ3^6 + 9*m3^2*mQ3^8 - 14*m3^8*mU3^2 - 76*m3^6*mQ3^2*mU3^2 + 
        65*m3^4*mQ3^4*mU3^2 - 26*m3^2*mQ3^6*mU3^2 + 3*mQ3^8*mU3^2 + 
        34*m3^6*mU3^4 + 65*m3^4*mQ3^2*mU3^4 - 34*m3^2*mQ3^4*mU3^4 + 
        7*mQ3^6*mU3^4 - 29*m3^4*mU3^6 - 26*m3^2*mQ3^2*mU3^6 + 7*mQ3^4*mU3^6 + 
        9*m3^2*mU3^8 + 3*mQ3^2*mU3^8)*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) + 
     (16*a3^2*at*k^(3/2)*mt^2*(12*m3^10 - 14*m3^8*mQ3^2 + 34*m3^6*mQ3^4 - 
        29*m3^4*mQ3^6 + 9*m3^2*mQ3^8 - 14*m3^8*mU3^2 - 76*m3^6*mQ3^2*mU3^2 + 
        65*m3^4*mQ3^4*mU3^2 - 26*m3^2*mQ3^6*mU3^2 + 3*mQ3^8*mU3^2 + 
        34*m3^6*mU3^4 + 65*m3^4*mQ3^2*mU3^4 - 34*m3^2*mQ3^4*mU3^4 + 
        7*mQ3^6*mU3^4 - 29*m3^4*mU3^6 - 26*m3^2*mQ3^2*mU3^6 + 7*mQ3^4*mU3^6 + 
        9*m3^2*mU3^8 + 3*mQ3^2*mU3^8)*Log[mU3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3))*PolyLog[2, 1 - mQ3^2/mU3^2] + 
   ((160*a3^2*at*k^(3/2)*mt^2*(-msq^2 + mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 
        3*m3^2*mU3^2 + msq^2*mU3^2 + mU3^4)*Log[mQ3^2/MR^2])/
      ((mQ3^2 - mU3^2)*(-m3^2 + mU3^2)^3) - 
     (160*a3^2*at*k^(3/2)*mt^2*(-msq^2 + mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 
        3*m3^2*mU3^2 + msq^2*mU3^2 + mU3^4)*Log[mU3^2/MR^2])/
      ((mQ3^2 - mU3^2)*(-m3^2 + mU3^2)^3))*PolyLog[2, 1 - msq^2/mU3^2]) + 
 Xt^4*((12*at*Sqrt[k]*mt^2)/(mQ3^2 - mU3^2)^2 + 
   (32*a3*at*k*mt^2*(m3^2*mQ3^2 + m3^2*mU3^2 + 6*mQ3^2*mU3^2))/
    (mQ3^2*mU3^2*(mQ3^2 - mU3^2)^2) + 
   (8*a3^2*at*k^(3/2)*mt^2*(-44*m3^10*mQ3^2 + 56*m3^8*mQ3^4 - 12*m3^6*mQ3^6 - 
      44*m3^10*mU3^2 - 455*m3^8*mQ3^2*mU3^2 + 558*m3^6*mQ3^4*mU3^2 - 
      173*m3^4*mQ3^6*mU3^2 + 18*m3^2*mQ3^8*mU3^2 + 360*m3^6*mQ3^2*msq^2*
       mU3^2 - 300*m3^4*mQ3^4*msq^2*mU3^2 + 180*m3^2*mQ3^6*msq^2*mU3^2 + 
      56*m3^8*mU3^4 + 558*m3^6*mQ3^2*mU3^4 - 168*m3^4*mQ3^4*mU3^4 - 
      236*m3^2*mQ3^6*mU3^4 + 6*mQ3^8*mU3^4 - 300*m3^4*mQ3^2*msq^2*mU3^4 - 
      240*m3^2*mQ3^4*msq^2*mU3^4 + 60*mQ3^6*msq^2*mU3^4 - 12*m3^6*mU3^6 - 
      173*m3^4*mQ3^2*mU3^6 - 236*m3^2*mQ3^4*mU3^6 + 277*mQ3^6*mU3^6 + 
      180*m3^2*mQ3^2*msq^2*mU3^6 + 60*mQ3^4*msq^2*mU3^6 + 
      18*m3^2*mQ3^2*mU3^8 + 6*mQ3^4*mU3^8))/(3*mQ3^2*(-m3^2 + mQ3^2)^2*mU3^2*
     (m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2) - 
   (32*a3*at*k*mt^2*(m3^2*mQ3^2 + m3^2*mU3^2)*Log[m3^2/MR^2])/
    (mQ3^2*mU3^2*(mQ3^2 - mU3^2)^2) + 
   (16*a3^2*at*k^(3/2)*mt^2*(66*m3^14*mQ3^2 - 150*m3^12*mQ3^4 + 
      118*m3^10*mQ3^6 - 34*m3^8*mQ3^8 + 66*m3^14*mU3^2 + 
      1876*m3^12*mQ3^2*mU3^2 - 4190*m3^10*mQ3^4*mU3^2 + 
      2756*m3^8*mQ3^6*mU3^2 - 465*m3^6*mQ3^8*mU3^2 + 9*m3^4*mQ3^10*mU3^2 - 
      180*m3^10*mQ3^2*msq^2*mU3^2 + 240*m3^8*mQ3^4*msq^2*mU3^2 - 
      270*m3^6*mQ3^6*msq^2*mU3^2 + 90*m3^4*mQ3^8*msq^2*mU3^2 - 
      150*m3^12*mU3^4 - 4190*m3^10*mQ3^2*mU3^4 + 8964*m3^8*mQ3^4*mU3^4 - 
      5645*m3^6*mQ3^6*mU3^4 + 850*m3^4*mQ3^8*mU3^4 + 3*m3^2*mQ3^10*mU3^4 + 
      240*m3^8*mQ3^2*msq^2*mU3^4 + 180*m3^6*mQ3^4*msq^2*mU3^4 - 
      90*m3^4*mQ3^6*msq^2*mU3^4 + 30*m3^2*mQ3^8*msq^2*mU3^4 + 
      118*m3^10*mU3^6 + 2756*m3^8*mQ3^2*mU3^6 - 5645*m3^6*mQ3^4*mU3^6 + 
      3522*m3^4*mQ3^6*mU3^6 - 559*m3^2*mQ3^8*mU3^6 - 
      270*m3^6*mQ3^2*msq^2*mU3^6 - 90*m3^4*mQ3^4*msq^2*mU3^6 - 
      34*m3^8*mU3^8 - 465*m3^6*mQ3^2*mU3^8 + 850*m3^4*mQ3^4*mU3^8 - 
      559*m3^2*mQ3^6*mU3^8 + 120*mQ3^8*mU3^8 + 90*m3^4*mQ3^2*msq^2*mU3^8 + 
      30*m3^2*mQ3^4*msq^2*mU3^8 + 9*m3^4*mQ3^2*mU3^10 + 3*m3^2*mQ3^4*mU3^10)*
     Log[m3^2/MR^2])/(3*mQ3^2*(m3^2 - mQ3^2)^3*mU3^2*(m3^2 - mU3^2)^3*
     (mQ3^2 - mU3^2)^2) - (16*a3^2*at*k^(3/2)*mt^2*
     (44*m3^18*mQ3^2 - 144*m3^16*mQ3^4 + 184*m3^14*mQ3^6 - 112*m3^12*mQ3^8 + 
      28*m3^10*mQ3^10 + 44*m3^18*mU3^2 - 2*m3^16*mQ3^2*mU3^2 - 
      294*m3^14*mQ3^4*mU3^2 + 277*m3^12*mQ3^6*mU3^2 + 4*m3^10*mQ3^8*mU3^2 - 
      25*m3^8*mQ3^10*mU3^2 - 144*m3^16*mU3^4 - 294*m3^14*mQ3^2*mU3^4 + 
      1932*m3^12*mQ3^4*mU3^4 - 2012*m3^10*mQ3^6*mU3^4 + 
      572*m3^8*mQ3^8*mU3^4 - 70*m3^6*mQ3^10*mU3^4 + 184*m3^14*mU3^6 + 
      277*m3^12*mQ3^2*mU3^6 - 2012*m3^10*mQ3^4*mU3^6 + 
      1712*m3^8*mQ3^6*mU3^6 - 88*m3^6*mQ3^8*mU3^6 - 49*m3^4*mQ3^10*mU3^6 - 
      112*m3^12*mU3^8 + 4*m3^10*mQ3^2*mU3^8 + 572*m3^8*mQ3^4*mU3^8 - 
      88*m3^6*mQ3^6*mU3^8 - 548*m3^4*mQ3^8*mU3^8 + 156*m3^2*mQ3^10*mU3^8 + 
      28*m3^10*mU3^10 - 25*m3^8*mQ3^2*mU3^10 - 70*m3^6*mQ3^4*mU3^10 - 
      49*m3^4*mQ3^6*mU3^10 + 156*m3^2*mQ3^8*mU3^10 - 36*mQ3^10*mU3^10)*
     Log[m3^2/MR^2]^2)/(3*mQ3^2*(m3^2 - mQ3^2)^4*mU3^2*(m3^2 - mU3^2)^4*
     (mQ3^2 - mU3^2)^2) - (32*a3*at*k*mt^2*(2*m3^2 + 7*mQ3^2 + 3*mU3^2)*
     Log[mQ3^2/MR^2])/(mQ3^2 - mU3^2)^3 + 
   (4*a3^2*at*k^(3/2)*mt^2*(4*m3^12*mQ3^4 - 140*m3^10*mQ3^6 + 
      204*m3^8*mQ3^8 - 68*m3^6*mQ3^10 + 176*m3^12*mQ3^2*mU3^2 - 
      4793*m3^10*mQ3^4*mU3^2 + 7283*m3^8*mQ3^6*mU3^2 - 
      1881*m3^6*mQ3^8*mU3^2 - 975*m3^4*mQ3^10*mU3^2 - 18*m3^2*mQ3^12*mU3^2 - 
      360*m3^8*mQ3^4*msq^2*mU3^2 + 1020*m3^6*mQ3^6*msq^2*mU3^2 - 
      360*m3^4*mQ3^8*msq^2*mU3^2 + 180*m3^2*mQ3^10*msq^2*mU3^2 - 
      4*m3^12*mU3^4 + 667*m3^10*mQ3^2*mU3^4 + 10101*m3^8*mQ3^4*mU3^4 - 
      17667*m3^6*mQ3^6*mU3^4 + 6073*m3^4*mQ3^8*mU3^4 + 
      1508*m3^2*mQ3^10*mU3^4 - 6*mQ3^12*mU3^4 - 360*m3^8*mQ3^2*msq^2*mU3^4 + 
      600*m3^6*mQ3^4*msq^2*mU3^4 - 1380*m3^4*mQ3^6*msq^2*mU3^4 - 
      360*m3^2*mQ3^8*msq^2*mU3^4 + 60*mQ3^10*msq^2*mU3^4 + 8*m3^10*mU3^6 - 
      1222*m3^8*mQ3^2*mU3^6 - 6913*m3^6*mQ3^4*mU3^6 + 
      13435*m3^4*mQ3^6*mU3^6 - 5659*m3^2*mQ3^8*mU3^6 - 417*mQ3^10*mU3^6 + 
      300*m3^6*mQ3^2*msq^2*mU3^6 + 480*m3^4*mQ3^4*msq^2*mU3^6 + 
      420*m3^2*mQ3^6*msq^2*mU3^6 + 240*mQ3^8*msq^2*mU3^6 - 4*m3^8*mU3^8 + 
      493*m3^6*mQ3^2*mU3^8 + 1361*m3^4*mQ3^4*mU3^8 - 2665*m3^2*mQ3^6*mU3^8 + 
      1167*mQ3^8*mU3^8 - 180*m3^4*mQ3^2*msq^2*mU3^8 - 
      240*m3^2*mQ3^4*msq^2*mU3^8 - 60*mQ3^6*msq^2*mU3^8 - 
      18*m3^4*mQ3^2*mU3^10 - 24*m3^2*mQ3^4*mU3^10 - 6*mQ3^6*mU3^10)*
     Log[mQ3^2/MR^2])/(3*mQ3^2*(m3^2 - mQ3^2)^3*mU3^2*(m3^2 - mU3^2)^2*
     (mQ3^2 - mU3^2)^3) - (8*a3^2*at*k^(3/2)*mt^2*
     (2*m3^16*mQ3^4 - 72*m3^14*mQ3^6 + 172*m3^12*mQ3^8 - 136*m3^10*mQ3^10 + 
      34*m3^8*mQ3^12 - 80*m3^16*mQ3^2*mU3^2 + 3178*m3^14*mQ3^4*mU3^2 - 
      9110*m3^12*mQ3^6*mU3^2 + 9926*m3^10*mQ3^8*mU3^2 - 
      4623*m3^8*mQ3^10*mU3^2 + 702*m3^6*mQ3^12*mU3^2 - 9*m3^4*mQ3^14*mU3^2 - 
      180*m3^12*mQ3^4*msq^2*mU3^2 + 420*m3^10*mQ3^6*msq^2*mU3^2 - 
      510*m3^8*mQ3^8*msq^2*mU3^2 + 360*m3^6*mQ3^10*msq^2*mU3^2 - 
      90*m3^4*mQ3^12*msq^2*mU3^2 - 2*m3^16*mU3^4 + 2152*m3^14*mQ3^2*mU3^4 - 
      12962*m3^12*mQ3^4*mU3^4 + 25584*m3^10*mQ3^6*mU3^4 - 
      22990*m3^8*mQ3^8*mU3^4 + 9553*m3^6*mQ3^10*mU3^4 - 
      1268*m3^4*mQ3^12*mU3^4 - 3*m3^2*mQ3^14*mU3^4 - 
      180*m3^12*mQ3^2*msq^2*mU3^4 + 660*m3^10*mQ3^4*msq^2*mU3^4 - 
      570*m3^8*mQ3^6*msq^2*mU3^4 + 90*m3^6*mQ3^8*msq^2*mU3^4 + 
      30*m3^4*mQ3^10*msq^2*mU3^4 - 30*m3^2*mQ3^12*msq^2*mU3^4 + 
      6*m3^14*mU3^6 - 4406*m3^12*mQ3^2*mU3^6 + 17588*m3^10*mQ3^4*mU3^6 - 
      25638*m3^8*mQ3^6*mU3^6 + 18070*m3^6*mQ3^8*mU3^6 - 
      6600*m3^4*mQ3^10*mU3^6 + 884*m3^2*mQ3^12*mU3^6 + 
      240*m3^10*mQ3^2*msq^2*mU3^6 - 330*m3^8*mQ3^4*msq^2*mU3^6 - 
      90*m3^6*mQ3^6*msq^2*mU3^6 + 210*m3^4*mQ3^8*msq^2*mU3^6 - 
      30*m3^2*mQ3^10*msq^2*mU3^6 - 6*m3^12*mU3^8 + 2996*m3^10*mQ3^2*mU3^8 - 
      9676*m3^8*mQ3^4*mU3^8 + 10862*m3^6*mQ3^6*mU3^8 - 
      5474*m3^4*mQ3^8*mU3^8 + 1602*m3^2*mQ3^10*mU3^8 - 240*mQ3^12*mU3^8 - 
      270*m3^8*mQ3^2*msq^2*mU3^8 + 270*m3^6*mQ3^4*msq^2*mU3^8 + 
      30*m3^4*mQ3^6*msq^2*mU3^8 - 30*m3^2*mQ3^8*msq^2*mU3^8 + 
      2*m3^10*mU3^10 - 643*m3^8*mQ3^2*mU3^10 + 1828*m3^6*mQ3^4*mU3^10 - 
      1723*m3^4*mQ3^6*mU3^10 + 616*m3^2*mQ3^8*mU3^10 - 96*mQ3^10*mU3^10 + 
      90*m3^6*mQ3^2*msq^2*mU3^10 - 60*m3^4*mQ3^4*msq^2*mU3^10 - 
      30*m3^2*mQ3^6*msq^2*mU3^10 + 9*m3^6*mQ3^2*mU3^12 - 
      6*m3^4*mQ3^4*mU3^12 - 3*m3^2*mQ3^6*mU3^12)*Log[m3^2/MR^2]*
     Log[mQ3^2/MR^2])/(3*mQ3^2*(m3^2 - mQ3^2)^4*mU3^2*(m3^2 - mU3^2)^3*
     (mQ3^2 - mU3^2)^3) + 
   (at*Sqrt[k]*mt^2*((-6*mQ3^2)/(mQ3^2 - mU3^2)^3 - 
       (6*mU3^2)/(mQ3^2 - mU3^2)^3) + (8*a3^2*at*k^(3/2)*mt^2*
       (1024*m3^18 - 2786*m3^16*mQ3^2 + 1826*m3^14*mQ3^4 + 549*m3^12*mQ3^6 - 
        664*m3^10*mQ3^8 + 55*m3^8*mQ3^10 - 2786*m3^16*mU3^2 + 
        8004*m3^14*mQ3^2*mU3^2 - 5903*m3^12*mQ3^4*mU3^2 - 
        868*m3^10*mQ3^6*mU3^2 + 1699*m3^8*mQ3^8*mU3^2 - 
        158*m3^6*mQ3^10*mU3^2 + 1826*m3^14*mU3^4 - 5903*m3^12*mQ3^2*mU3^4 + 
        5224*m3^10*mQ3^4*mU3^4 - 292*m3^8*mQ3^6*mU3^4 - 
        846*m3^6*mQ3^8*mU3^4 - m3^4*mQ3^10*mU3^4 + 549*m3^12*mU3^6 - 
        868*m3^10*mQ3^2*mU3^6 - 292*m3^8*mQ3^4*mU3^6 + 928*m3^6*mQ3^6*mU3^6 - 
        453*m3^4*mQ3^8*mU3^6 + 144*m3^2*mQ3^10*mU3^6 - 664*m3^10*mU3^8 + 
        1699*m3^8*mQ3^2*mU3^8 - 846*m3^6*mQ3^4*mU3^8 - 453*m3^4*mQ3^6*mU3^8 + 
        288*m3^2*mQ3^8*mU3^8 - 36*mQ3^10*mU3^8 + 55*m3^8*mU3^10 - 
        158*m3^6*mQ3^2*mU3^10 - m3^4*mQ3^4*mU3^10 + 144*m3^2*mQ3^6*mU3^10 - 
        36*mQ3^8*mU3^10)*Log[m3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^4*
       (m3^2 - mU3^2)^4*(mQ3^2 - mU3^2)^3))*Log[mQ3^2/MR^2] - 
   (8*a3^2*at*k^(3/2)*mt^2*(176*m3^16*mQ3^2 - 2586*m3^14*mQ3^4 + 
      5691*m3^12*mQ3^6 - 4056*m3^10*mQ3^8 + 416*m3^8*mQ3^10 + 
      329*m3^6*mQ3^12 + 47*m3^4*mQ3^14 - 9*m3^2*mQ3^16 + 
      60*m3^12*mQ3^4*msq^2 + 210*m3^10*mQ3^6*msq^2 - 240*m3^8*mQ3^8*msq^2 - 
      30*m3^6*mQ3^10*msq^2 - 90*m3^10*mQ3^4*msq^4 + 60*m3^8*mQ3^6*msq^4 + 
      30*m3^6*mQ3^8*msq^4 - 176*m3^16*mU3^2 - 296*m3^14*mQ3^2*mU3^2 + 
      8817*m3^12*mQ3^4*mU3^2 - 19237*m3^10*mQ3^6*mU3^2 + 
      13306*m3^8*mQ3^8*mU3^2 - 1117*m3^6*mQ3^10*mU3^2 - 
      1384*m3^4*mQ3^12*mU3^2 + 50*m3^2*mQ3^14*mU3^2 - 3*mQ3^16*mU3^2 - 
      120*m3^12*mQ3^2*msq^2*mU3^2 - 420*m3^10*mQ3^4*msq^2*mU3^2 - 
      270*m3^8*mQ3^6*msq^2*mU3^2 + 720*m3^6*mQ3^8*msq^2*mU3^2 + 
      90*m3^4*mQ3^10*msq^2*mU3^2 + 180*m3^10*mQ3^2*msq^4*mU3^2 + 
      150*m3^8*mQ3^4*msq^4*mU3^2 - 240*m3^6*mQ3^6*msq^4*mU3^2 - 
      90*m3^4*mQ3^8*msq^4*mU3^2 + 498*m3^14*mU3^4 + 357*m3^12*mQ3^2*mU3^4 - 
      12407*m3^10*mQ3^4*mU3^4 + 25376*m3^8*mQ3^6*mU3^4 - 
      17186*m3^6*mQ3^8*mU3^4 + 2382*m3^4*mQ3^10*mU3^4 + 
      1061*m3^2*mQ3^12*mU3^4 - mQ3^14*mU3^4 + 60*m3^12*msq^2*mU3^4 + 
      390*m3^10*mQ3^2*msq^2*mU3^4 + 780*m3^8*mQ3^4*msq^2*mU3^4 - 
      420*m3^6*mQ3^6*msq^2*mU3^4 - 720*m3^4*mQ3^8*msq^2*mU3^4 - 
      90*m3^2*mQ3^10*msq^2*mU3^4 - 90*m3^10*msq^4*mU3^4 - 
      480*m3^8*mQ3^2*msq^4*mU3^4 + 120*m3^6*mQ3^4*msq^4*mU3^4 + 
      360*m3^4*mQ3^6*msq^4*mU3^4 + 90*m3^2*mQ3^8*msq^4*mU3^4 - 
      481*m3^12*mU3^6 - 691*m3^10*mQ3^2*mU3^6 + 8970*m3^8*mQ3^4*mU3^6 - 
      15256*m3^6*mQ3^6*mU3^6 + 8973*m3^4*mQ3^8*mU3^6 - 
      1253*m3^2*mQ3^10*mU3^6 - 342*mQ3^12*mU3^6 - 180*m3^10*msq^2*mU3^6 - 
      450*m3^8*mQ3^2*msq^2*mU3^6 - 420*m3^6*mQ3^4*msq^2*mU3^6 + 
      780*m3^4*mQ3^6*msq^2*mU3^6 + 240*m3^2*mQ3^8*msq^2*mU3^6 + 
      30*mQ3^10*msq^2*mU3^6 + 270*m3^8*msq^4*mU3^6 + 
      360*m3^6*mQ3^2*msq^4*mU3^6 - 360*m3^4*mQ3^4*msq^4*mU3^6 - 
      240*m3^2*mQ3^6*msq^4*mU3^6 - 30*mQ3^8*msq^4*mU3^6 + 151*m3^10*mU3^8 + 
      816*m3^8*mQ3^2*mU3^8 - 3855*m3^6*mQ3^4*mU3^8 + 4746*m3^4*mQ3^6*mU3^8 - 
      2062*m3^2*mQ3^8*mU3^8 + 244*mQ3^10*mU3^8 + 180*m3^8*msq^2*mU3^8 + 
      210*m3^6*mQ3^2*msq^2*mU3^8 - 120*m3^4*mQ3^4*msq^2*mU3^8 - 
      270*m3^2*mQ3^6*msq^2*mU3^8 - 270*m3^6*msq^4*mU3^8 + 
      210*m3^2*mQ3^4*msq^4*mU3^8 + 60*mQ3^6*msq^4*mU3^8 - 4*m3^8*mU3^10 - 
      275*m3^6*mQ3^2*mU3^10 + 731*m3^4*mQ3^4*mU3^10 - 645*m3^2*mQ3^6*mU3^10 + 
      185*mQ3^8*mU3^10 - 60*m3^6*msq^2*mU3^10 - 30*m3^4*mQ3^2*msq^2*mU3^10 + 
      120*m3^2*mQ3^4*msq^2*mU3^10 - 30*mQ3^6*msq^2*mU3^10 + 
      90*m3^4*msq^4*mU3^10 - 60*m3^2*mQ3^2*msq^4*mU3^10 - 
      30*mQ3^4*msq^4*mU3^10 + 9*m3^4*mQ3^2*mU3^12 - 6*m3^2*mQ3^4*mU3^12 - 
      3*mQ3^6*mU3^12)*Log[mQ3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*
     (mQ3^2 - mU3^2)^4) + 
   ((32*a3*at*k*mt^2*(m3^2*mQ3^2 + 2*mQ3^4 - m3^2*mU3^2 + 2*mQ3^2*mU3^2))/
      (mQ3^2 - mU3^2)^4 - (16*a3^2*at*k^(3/2)*mt^2*(788*m3^16*mQ3^2 - 
        2556*m3^14*mQ3^4 + 2618*m3^12*mQ3^6 - 835*m3^10*mQ3^8 + 
        58*m3^8*mQ3^10 - 69*m3^6*mQ3^12 - 148*m3^16*mU3^2 - 
        1796*m3^14*mQ3^2*mU3^2 + 7366*m3^12*mQ3^4*mU3^2 - 
        8472*m3^10*mQ3^6*mU3^2 + 3031*m3^8*mQ3^8*mU3^2 - 
        116*m3^6*mQ3^10*mU3^2 + 123*m3^4*mQ3^12*mU3^2 + 336*m3^14*mU3^4 + 
        786*m3^12*mQ3^2*mU3^4 - 6232*m3^10*mQ3^4*mU3^4 + 
        8712*m3^8*mQ3^6*mU3^4 - 3742*m3^6*mQ3^8*mU3^4 + 
        238*m3^4*mQ3^10*mU3^4 - 90*m3^2*mQ3^12*mU3^4 - 194*m3^12*mU3^6 + 
        452*m3^10*mQ3^2*mU3^6 + 936*m3^8*mQ3^4*mU3^6 - 
        2648*m3^6*mQ3^6*mU3^6 + 1612*m3^4*mQ3^8*mU3^6 - 
        180*m3^2*mQ3^10*mU3^6 + 30*mQ3^12*mU3^6 - m3^10*mU3^8 - 
        190*m3^8*mQ3^2*mU3^8 + 479*m3^6*mQ3^4*mU3^8 - 234*m3^4*mQ3^6*mU3^8 - 
        90*m3^2*mQ3^8*mU3^8 + 24*mQ3^10*mU3^8 + 13*m3^8*mU3^10 - 
        64*m3^6*mQ3^2*mU3^10 + 37*m3^4*mQ3^4*mU3^10 + 24*m3^2*mQ3^6*mU3^10 - 
        6*mQ3^8*mU3^10)*Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*
       (mQ3^2 - mU3^2)^4))*Log[mQ3^2/MR^2]^2 + 
   (4*a3^2*at*k^(3/2)*mt^2*(1156*m3^16*mQ3^2 - 4276*m3^14*mQ3^4 + 
      4738*m3^12*mQ3^6 - 1189*m3^10*mQ3^8 - 556*m3^8*mQ3^10 + 
      106*m3^6*mQ3^12 + 38*m3^4*mQ3^14 - 9*m3^2*mQ3^16 + 
      60*m3^12*mQ3^4*msq^2 + 120*m3^10*mQ3^6*msq^2 - 180*m3^8*mQ3^8*msq^2 - 
      90*m3^10*mQ3^4*msq^4 + 60*m3^8*mQ3^6*msq^4 + 30*m3^6*mQ3^8*msq^4 + 
      124*m3^16*mU3^2 - 4340*m3^14*mQ3^2*mU3^2 + 15130*m3^12*mQ3^4*mU3^2 - 
      16510*m3^10*mQ3^6*mU3^2 + 4017*m3^8*mQ3^8*mU3^2 + 
      2156*m3^6*mQ3^10*mU3^2 - 636*m3^4*mQ3^12*mU3^2 + 38*m3^2*mQ3^14*mU3^2 - 
      3*mQ3^16*mU3^2 - 180*m3^10*mQ3^4*msq^2*mU3^2 - 
      360*m3^8*mQ3^6*msq^2*mU3^2 + 540*m3^6*mQ3^8*msq^2*mU3^2 + 
      270*m3^8*mQ3^4*msq^4*mU3^2 - 180*m3^6*mQ3^6*msq^4*mU3^2 - 
      90*m3^4*mQ3^8*msq^4*mU3^2 - 328*m3^14*mU3^4 + 6138*m3^12*mQ3^2*mU3^4 - 
      19952*m3^10*mQ3^4*mU3^4 + 21392*m3^8*mQ3^6*mU3^4 - 
      5480*m3^6*mQ3^8*mU3^4 - 2278*m3^4*mQ3^10*mU3^4 + 
      528*m3^2*mQ3^12*mU3^4 - 4*mQ3^14*mU3^4 - 60*m3^12*msq^2*mU3^4 - 
      120*m3^10*mQ3^2*msq^2*mU3^4 + 360*m3^8*mQ3^4*msq^2*mU3^4 + 
      360*m3^6*mQ3^6*msq^2*mU3^4 - 540*m3^4*mQ3^8*msq^2*mU3^4 + 
      90*m3^10*msq^4*mU3^4 - 60*m3^8*mQ3^2*msq^4*mU3^4 - 
      300*m3^6*mQ3^4*msq^4*mU3^4 + 180*m3^4*mQ3^6*msq^4*mU3^4 + 
      90*m3^2*mQ3^8*msq^4*mU3^4 + 250*m3^12*mU3^6 - 4006*m3^10*mQ3^2*mU3^6 + 
      11990*m3^8*mQ3^4*mU3^6 - 12428*m3^6*mQ3^6*mU3^6 + 
      3398*m3^4*mQ3^8*mU3^6 + 978*m3^2*mQ3^10*mU3^6 - 166*mQ3^12*mU3^6 + 
      180*m3^10*msq^2*mU3^6 + 360*m3^8*mQ3^2*msq^2*mU3^6 - 
      600*m3^6*mQ3^4*msq^2*mU3^6 - 120*m3^4*mQ3^6*msq^2*mU3^6 + 
      180*m3^2*mQ3^8*msq^2*mU3^6 - 270*m3^8*msq^4*mU3^6 + 
      180*m3^6*mQ3^2*msq^4*mU3^6 + 180*m3^4*mQ3^4*msq^4*mU3^6 - 
      60*m3^2*mQ3^6*msq^4*mU3^6 - 30*mQ3^8*msq^4*mU3^6 + 9*m3^10*mU3^8 + 
      1040*m3^8*mQ3^2*mU3^8 - 2934*m3^6*mQ3^4*mU3^8 + 2772*m3^4*mQ3^6*mU3^8 - 
      779*m3^2*mQ3^8*mU3^8 - 132*mQ3^10*mU3^8 - 180*m3^8*msq^2*mU3^8 - 
      360*m3^6*mQ3^2*msq^2*mU3^8 + 540*m3^4*mQ3^4*msq^2*mU3^8 + 
      270*m3^6*msq^4*mU3^8 - 180*m3^4*mQ3^2*msq^4*mU3^8 - 
      90*m3^2*mQ3^4*msq^4*mU3^8 - 43*m3^8*mU3^10 - 60*m3^6*mQ3^2*mU3^10 + 
      210*m3^4*mQ3^4*mU3^10 - 132*m3^2*mQ3^6*mU3^10 + 33*mQ3^8*mU3^10 + 
      60*m3^6*msq^2*mU3^10 + 120*m3^4*mQ3^2*msq^2*mU3^10 - 
      180*m3^2*mQ3^4*msq^2*mU3^10 - 90*m3^4*msq^4*mU3^10 + 
      60*m3^2*mQ3^2*msq^4*mU3^10 + 30*mQ3^4*msq^4*mU3^10)*Log[mQ3^2/MR^2]^3)/
    (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^4) + 
   (160*a3^2*at*k^(3/2)*mt^2*(m3^10*mQ3^2 - 2*m3^8*mQ3^4 + m3^6*mQ3^6 + 
      m3^10*mU3^2 + 14*m3^8*mQ3^2*mU3^2 - 30*m3^6*mQ3^4*mU3^2 + 
      15*m3^4*mQ3^6*mU3^2 + 18*m3^6*mQ3^2*msq^2*mU3^2 - 
      15*m3^4*mQ3^4*msq^2*mU3^2 + 9*m3^2*mQ3^6*msq^2*mU3^2 - 2*m3^8*mU3^4 - 
      30*m3^6*mQ3^2*mU3^4 + 64*m3^4*mQ3^4*mU3^4 - 32*m3^2*mQ3^6*mU3^4 - 
      15*m3^4*mQ3^2*msq^2*mU3^4 - 12*m3^2*mQ3^4*msq^2*mU3^4 + 
      3*mQ3^6*msq^2*mU3^4 + m3^6*mU3^6 + 15*m3^4*mQ3^2*mU3^6 - 
      32*m3^2*mQ3^4*mU3^6 + 16*mQ3^6*mU3^6 + 9*m3^2*mQ3^2*msq^2*mU3^6 + 
      3*mQ3^4*msq^2*mU3^6)*Log[msq^2/MR^2])/(3*mQ3^2*(-m3^2 + mQ3^2)^2*mU3^2*
     (m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2) + 
   (4*a3^2*at*k^(3/2)*mt^2*(-4*m3^12*mQ3^4 + 8*m3^10*mQ3^6 - 4*m3^8*mQ3^8 + 
      176*m3^12*mQ3^2*mU3^2 + 667*m3^10*mQ3^4*mU3^2 - 1222*m3^8*mQ3^6*mU3^2 + 
      493*m3^6*mQ3^8*mU3^2 - 18*m3^4*mQ3^10*mU3^2 - 360*m3^8*mQ3^4*msq^2*
       mU3^2 + 300*m3^6*mQ3^6*msq^2*mU3^2 - 180*m3^4*mQ3^8*msq^2*mU3^2 + 
      4*m3^12*mU3^4 - 4793*m3^10*mQ3^2*mU3^4 + 10101*m3^8*mQ3^4*mU3^4 - 
      6913*m3^6*mQ3^6*mU3^4 + 1361*m3^4*mQ3^8*mU3^4 - 24*m3^2*mQ3^10*mU3^4 - 
      360*m3^8*mQ3^2*msq^2*mU3^4 + 600*m3^6*mQ3^4*msq^2*mU3^4 + 
      480*m3^4*mQ3^6*msq^2*mU3^4 - 240*m3^2*mQ3^8*msq^2*mU3^4 - 
      140*m3^10*mU3^6 + 7283*m3^8*mQ3^2*mU3^6 - 17667*m3^6*mQ3^4*mU3^6 + 
      13435*m3^4*mQ3^6*mU3^6 - 2665*m3^2*mQ3^8*mU3^6 - 6*mQ3^10*mU3^6 + 
      1020*m3^6*mQ3^2*msq^2*mU3^6 - 1380*m3^4*mQ3^4*msq^2*mU3^6 + 
      420*m3^2*mQ3^6*msq^2*mU3^6 - 60*mQ3^8*msq^2*mU3^6 + 204*m3^8*mU3^8 - 
      1881*m3^6*mQ3^2*mU3^8 + 6073*m3^4*mQ3^4*mU3^8 - 5659*m3^2*mQ3^6*mU3^8 + 
      1167*mQ3^8*mU3^8 - 360*m3^4*mQ3^2*msq^2*mU3^8 - 
      360*m3^2*mQ3^4*msq^2*mU3^8 + 240*mQ3^6*msq^2*mU3^8 - 68*m3^6*mU3^10 - 
      975*m3^4*mQ3^2*mU3^10 + 1508*m3^2*mQ3^4*mU3^10 - 417*mQ3^6*mU3^10 + 
      180*m3^2*mQ3^2*msq^2*mU3^10 + 60*mQ3^4*msq^2*mU3^10 - 
      18*m3^2*mQ3^2*mU3^12 - 6*mQ3^4*mU3^12)*Log[mU3^2/MR^2])/
    (3*mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*(m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)^3) + 
   ((-8*a3^2*at*k^(3/2)*mt^2*(2*m3^16*mQ3^4 - 6*m3^14*mQ3^6 + 6*m3^12*mQ3^8 - 
        2*m3^10*mQ3^10 + 80*m3^16*mQ3^2*mU3^2 - 2152*m3^14*mQ3^4*mU3^2 + 
        4406*m3^12*mQ3^6*mU3^2 - 2996*m3^10*mQ3^8*mU3^2 + 
        643*m3^8*mQ3^10*mU3^2 - 9*m3^6*mQ3^12*mU3^2 + 180*m3^12*mQ3^4*msq^2*
         mU3^2 - 240*m3^10*mQ3^6*msq^2*mU3^2 + 270*m3^8*mQ3^8*msq^2*mU3^2 - 
        90*m3^6*mQ3^10*msq^2*mU3^2 - 2*m3^16*mU3^4 - 3178*m3^14*mQ3^2*mU3^4 + 
        12962*m3^12*mQ3^4*mU3^4 - 17588*m3^10*mQ3^6*mU3^4 + 
        9676*m3^8*mQ3^8*mU3^4 - 1828*m3^6*mQ3^10*mU3^4 + 
        6*m3^4*mQ3^12*mU3^4 + 180*m3^12*mQ3^2*msq^2*mU3^4 - 
        660*m3^10*mQ3^4*msq^2*mU3^4 + 330*m3^8*mQ3^6*msq^2*mU3^4 - 
        270*m3^6*mQ3^8*msq^2*mU3^4 + 60*m3^4*mQ3^10*msq^2*mU3^4 + 
        72*m3^14*mU3^6 + 9110*m3^12*mQ3^2*mU3^6 - 25584*m3^10*mQ3^4*mU3^6 + 
        25638*m3^8*mQ3^6*mU3^6 - 10862*m3^6*mQ3^8*mU3^6 + 
        1723*m3^4*mQ3^10*mU3^6 + 3*m3^2*mQ3^12*mU3^6 - 420*m3^10*mQ3^2*msq^2*
         mU3^6 + 570*m3^8*mQ3^4*msq^2*mU3^6 + 90*m3^6*mQ3^6*msq^2*mU3^6 - 
        30*m3^4*mQ3^8*msq^2*mU3^6 + 30*m3^2*mQ3^10*msq^2*mU3^6 - 
        172*m3^12*mU3^8 - 9926*m3^10*mQ3^2*mU3^8 + 22990*m3^8*mQ3^4*mU3^8 - 
        18070*m3^6*mQ3^6*mU3^8 + 5474*m3^4*mQ3^8*mU3^8 - 
        616*m3^2*mQ3^10*mU3^8 + 510*m3^8*mQ3^2*msq^2*mU3^8 - 
        90*m3^6*mQ3^4*msq^2*mU3^8 - 210*m3^4*mQ3^6*msq^2*mU3^8 + 
        30*m3^2*mQ3^8*msq^2*mU3^8 + 136*m3^10*mU3^10 + 
        4623*m3^8*mQ3^2*mU3^10 - 9553*m3^6*mQ3^4*mU3^10 + 
        6600*m3^4*mQ3^6*mU3^10 - 1602*m3^2*mQ3^8*mU3^10 + 96*mQ3^10*mU3^10 - 
        360*m3^6*mQ3^2*msq^2*mU3^10 - 30*m3^4*mQ3^4*msq^2*mU3^10 + 
        30*m3^2*mQ3^6*msq^2*mU3^10 - 34*m3^8*mU3^12 - 702*m3^6*mQ3^2*mU3^12 + 
        1268*m3^4*mQ3^4*mU3^12 - 884*m3^2*mQ3^6*mU3^12 + 240*mQ3^8*mU3^12 + 
        90*m3^4*mQ3^2*msq^2*mU3^12 + 30*m3^2*mQ3^4*msq^2*mU3^12 + 
        9*m3^4*mQ3^2*mU3^14 + 3*m3^2*mQ3^4*mU3^14)*Log[m3^2/MR^2])/
      (3*mQ3^2*(m3^2 - mQ3^2)^3*mU3^2*(m3^2 - mU3^2)^4*(mQ3^2 - mU3^2)^3) - 
     (8*a3^2*at*k^(3/2)*mt^2*(816*m3^12*mQ3^4 - 1060*m3^10*mQ3^6 + 
        22*m3^8*mQ3^8 + 201*m3^6*mQ3^10 + 67*m3^4*mQ3^12 - 18*m3^2*mQ3^14 - 
        90*m3^8*mQ3^6*msq^2 - 30*m3^6*mQ3^8*msq^2 + 3136*m3^12*mQ3^2*mU3^2 - 
        10916*m3^10*mQ3^4*mU3^2 + 11716*m3^8*mQ3^6*mU3^2 - 
        3835*m3^6*mQ3^8*mU3^2 - 234*m3^4*mQ3^10*mU3^2 + 
        91*m3^2*mQ3^12*mU3^2 - 6*mQ3^14*mU3^2 + 90*m3^8*mQ3^4*msq^2*mU3^2 + 
        360*m3^4*mQ3^8*msq^2*mU3^2 - 90*m3^2*mQ3^10*msq^2*mU3^2 + 
        816*m3^12*mU3^4 - 10988*m3^10*mQ3^2*mU3^4 + 24976*m3^8*mQ3^4*mU3^4 - 
        21226*m3^6*mQ3^6*mU3^4 + 6899*m3^4*mQ3^8*mU3^4 - 
        575*m3^2*mQ3^10*mU3^4 - 2*mQ3^12*mU3^4 + 90*m3^8*mQ3^2*msq^2*mU3^4 + 
        60*m3^6*mQ3^4*msq^2*mU3^4 - 360*m3^4*mQ3^6*msq^2*mU3^4 - 
        30*mQ3^10*msq^2*mU3^4 - 1036*m3^10*mU3^6 + 11772*m3^8*mQ3^2*mU3^6 - 
        20498*m3^6*mQ3^4*mU3^6 + 12164*m3^4*mQ3^6*mU3^6 - 
        2298*m3^2*mQ3^8*mU3^6 + 216*mQ3^10*mU3^6 - 90*m3^8*msq^2*mU3^6 - 
        360*m3^4*mQ3^4*msq^2*mU3^6 + 180*m3^2*mQ3^6*msq^2*mU3^6 + 
        30*mQ3^8*msq^2*mU3^6 - 6*m3^8*mU3^8 - 4191*m3^6*mQ3^2*mU3^8 + 
        6465*m3^4*mQ3^4*mU3^8 - 2320*m3^2*mQ3^6*mU3^8 - 248*mQ3^8*mU3^8 - 
        30*m3^6*msq^2*mU3^8 + 360*m3^4*mQ3^2*msq^2*mU3^8 + 
        30*mQ3^6*msq^2*mU3^8 + 269*m3^6*mU3^10 + 70*m3^4*mQ3^2*mU3^10 - 
        433*m3^2*mQ3^4*mU3^10 + 206*mQ3^6*mU3^10 - 90*m3^2*mQ3^2*msq^2*
         mU3^10 - 30*mQ3^4*msq^2*mU3^10 + 9*m3^4*mU3^12 - 
        15*m3^2*mQ3^2*mU3^12 - 6*mQ3^4*mU3^12)*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)^4))*
    Log[mU3^2/MR^2] + ((32*a3*at*k*mt^2*(2*m3^2 + 3*mQ3^2 + 7*mU3^2))/
      (mQ3^2 - mU3^2)^3 + at*Sqrt[k]*mt^2*((6*mQ3^2)/(mQ3^2 - mU3^2)^3 + 
       (6*mU3^2)/(mQ3^2 - mU3^2)^3) - (8*a3^2*at*k^(3/2)*mt^2*
       (1024*m3^18 - 2786*m3^16*mQ3^2 + 1826*m3^14*mQ3^4 + 549*m3^12*mQ3^6 - 
        664*m3^10*mQ3^8 + 55*m3^8*mQ3^10 - 2786*m3^16*mU3^2 + 
        8004*m3^14*mQ3^2*mU3^2 - 5903*m3^12*mQ3^4*mU3^2 - 
        868*m3^10*mQ3^6*mU3^2 + 1699*m3^8*mQ3^8*mU3^2 - 
        158*m3^6*mQ3^10*mU3^2 + 1826*m3^14*mU3^4 - 5903*m3^12*mQ3^2*mU3^4 + 
        5224*m3^10*mQ3^4*mU3^4 - 292*m3^8*mQ3^6*mU3^4 - 
        846*m3^6*mQ3^8*mU3^4 - m3^4*mQ3^10*mU3^4 + 549*m3^12*mU3^6 - 
        868*m3^10*mQ3^2*mU3^6 - 292*m3^8*mQ3^4*mU3^6 + 928*m3^6*mQ3^6*mU3^6 - 
        453*m3^4*mQ3^8*mU3^6 + 144*m3^2*mQ3^10*mU3^6 - 664*m3^10*mU3^8 + 
        1699*m3^8*mQ3^2*mU3^8 - 846*m3^6*mQ3^4*mU3^8 - 453*m3^4*mQ3^6*mU3^8 + 
        288*m3^2*mQ3^8*mU3^8 - 36*mQ3^10*mU3^8 + 55*m3^8*mU3^10 - 
        158*m3^6*mQ3^2*mU3^10 - m3^4*mQ3^4*mU3^10 + 144*m3^2*mQ3^6*mU3^10 - 
        36*mQ3^8*mU3^10)*Log[m3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^4*
       (m3^2 - mU3^2)^4*(mQ3^2 - mU3^2)^3) + 
     ((-64*a3*at*k*mt^2*(mQ3^4 + 2*mQ3^2*mU3^2 + mU3^4))/(mQ3^2 - mU3^2)^4 + 
       (32*a3^2*at*k^(3/2)*mt^2*(320*m3^18*mQ3^2 - 1036*m3^16*mQ3^4 + 
          1044*m3^14*mQ3^6 - 321*m3^12*mQ3^8 + 36*m3^10*mQ3^10 - 
          41*m3^8*mQ3^12 + 320*m3^18*mU3^2 - 2584*m3^16*mQ3^2*mU3^2 + 
          6252*m3^14*mQ3^4*mU3^2 - 5712*m3^12*mQ3^6*mU3^2 + 
          1612*m3^10*mQ3^8*mU3^2 - 24*m3^8*mQ3^10*mU3^2 + 
          128*m3^6*mQ3^12*mU3^2 - 1036*m3^16*mU3^4 + 6252*m3^14*mQ3^2*mU3^4 - 
          13598*m3^12*mQ3^4*mU3^4 + 12176*m3^10*mQ3^6*mU3^4 - 
          3615*m3^8*mQ3^8*mU3^4 - 44*m3^6*mQ3^10*mU3^4 - 
          125*m3^4*mQ3^12*mU3^4 + 1044*m3^14*mU3^6 - 5712*m3^12*mQ3^2*mU3^6 + 
          12176*m3^10*mQ3^4*mU3^6 - 11360*m3^8*mQ3^6*mU3^6 + 
          3884*m3^6*mQ3^8*mU3^6 - 80*m3^4*mQ3^10*mU3^6 + 
          48*m3^2*mQ3^12*mU3^6 - 321*m3^12*mU3^8 + 1612*m3^10*mQ3^2*mU3^8 - 
          3615*m3^8*mQ3^4*mU3^8 + 3884*m3^6*mQ3^6*mU3^8 - 
          1702*m3^4*mQ3^8*mU3^8 + 144*m3^2*mQ3^10*mU3^8 - 12*mQ3^12*mU3^8 + 
          36*m3^10*mU3^10 - 24*m3^8*mQ3^2*mU3^10 - 44*m3^6*mQ3^4*mU3^10 - 
          80*m3^4*mQ3^6*mU3^10 + 144*m3^2*mQ3^8*mU3^10 - 24*mQ3^10*mU3^10 - 
          41*m3^8*mU3^12 + 128*m3^6*mQ3^2*mU3^12 - 125*m3^4*mQ3^4*mU3^12 + 
          48*m3^2*mQ3^6*mU3^12 - 12*mQ3^8*mU3^12)*Log[m3^2/MR^2])/
        (3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^4*(mQ3^2 - mU3^2)^4))*
      Log[mQ3^2/MR^2] - (4*a3^2*at*k^(3/2)*mt^2*(124*m3^16*mQ3^2 - 
        540*m3^14*mQ3^4 - 178*m3^12*mQ3^6 + 1549*m3^10*mQ3^8 - 
        1004*m3^8*mQ3^10 - 30*m3^6*mQ3^12 + 114*m3^4*mQ3^14 - 
        27*m3^2*mQ3^16 + 60*m3^12*mQ3^4*msq^2 + 120*m3^10*mQ3^6*msq^2 - 
        180*m3^8*mQ3^8*msq^2 - 90*m3^10*mQ3^4*msq^4 + 60*m3^8*mQ3^6*msq^4 + 
        30*m3^6*mQ3^8*msq^4 + 1156*m3^16*mU3^2 - 5372*m3^14*mQ3^2*mU3^2 + 
        9158*m3^12*mQ3^4*mU3^2 - 3530*m3^10*mQ3^6*mU3^2 - 
        4889*m3^8*mQ3^8*mU3^2 + 3876*m3^6*mQ3^10*mU3^2 - 
        528*m3^4*mQ3^12*mU3^2 + 114*m3^2*mQ3^14*mU3^2 - 9*mQ3^16*mU3^2 - 
        180*m3^10*mQ3^4*msq^2*mU3^2 - 360*m3^8*mQ3^6*msq^2*mU3^2 + 
        540*m3^6*mQ3^8*msq^2*mU3^2 + 270*m3^8*mQ3^4*msq^4*mU3^2 - 
        180*m3^6*mQ3^6*msq^4*mU3^2 - 90*m3^4*mQ3^8*msq^4*mU3^2 - 
        3032*m3^14*mU3^4 + 14790*m3^12*mQ3^2*mU3^4 - 25872*m3^10*mQ3^4*
         mU3^4 + 15488*m3^8*mQ3^6*mU3^4 + 2576*m3^6*mQ3^8*mU3^4 - 
        3994*m3^4*mQ3^10*mU3^4 + 72*m3^2*mQ3^12*mU3^4 - 12*mQ3^14*mU3^4 - 
        60*m3^12*msq^2*mU3^4 - 120*m3^10*mQ3^2*msq^2*mU3^4 + 
        360*m3^8*mQ3^4*msq^2*mU3^4 + 360*m3^6*mQ3^6*msq^2*mU3^4 - 
        540*m3^4*mQ3^8*msq^2*mU3^4 + 90*m3^10*msq^4*mU3^4 - 
        60*m3^8*mQ3^2*msq^4*mU3^4 - 300*m3^6*mQ3^4*msq^4*mU3^4 + 
        180*m3^4*mQ3^6*msq^4*mU3^4 + 90*m3^2*mQ3^8*msq^4*mU3^4 + 
        2486*m3^12*mU3^6 - 13250*m3^10*mQ3^2*mU3^6 + 25026*m3^8*mQ3^4*mU3^6 - 
        18644*m3^6*mQ3^6*mU3^6 + 2738*m3^4*mQ3^8*mU3^6 + 
        1654*m3^2*mQ3^10*mU3^6 + 6*mQ3^12*mU3^6 + 180*m3^10*msq^2*mU3^6 + 
        360*m3^8*mQ3^2*msq^2*mU3^6 - 600*m3^6*mQ3^4*msq^2*mU3^6 - 
        120*m3^4*mQ3^6*msq^2*mU3^6 + 180*m3^2*mQ3^8*msq^2*mU3^6 - 
        270*m3^8*msq^4*mU3^6 + 180*m3^6*mQ3^2*msq^4*mU3^6 + 
        180*m3^4*mQ3^4*msq^4*mU3^6 - 60*m3^2*mQ3^6*msq^4*mU3^6 - 
        30*mQ3^8*msq^4*mU3^6 - 545*m3^10*mU3^8 + 3272*m3^8*mQ3^2*mU3^8 - 
        6734*m3^6*mQ3^4*mU3^8 + 5828*m3^4*mQ3^6*mU3^8 - 
        1721*m3^2*mQ3^8*mU3^8 - 124*mQ3^10*mU3^8 - 180*m3^8*msq^2*mU3^8 - 
        360*m3^6*mQ3^2*msq^2*mU3^8 + 540*m3^4*mQ3^4*msq^2*mU3^8 + 
        270*m3^6*msq^4*mU3^8 - 180*m3^4*mQ3^2*msq^4*mU3^8 - 
        90*m3^2*mQ3^4*msq^4*mU3^8 - 53*m3^8*mU3^10 + 316*m3^6*mQ3^2*mU3^10 - 
        654*m3^4*mQ3^4*mU3^10 + 532*m3^2*mQ3^6*mU3^10 - 133*mQ3^8*mU3^10 + 
        60*m3^6*msq^2*mU3^10 + 120*m3^4*mQ3^2*msq^2*mU3^10 - 
        180*m3^2*mQ3^4*msq^2*mU3^10 - 90*m3^4*msq^4*mU3^10 + 
        60*m3^2*mQ3^2*msq^4*mU3^10 + 30*mQ3^4*msq^4*mU3^10)*
       Log[mQ3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^4*(m3^2 - mU3^2)^3*
       (mQ3^2 - mU3^2)^4))*Log[mU3^2/MR^2] - 
   (8*a3^2*at*k^(3/2)*mt^2*(-176*m3^14*mQ3^2 + 322*m3^12*mQ3^4 - 
      147*m3^10*mQ3^6 - 10*m3^8*mQ3^8 + 20*m3^6*mQ3^10 - 9*m3^4*mQ3^12 + 
      60*m3^10*mQ3^4*msq^2 - 120*m3^8*mQ3^6*msq^2 + 60*m3^6*mQ3^8*msq^2 - 
      90*m3^8*mQ3^4*msq^4 + 180*m3^6*mQ3^6*msq^4 - 90*m3^4*mQ3^8*msq^4 + 
      176*m3^14*mU3^2 - 120*m3^12*mQ3^2*mU3^2 + 201*m3^10*mQ3^4*mU3^2 - 
      474*m3^8*mQ3^6*mU3^2 + 178*m3^6*mQ3^8*mU3^2 + 21*m3^4*mQ3^10*mU3^2 + 
      6*m3^2*mQ3^12*mU3^2 - 120*m3^10*mQ3^2*msq^2*mU3^2 + 
      270*m3^8*mQ3^4*msq^2*mU3^2 - 180*m3^6*mQ3^6*msq^2*mU3^2 + 
      30*m3^4*mQ3^8*msq^2*mU3^2 + 180*m3^8*mQ3^2*msq^4*mU3^2 - 
      300*m3^6*mQ3^4*msq^4*mU3^2 + 60*m3^4*mQ3^6*msq^4*mU3^2 + 
      60*m3^2*mQ3^8*msq^4*mU3^2 - 2586*m3^12*mU3^4 + 6267*m3^10*mQ3^2*mU3^4 - 
      6104*m3^8*mQ3^4*mU3^4 + 3202*m3^6*mQ3^6*mU3^4 - 692*m3^4*mQ3^8*mU3^4 - 
      42*m3^2*mQ3^10*mU3^4 + 3*mQ3^12*mU3^4 + 60*m3^10*msq^2*mU3^4 - 
      360*m3^8*mQ3^2*msq^2*mU3^4 + 420*m3^6*mQ3^4*msq^2*mU3^4 - 
      120*m3^2*mQ3^8*msq^2*mU3^4 - 90*m3^8*msq^4*mU3^4 + 
      60*m3^6*mQ3^2*msq^4*mU3^4 + 180*m3^4*mQ3^4*msq^4*mU3^4 - 
      180*m3^2*mQ3^6*msq^4*mU3^4 + 30*mQ3^8*msq^4*mU3^4 + 5679*m3^10*mU3^6 - 
      13622*m3^8*mQ3^2*mU3^6 + 11390*m3^6*mQ3^4*mU3^6 - 
      4230*m3^4*mQ3^6*mU3^6 + 722*m3^2*mQ3^8*mU3^6 + mQ3^10*mU3^6 + 
      210*m3^8*msq^2*mU3^6 - 60*m3^6*mQ3^2*msq^2*mU3^6 - 
      480*m3^4*mQ3^4*msq^2*mU3^6 + 300*m3^2*mQ3^6*msq^2*mU3^6 + 
      30*mQ3^8*msq^2*mU3^6 + 60*m3^6*msq^4*mU3^6 - 180*m3^4*mQ3^2*msq^4*
       mU3^6 + 180*m3^2*mQ3^4*msq^4*mU3^6 - 60*mQ3^6*msq^4*mU3^6 - 
      4030*m3^8*mU3^8 + 9482*m3^6*mQ3^2*mU3^8 - 7123*m3^4*mQ3^4*mU3^8 + 
      1861*m3^2*mQ3^6*mU3^8 - 190*mQ3^8*mU3^8 - 240*m3^6*msq^2*mU3^8 + 
      480*m3^4*mQ3^2*msq^2*mU3^8 - 240*m3^2*mQ3^4*msq^2*mU3^8 + 
      30*m3^4*msq^4*mU3^8 - 60*m3^2*mQ3^2*msq^4*mU3^8 + 
      30*mQ3^4*msq^4*mU3^8 + 368*m3^6*mU3^10 - 1079*m3^4*mQ3^2*mU3^10 + 
      1015*m3^2*mQ3^4*mU3^10 - 244*mQ3^6*mU3^10 - 30*m3^4*msq^2*mU3^10 + 
      60*m3^2*mQ3^2*msq^2*mU3^10 - 30*mQ3^4*msq^2*mU3^10 + 392*m3^4*mU3^12 - 
      787*m3^2*mQ3^2*mU3^12 + 347*mQ3^4*mU3^12 + 9*m3^2*mU3^14 + 
      3*mQ3^2*mU3^14)*Log[mU3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^4*
     (-mQ3^2 + mU3^2)^4) + 
   ((32*a3*at*k*mt^2*(-(m3^2*mQ3^2) + m3^2*mU3^2 + 2*mQ3^2*mU3^2 + 2*mU3^4))/
      (-mQ3^2 + mU3^2)^4 + (16*a3^2*at*k^(3/2)*mt^2*(148*m3^16*mQ3^2 - 
        336*m3^14*mQ3^4 + 194*m3^12*mQ3^6 + m3^10*mQ3^8 - 13*m3^8*mQ3^10 - 
        788*m3^16*mU3^2 + 1796*m3^14*mQ3^2*mU3^2 - 786*m3^12*mQ3^4*mU3^2 - 
        452*m3^10*mQ3^6*mU3^2 + 190*m3^8*mQ3^8*mU3^2 + 64*m3^6*mQ3^10*mU3^2 + 
        2556*m3^14*mU3^4 - 7366*m3^12*mQ3^2*mU3^4 + 6232*m3^10*mQ3^4*mU3^4 - 
        936*m3^8*mQ3^6*mU3^4 - 479*m3^6*mQ3^8*mU3^4 - 37*m3^4*mQ3^10*mU3^4 - 
        2618*m3^12*mU3^6 + 8472*m3^10*mQ3^2*mU3^6 - 8712*m3^8*mQ3^4*mU3^6 + 
        2648*m3^6*mQ3^6*mU3^6 + 234*m3^4*mQ3^8*mU3^6 - 24*m3^2*mQ3^10*mU3^6 + 
        835*m3^10*mU3^8 - 3031*m3^8*mQ3^2*mU3^8 + 3742*m3^6*mQ3^4*mU3^8 - 
        1612*m3^4*mQ3^6*mU3^8 + 90*m3^2*mQ3^8*mU3^8 + 6*mQ3^10*mU3^8 - 
        58*m3^8*mU3^10 + 116*m3^6*mQ3^2*mU3^10 - 238*m3^4*mQ3^4*mU3^10 + 
        180*m3^2*mQ3^6*mU3^10 - 24*mQ3^8*mU3^10 + 69*m3^6*mU3^12 - 
        123*m3^4*mQ3^2*mU3^12 + 90*m3^2*mQ3^4*mU3^12 - 30*mQ3^6*mU3^12)*
       Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^4*
       (mQ3^2 - mU3^2)^4) - (4*a3^2*at*k^(3/2)*mt^2*(1156*m3^16*mQ3^2 - 
        3032*m3^14*mQ3^4 + 2450*m3^12*mQ3^6 - 503*m3^10*mQ3^8 - 
        155*m3^8*mQ3^10 + 87*m3^6*mQ3^12 - 27*m3^4*mQ3^14 - 
        60*m3^12*mQ3^4*msq^2 + 180*m3^10*mQ3^6*msq^2 - 180*m3^8*mQ3^8*msq^2 + 
        60*m3^6*mQ3^10*msq^2 + 90*m3^10*mQ3^4*msq^4 - 270*m3^8*mQ3^6*msq^4 + 
        270*m3^6*mQ3^8*msq^4 - 90*m3^4*mQ3^10*msq^4 + 124*m3^16*mU3^2 - 
        5372*m3^14*mQ3^2*mU3^2 + 14826*m3^12*mQ3^4*mU3^2 - 
        13214*m3^10*mQ3^6*mU3^2 + 3560*m3^8*mQ3^8*mU3^2 + 
        136*m3^6*mQ3^10*mU3^2 + 18*m3^4*mQ3^12*mU3^2 + 18*m3^2*mQ3^14*mU3^2 - 
        120*m3^10*mQ3^4*msq^2*mU3^2 + 360*m3^8*mQ3^6*msq^2*mU3^2 - 
        360*m3^6*mQ3^8*msq^2*mU3^2 + 120*m3^4*mQ3^10*msq^2*mU3^2 - 
        60*m3^8*mQ3^4*msq^4*mU3^2 + 180*m3^6*mQ3^6*msq^4*mU3^2 - 
        180*m3^4*mQ3^8*msq^4*mU3^2 + 60*m3^2*mQ3^10*msq^4*mU3^2 - 
        540*m3^14*mU3^4 + 9194*m3^12*mQ3^2*mU3^4 - 25992*m3^10*mQ3^4*mU3^4 + 
        24798*m3^8*mQ3^6*mU3^4 - 7151*m3^6*mQ3^8*mU3^4 - 
        321*m3^4*mQ3^10*mU3^4 - 117*m3^2*mQ3^12*mU3^4 + 9*mQ3^14*mU3^4 + 
        60*m3^12*msq^2*mU3^4 - 180*m3^10*mQ3^2*msq^2*mU3^4 + 
        360*m3^8*mQ3^4*msq^2*mU3^4 - 600*m3^6*mQ3^6*msq^2*mU3^4 + 
        540*m3^4*mQ3^8*msq^2*mU3^4 - 180*m3^2*mQ3^10*msq^2*mU3^4 - 
        90*m3^10*msq^4*mU3^4 + 270*m3^8*mQ3^2*msq^4*mU3^4 - 
        300*m3^6*mQ3^4*msq^4*mU3^4 + 180*m3^4*mQ3^6*msq^4*mU3^4 - 
        90*m3^2*mQ3^8*msq^4*mU3^4 + 30*mQ3^10*msq^4*mU3^4 - 214*m3^12*mU3^6 - 
        3566*m3^10*mQ3^2*mU3^6 + 15344*m3^8*mQ3^4*mU3^6 - 
        17852*m3^6*mQ3^6*mU3^6 + 5786*m3^4*mQ3^8*mU3^6 + 
        490*m3^2*mQ3^10*mU3^6 + 12*mQ3^12*mU3^6 + 120*m3^10*msq^2*mU3^6 - 
        360*m3^8*mQ3^2*msq^2*mU3^6 + 360*m3^6*mQ3^4*msq^2*mU3^6 - 
        120*m3^4*mQ3^6*msq^2*mU3^6 + 60*m3^8*msq^4*mU3^6 - 
        180*m3^6*mQ3^2*msq^4*mU3^6 + 180*m3^4*mQ3^4*msq^4*mU3^6 - 
        60*m3^2*mQ3^6*msq^4*mU3^6 + 1627*m3^10*mU3^8 - 
        4559*m3^8*mQ3^2*mU3^8 + 2717*m3^6*mQ3^4*mU3^8 + 
        2045*m3^4*mQ3^6*mU3^8 - 1568*m3^2*mQ3^8*mU3^8 - 142*mQ3^10*mU3^8 - 
        180*m3^8*msq^2*mU3^8 + 540*m3^6*mQ3^2*msq^2*mU3^8 - 
        540*m3^4*mQ3^4*msq^2*mU3^8 + 180*m3^2*mQ3^6*msq^2*mU3^8 + 
        30*m3^6*msq^4*mU3^8 - 90*m3^4*mQ3^2*msq^4*mU3^8 + 
        90*m3^2*mQ3^4*msq^4*mU3^8 - 30*mQ3^6*msq^4*mU3^8 - 1148*m3^8*mU3^10 + 
        3264*m3^6*mQ3^2*mU3^10 - 3856*m3^4*mQ3^4*mU3^10 + 
        1792*m3^2*mQ3^6*mU3^10 - 148*mQ3^8*mU3^10 + 159*m3^6*mU3^12 - 
        141*m3^4*mQ3^2*mU3^12 + 9*m3^2*mQ3^4*mU3^12 - 3*mQ3^6*mU3^12)*
       Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^4*
       (-mQ3^2 + mU3^2)^4))*Log[mU3^2/MR^2]^2 + 
   (4*a3^2*at*k^(3/2)*mt^2*(124*m3^10*mQ3^2 + 44*m3^8*mQ3^4 - 2*m3^6*mQ3^6 + 
      9*m3^4*mQ3^8 - 60*m3^6*mQ3^4*msq^2 + 90*m3^4*mQ3^4*msq^4 + 
      1156*m3^10*mU3^2 - 872*m3^8*mQ3^2*mU3^2 + 66*m3^6*mQ3^4*mU3^2 - 
      24*m3^4*mQ3^6*mU3^2 - 6*m3^2*mQ3^8*mU3^2 - 120*m3^4*mQ3^4*msq^2*mU3^2 - 
      60*m3^2*mQ3^4*msq^4*mU3^2 - 4276*m3^8*mU3^4 + 2314*m3^6*mQ3^2*mU3^4 - 
      222*m3^4*mQ3^4*mU3^4 + 30*m3^2*mQ3^6*mU3^4 - 3*mQ3^8*mU3^4 + 
      60*m3^6*msq^2*mU3^4 + 180*m3^2*mQ3^4*msq^2*mU3^4 - 
      90*m3^4*msq^4*mU3^4 - 30*mQ3^4*msq^4*mU3^4 + 4726*m3^6*mU3^6 - 
      2344*m3^4*mQ3^2*mU3^6 + 134*m3^2*mQ3^4*mU3^6 - 4*mQ3^6*mU3^6 + 
      120*m3^4*msq^2*mU3^6 + 60*m3^2*msq^4*mU3^6 - 1163*m3^4*mU3^8 + 
      638*m3^2*mQ3^2*mU3^8 - 30*mQ3^4*mU3^8 - 180*m3^2*msq^2*mU3^8 + 
      30*msq^4*mU3^8 - 604*m3^2*mU3^10 + 140*mQ3^2*mU3^10 + 169*mU3^12)*
     Log[mU3^2/MR^2]^3)/(3*(m3^2 - mU3^2)^4*(-mQ3^2 + mU3^2)^4) + 
   Log[msq^2/MR^2]^2*((-48*a3^2*at*k^(3/2)*mt^2*
       ((-5*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 
           3*m3^2*msq^2 + mQ3^2*msq^2))/(6*(-m3^2 + mQ3^2)^3) - 
        (5*(-2*msq^2 + 2*mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
           msq^2*mU3^2 + mU3^4))/(6*(-m3^2 + mU3^2)^3)))/(mQ3^2 - mU3^2)^2 + 
     (24*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*
       ((-5*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 
           3*m3^2*msq^2 + mQ3^2*msq^2))/(6*(-m3^2 + mQ3^2)^3) - 
        (5*(-2*msq^2 + 2*mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
           msq^2*mU3^2 + mU3^4))/(6*(-m3^2 + mU3^2)^3))*Log[mQ3^2/MR^2])/
      (mQ3^2 - mU3^2)^3 - (24*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*
       ((-5*(2*mQ3^2 - 2*msq^2)*(-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 
           3*m3^2*msq^2 + mQ3^2*msq^2))/(6*(-m3^2 + mQ3^2)^3) - 
        (5*(-2*msq^2 + 2*mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
           msq^2*mU3^2 + mU3^4))/(6*(-m3^2 + mU3^2)^3))*Log[mU3^2/MR^2])/
      (mQ3^2 - mU3^2)^3) + Log[msq^2/MR^2]*
    ((-160*a3^2*at*k^(3/2)*mt^2*(m3^14*mQ3^2 - 3*m3^12*mQ3^4 + 
        3*m3^10*mQ3^6 - m3^8*mQ3^8 + m3^14*mU3^2 - 10*m3^12*mQ3^2*mU3^2 + 
        32*m3^10*mQ3^4*mU3^2 - 16*m3^8*mQ3^6*mU3^2 + 5*m3^6*mQ3^8*mU3^2 - 
        3*m3^12*mU3^4 + 32*m3^10*mQ3^2*mU3^4 - 102*m3^8*mQ3^4*mU3^4 + 
        54*m3^6*mQ3^6*mU3^4 - 17*m3^4*mQ3^8*mU3^4 + 3*m3^10*mU3^6 - 
        16*m3^8*mQ3^2*mU3^6 + 54*m3^6*mQ3^4*mU3^6 - 6*m3^4*mQ3^6*mU3^6 + 
        m3^2*mQ3^8*mU3^6 - m3^8*mU3^8 + 5*m3^6*mQ3^2*mU3^8 - 
        17*m3^4*mQ3^4*mU3^8 + m3^2*mQ3^6*mU3^8)*Log[m3^2/MR^2])/
      (3*mQ3^2*(m3^2 - mQ3^2)^3*mU3^2*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^2) + 
     (a3^2*at*k^(3/2)*mt^2*((160*(2*m3^4*msq^2 + 6*m3^2*mQ3^2*msq^2 - 
            3*m3^2*msq^4 - mQ3^2*msq^4))/((m3^2 - mQ3^2)^3*
           (mQ3^2 - mU3^2)^2) - (80*(mQ3^2 + mU3^2)*(11*m3^8 - 
            22*m3^6*mQ3^2 + 11*m3^4*mQ3^4 + 18*m3^6*msq^2 - 
            15*m3^4*mQ3^2*msq^2 + 9*m3^2*mQ3^4*msq^2 - 22*m3^6*mU3^2 + 
            44*m3^4*mQ3^2*mU3^2 - 22*m3^2*mQ3^4*mU3^2 - 15*m3^4*msq^2*mU3^2 - 
            12*m3^2*mQ3^2*msq^2*mU3^2 + 3*mQ3^4*msq^2*mU3^2 + 11*m3^4*mU3^4 - 
            22*m3^2*mQ3^2*mU3^4 + 11*mQ3^4*mU3^4 + 9*m3^2*msq^2*mU3^4 + 
            3*mQ3^2*msq^2*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*
           (mQ3^2 - mU3^2)^3) - (80*(4*m3^8 - 3*mQ3^4*mU3^2*
             (5*mQ3^2 + mU3^2) + m3^6*(7*mQ3^2 + 3*mU3^2) - 
            m3^4*(29*mQ3^4 + 15*mQ3^2*mU3^2 + 6*mU3^4) + 
            m3^2*(16*mQ3^6 + 31*mQ3^4*mU3^2 + 7*mQ3^2*mU3^4)))/
          (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)^3)) - 
       (160*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*(2*m3^12 - 10*m3^10*mQ3^2 + 
          3*m3^8*mQ3^4 - m3^6*mQ3^6 - 10*m3^10*mU3^2 + 42*m3^8*mQ3^2*mU3^2 - 
          21*m3^6*mQ3^4*mU3^2 + 7*m3^4*mQ3^6*mU3^2 + 3*m3^8*mU3^4 - 
          21*m3^6*mQ3^2*mU3^4 - m3^6*mU3^6 + 7*m3^4*mQ3^2*mU3^6)*
         Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
         (mQ3^2 - mU3^2)^3))*Log[mQ3^2/MR^2] + a3^2*at*k^(3/2)*mt^2*
      ((-80*(2*m3^4*msq^2 + 6*m3^2*mQ3^2*msq^2 - 3*m3^2*msq^4 - mQ3^2*msq^4)*
         (mQ3^2 + mU3^2))/((m3^2 - mQ3^2)^3*(mQ3^2 - mU3^2)^3) + 
       (80*(5*mQ3^8 + 8*m3^4*mQ3^2*mU3^2 + 4*mQ3^6*mU3^2 - mQ3^4*mU3^4 + 
          2*m3^6*(mQ3^2 - mU3^2) - 2*m3^2*(4*mQ3^6 + 5*mQ3^4*mU3^2 - 
            mQ3^2*mU3^4)))/(3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)^4))*
      Log[mQ3^2/MR^2]^2 + 
     (a3^2*at*k^(3/2)*mt^2*((160*(2*m3^4*msq^2 - 3*m3^2*msq^4 + 
            6*m3^2*msq^2*mU3^2 - msq^4*mU3^2))/((m3^2 - mU3^2)^3*
           (-mQ3^2 + mU3^2)^2) + (80*(mQ3^2 + mU3^2)*(11*m3^8 - 
            22*m3^6*mQ3^2 + 11*m3^4*mQ3^4 + 18*m3^6*msq^2 - 
            15*m3^4*mQ3^2*msq^2 + 9*m3^2*mQ3^4*msq^2 - 22*m3^6*mU3^2 + 
            44*m3^4*mQ3^2*mU3^2 - 22*m3^2*mQ3^4*mU3^2 - 15*m3^4*msq^2*mU3^2 - 
            12*m3^2*mQ3^2*msq^2*mU3^2 + 3*mQ3^4*msq^2*mU3^2 + 11*m3^4*mU3^4 - 
            22*m3^2*mQ3^2*mU3^4 + 11*mQ3^4*mU3^4 + 9*m3^2*msq^2*mU3^4 + 
            3*mQ3^2*msq^2*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*
           (mQ3^2 - mU3^2)^3) + (80*(4*m3^8 - 3*mQ3^2*mU3^4*
             (mQ3^2 + 5*mU3^2) + m3^6*(3*mQ3^2 + 7*mU3^2) - 
            m3^4*(6*mQ3^4 + 15*mQ3^2*mU3^2 + 29*mU3^4) + 
            m3^2*(7*mQ3^4*mU3^2 + 31*mQ3^2*mU3^4 + 16*mU3^6)))/
          (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3)) + 
       (160*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*(2*m3^12 - 10*m3^10*mQ3^2 + 
          3*m3^8*mQ3^4 - m3^6*mQ3^6 - 10*m3^10*mU3^2 + 42*m3^8*mQ3^2*mU3^2 - 
          21*m3^6*mQ3^4*mU3^2 + 7*m3^4*mQ3^6*mU3^2 + 3*m3^8*mU3^4 - 
          21*m3^6*mQ3^2*mU3^4 - m3^6*mU3^6 + 7*m3^4*mQ3^2*mU3^6)*
         Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
         (mQ3^2 - mU3^2)^3) + a3^2*at*k^(3/2)*mt^2*
        ((80*(2*m3^4*msq^2 + 6*m3^2*mQ3^2*msq^2 - 3*m3^2*msq^4 - mQ3^2*msq^4)*
           (mQ3^2 + mU3^2))/((m3^2 - mQ3^2)^3*(mQ3^2 - mU3^2)^3) + 
         (80*(mQ3^2 + mU3^2)*(2*m3^4*msq^2 - 3*m3^2*msq^4 + 
            6*m3^2*msq^2*mU3^2 - msq^4*mU3^2))/((m3^2 - mU3^2)^3*
           (-mQ3^2 + mU3^2)^3) - (80*(mQ3^2 + mU3^2)*
           (4*m3^8*(mQ3^2 + mU3^2) + 4*mQ3^4*mU3^4*(mQ3^2 + mU3^2) - 
            8*m3^2*mQ3^2*mU3^2*(mQ3^2 + mU3^2)^2 - 
            2*m3^6*(5*mQ3^4 + 6*mQ3^2*mU3^2 + 5*mU3^4) + 
            m3^4*(5*mQ3^6 + 19*mQ3^4*mU3^2 + 19*mQ3^2*mU3^4 + 5*mU3^6)))/
          (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^4))*
        Log[mQ3^2/MR^2])*Log[mU3^2/MR^2] + a3^2*at*k^(3/2)*mt^2*
      ((-80*(mQ3^2 + mU3^2)*(2*m3^4*msq^2 - 3*m3^2*msq^4 + 
          6*m3^2*msq^2*mU3^2 - msq^4*mU3^2))/((m3^2 - mU3^2)^3*
         (-mQ3^2 + mU3^2)^3) + (80*(8*m3^4*mQ3^2*mU3^2 - mQ3^4*mU3^4 + 
          4*mQ3^2*mU3^6 + 5*mU3^8 - 2*m3^6*(mQ3^2 - mU3^2) + 
          2*m3^2*mU3^2*(mQ3^4 - 5*mQ3^2*mU3^2 - 4*mU3^4)))/
        (3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^4))*Log[mU3^2/MR^2]^2) + 
   ((-32*a3*at*k*mt^2*(-2*m3^2 + mQ3^2 + mU3^2))/(mQ3^2 - mU3^2)^3 - 
     (16*a3^2*at*k^(3/2)*mt^2*(172*m3^12 - 372*m3^10*mQ3^2 + 138*m3^8*mQ3^4 + 
        3*m3^6*mQ3^6 + 20*m3^4*mQ3^8 - 9*m3^2*mQ3^10 - 468*m3^10*mU3^2 + 
        1084*m3^8*mQ3^2*mU3^2 - 313*m3^6*mQ3^4*mU3^2 - 149*m3^4*mQ3^6*mU3^2 + 
        41*m3^2*mQ3^8*mU3^2 - 3*mQ3^10*mU3^2 + 398*m3^8*mU3^4 - 
        1151*m3^6*mQ3^2*mU3^4 + 411*m3^4*mQ3^4*mU3^4 + 55*m3^2*mQ3^6*mU3^4 - 
        mQ3^8*mU3^4 - 59*m3^6*mU3^6 + 409*m3^4*mQ3^2*mU3^6 - 
        129*m3^2*mQ3^4*mU3^6 - 29*mQ3^6*mU3^6 - 31*m3^4*mU3^8 - 
        30*m3^2*mQ3^2*mU3^8 + 13*mQ3^4*mU3^8))/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^3) - 
     (64*a3^2*at*k^(3/2)*mt^2*(2*m3^2 - mQ3^2 - mU3^2)*
       (53*m3^8 - 50*m3^6*mQ3^2 - 7*m3^4*mQ3^4 - 50*m3^6*mU3^2 + 
        52*m3^4*mQ3^2*mU3^2 + 6*m3^2*mQ3^4*mU3^2 - 7*m3^4*mU3^4 + 
        6*m3^2*mQ3^2*mU3^4 - 3*mQ3^4*mU3^4)*Log[m3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) + 
     (8*a3^2*at*k^(3/2)*mt^2*(1156*m3^12*mQ3^2 - 1968*m3^10*mQ3^4 + 
        746*m3^8*mQ3^6 - 25*m3^6*mQ3^8 + 20*m3^4*mQ3^10 - 9*m3^2*mQ3^12 + 
        124*m3^12*mU3^2 - 4100*m3^10*mQ3^2*mU3^2 + 6466*m3^8*mQ3^4*mU3^2 - 
        2186*m3^6*mQ3^6*mU3^2 - 29*m3^4*mQ3^8*mU3^2 + 32*m3^2*mQ3^10*mU3^2 - 
        3*mQ3^12*mU3^2 - 332*m3^10*mU3^4 + 5326*m3^8*mQ3^2*mU3^4 - 
        7596*m3^6*mQ3^4*mU3^4 + 2210*m3^4*mQ3^6*mU3^4 - 20*m3^2*mQ3^8*mU3^4 - 
        4*mQ3^10*mU3^4 + 262*m3^8*mU3^6 - 2990*m3^6*mQ3^2*mU3^6 + 
        3712*m3^4*mQ3^4*mU3^6 - 774*m3^2*mQ3^6*mU3^6 + 14*mQ3^8*mU3^6 - 
        3*m3^6*mU3^8 + 526*m3^4*mQ3^2*mU3^8 - 543*m3^2*mQ3^4*mU3^8 + 
        4*mQ3^6*mU3^8 - 39*m3^4*mU3^10 + 34*m3^2*mQ3^2*mU3^10 - 
        11*mQ3^4*mU3^10)*Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^4) - 
     (160*a3^2*at*k^(3/2)*mt^2*(-2*m3^2 + mQ3^2 + mU3^2)*Log[msq^2/MR^2])/
      (3*(mQ3^2 - mU3^2)^3) - (8*a3^2*at*k^(3/2)*mt^2*
       (124*m3^12*mQ3^2 - 284*m3^10*mQ3^4 + 90*m3^8*mQ3^6 + 11*m3^6*mQ3^8 + 
        20*m3^4*mQ3^10 - 9*m3^2*mQ3^12 + 1156*m3^12*mU3^2 - 
        3068*m3^10*mQ3^2*mU3^2 + 2702*m3^8*mQ3^4*mU3^2 - 
        474*m3^6*mQ3^6*mU3^2 - 201*m3^4*mQ3^8*mU3^2 + 32*m3^2*mQ3^10*mU3^2 - 
        3*mQ3^12*mU3^2 - 3048*m3^10*mU3^4 + 7406*m3^8*mQ3^2*mU3^4 - 
        5516*m3^6*mQ3^4*mU3^4 + 882*m3^4*mQ3^6*mU3^4 + 184*m3^2*mQ3^8*mU3^4 - 
        4*mQ3^10*mU3^4 + 2602*m3^8*mU3^6 - 6126*m3^6*mQ3^2*mU3^6 + 
        3984*m3^4*mQ3^4*mU3^6 - 502*m3^2*mQ3^6*mU3^6 - 54*mQ3^8*mU3^6 - 
        695*m3^6*mU3^8 + 1718*m3^4*mQ3^2*mU3^8 - 883*m3^2*mQ3^4*mU3^8 + 
        4*mQ3^6*mU3^8 - 3*m3^4*mU3^10 - 102*m3^2*mQ3^2*mU3^10 + 
        57*mQ3^4*mU3^10)*Log[mU3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)^4))*PolyLog[2, 1 - m3^2/mQ3^2] + 
   (80*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*(2*m3^10 - 8*m3^8*mQ3^2 + 
      3*m3^6*mQ3^4 - m3^4*mQ3^6 + 6*m3^8*msq^2 - 8*m3^6*mQ3^2*msq^2 + 
      9*m3^4*mQ3^4*msq^2 - 3*m3^2*mQ3^6*msq^2 - 8*m3^8*mU3^2 + 
      30*m3^6*mQ3^2*mU3^2 - 15*m3^4*mQ3^4*mU3^2 + 5*m3^2*mQ3^6*mU3^2 - 
      8*m3^6*msq^2*mU3^2 - 6*m3^4*mQ3^2*msq^2*mU3^2 + 
      3*m3^2*mQ3^4*msq^2*mU3^2 - mQ3^6*msq^2*mU3^2 + 3*m3^6*mU3^4 - 
      15*m3^4*mQ3^2*mU3^4 + 9*m3^4*msq^2*mU3^4 + 3*m3^2*mQ3^2*msq^2*mU3^4 - 
      m3^4*mU3^6 + 5*m3^2*mQ3^2*mU3^6 - 3*m3^2*msq^2*mU3^6 - 
      mQ3^2*msq^2*mU3^6)*(-2*mQ3^2 + 2*mU3^2 + mQ3^2*Log[mQ3^2/MR^2] + 
      mU3^2*Log[mQ3^2/MR^2] - mQ3^2*Log[mU3^2/MR^2] - mU3^2*Log[mU3^2/MR^2])*
     PolyLog[2, 1 - m3^2/msq^2])/((m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
     (mQ3^2 - mU3^2)^3) + 
   ((160*a3^2*at*k^(3/2)*(mQ3^2 - msq^2)*(2*m3^4 + 3*m3^2*mQ3^2 - mQ3^4 - 
        3*m3^2*msq^2 - mQ3^2*msq^2)*mt^2)/((m3^2 - mQ3^2)^3*
       (mQ3^2 - mU3^2)^2) - (40*a3^2*at*k^(3/2)*(2*mQ3^2 - 2*msq^2)*
       (-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 3*m3^2*msq^2 + mQ3^2*msq^2)*mt^2*
       (mQ3^2 + mU3^2)*Log[mQ3^2/MR^2])/((-m3^2 + mQ3^2)^3*
       (mQ3^2 - mU3^2)^3) + (40*a3^2*at*k^(3/2)*(2*mQ3^2 - 2*msq^2)*
       (-2*m3^4 - 3*m3^2*mQ3^2 + mQ3^4 + 3*m3^2*msq^2 + mQ3^2*msq^2)*mt^2*
       (mQ3^2 + mU3^2)*Log[mU3^2/MR^2])/((-m3^2 + mQ3^2)^3*
       (mQ3^2 - mU3^2)^3))*PolyLog[2, 1 - msq^2/mQ3^2] + 
   ((32*a3*at*k*mt^2*(-2*m3^2 + mQ3^2 + mU3^2))/(mQ3^2 - mU3^2)^3 + 
     (16*a3^2*at*k^(3/2)*mt^2*(172*m3^12 - 468*m3^10*mQ3^2 + 398*m3^8*mQ3^4 - 
        59*m3^6*mQ3^6 - 31*m3^4*mQ3^8 - 372*m3^10*mU3^2 + 
        1084*m3^8*mQ3^2*mU3^2 - 1151*m3^6*mQ3^4*mU3^2 + 
        409*m3^4*mQ3^6*mU3^2 - 30*m3^2*mQ3^8*mU3^2 + 138*m3^8*mU3^4 - 
        313*m3^6*mQ3^2*mU3^4 + 411*m3^4*mQ3^4*mU3^4 - 129*m3^2*mQ3^6*mU3^4 + 
        13*mQ3^8*mU3^4 + 3*m3^6*mU3^6 - 149*m3^4*mQ3^2*mU3^6 + 
        55*m3^2*mQ3^4*mU3^6 - 29*mQ3^6*mU3^6 + 20*m3^4*mU3^8 + 
        41*m3^2*mQ3^2*mU3^8 - mQ3^4*mU3^8 - 9*m3^2*mU3^10 - 3*mQ3^2*mU3^10))/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) + 
     (64*a3^2*at*k^(3/2)*mt^2*(2*m3^2 - mQ3^2 - mU3^2)*
       (53*m3^8 - 50*m3^6*mQ3^2 - 7*m3^4*mQ3^4 - 50*m3^6*mU3^2 + 
        52*m3^4*mQ3^2*mU3^2 + 6*m3^2*mQ3^4*mU3^2 - 7*m3^4*mU3^4 + 
        6*m3^2*mQ3^2*mU3^4 - 3*mQ3^4*mU3^4)*Log[m3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^3) - 
     (8*a3^2*at*k^(3/2)*mt^2*(1156*m3^12*mQ3^2 - 3048*m3^10*mQ3^4 + 
        2602*m3^8*mQ3^6 - 695*m3^6*mQ3^8 - 3*m3^4*mQ3^10 + 124*m3^12*mU3^2 - 
        3068*m3^10*mQ3^2*mU3^2 + 7406*m3^8*mQ3^4*mU3^2 - 
        6126*m3^6*mQ3^6*mU3^2 + 1718*m3^4*mQ3^8*mU3^2 - 
        102*m3^2*mQ3^10*mU3^2 - 284*m3^10*mU3^4 + 2702*m3^8*mQ3^2*mU3^4 - 
        5516*m3^6*mQ3^4*mU3^4 + 3984*m3^4*mQ3^6*mU3^4 - 
        883*m3^2*mQ3^8*mU3^4 + 57*mQ3^10*mU3^4 + 90*m3^8*mU3^6 - 
        474*m3^6*mQ3^2*mU3^6 + 882*m3^4*mQ3^4*mU3^6 - 502*m3^2*mQ3^6*mU3^6 + 
        4*mQ3^8*mU3^6 + 11*m3^6*mU3^8 - 201*m3^4*mQ3^2*mU3^8 + 
        184*m3^2*mQ3^4*mU3^8 - 54*mQ3^6*mU3^8 + 20*m3^4*mU3^10 + 
        32*m3^2*mQ3^2*mU3^10 - 4*mQ3^4*mU3^10 - 9*m3^2*mU3^12 - 
        3*mQ3^2*mU3^12)*Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*
       (mQ3^2 - mU3^2)^4) + (160*a3^2*at*k^(3/2)*mt^2*
       (-2*m3^2 + mQ3^2 + mU3^2)*Log[msq^2/MR^2])/(3*(mQ3^2 - mU3^2)^3) + 
     (8*a3^2*at*k^(3/2)*mt^2*(124*m3^12*mQ3^2 - 332*m3^10*mQ3^4 + 
        262*m3^8*mQ3^6 - 3*m3^6*mQ3^8 - 39*m3^4*mQ3^10 + 1156*m3^12*mU3^2 - 
        4100*m3^10*mQ3^2*mU3^2 + 5326*m3^8*mQ3^4*mU3^2 - 
        2990*m3^6*mQ3^6*mU3^2 + 526*m3^4*mQ3^8*mU3^2 + 34*m3^2*mQ3^10*mU3^2 - 
        1968*m3^10*mU3^4 + 6466*m3^8*mQ3^2*mU3^4 - 7596*m3^6*mQ3^4*mU3^4 + 
        3712*m3^4*mQ3^6*mU3^4 - 543*m3^2*mQ3^8*mU3^4 - 11*mQ3^10*mU3^4 + 
        746*m3^8*mU3^6 - 2186*m3^6*mQ3^2*mU3^6 + 2210*m3^4*mQ3^4*mU3^6 - 
        774*m3^2*mQ3^6*mU3^6 + 4*mQ3^8*mU3^6 - 25*m3^6*mU3^8 - 
        29*m3^4*mQ3^2*mU3^8 - 20*m3^2*mQ3^4*mU3^8 + 14*mQ3^6*mU3^8 + 
        20*m3^4*mU3^10 + 32*m3^2*mQ3^2*mU3^10 - 4*mQ3^4*mU3^10 - 
        9*m3^2*mU3^12 - 3*mQ3^2*mU3^12)*Log[mU3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)^4))*
    PolyLog[2, 1 - m3^2/mU3^2] + 
   ((-16*a3^2*at*k^(3/2)*mt^2*(12*m3^10 - 14*m3^8*mQ3^2 + 34*m3^6*mQ3^4 - 
        29*m3^4*mQ3^6 + 9*m3^2*mQ3^8 - 14*m3^8*mU3^2 - 76*m3^6*mQ3^2*mU3^2 + 
        65*m3^4*mQ3^4*mU3^2 - 26*m3^2*mQ3^6*mU3^2 + 3*mQ3^8*mU3^2 + 
        34*m3^6*mU3^4 + 65*m3^4*mQ3^2*mU3^4 - 34*m3^2*mQ3^4*mU3^4 + 
        7*mQ3^6*mU3^4 - 29*m3^4*mU3^6 - 26*m3^2*mQ3^2*mU3^6 + 7*mQ3^4*mU3^6 + 
        9*m3^2*mU3^8 + 3*mQ3^2*mU3^8))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
       (mQ3^2 - mU3^2)) + (8*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*
       (12*m3^10 - 14*m3^8*mQ3^2 + 34*m3^6*mQ3^4 - 29*m3^4*mQ3^6 + 
        9*m3^2*mQ3^8 - 14*m3^8*mU3^2 - 76*m3^6*mQ3^2*mU3^2 + 
        65*m3^4*mQ3^4*mU3^2 - 26*m3^2*mQ3^6*mU3^2 + 3*mQ3^8*mU3^2 + 
        34*m3^6*mU3^4 + 65*m3^4*mQ3^2*mU3^4 - 34*m3^2*mQ3^4*mU3^4 + 
        7*mQ3^6*mU3^4 - 29*m3^4*mU3^6 - 26*m3^2*mQ3^2*mU3^6 + 7*mQ3^4*mU3^6 + 
        9*m3^2*mU3^8 + 3*mQ3^2*mU3^8)*Log[mQ3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)^2) - 
     (8*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*(12*m3^10 - 14*m3^8*mQ3^2 + 
        34*m3^6*mQ3^4 - 29*m3^4*mQ3^6 + 9*m3^2*mQ3^8 - 14*m3^8*mU3^2 - 
        76*m3^6*mQ3^2*mU3^2 + 65*m3^4*mQ3^4*mU3^2 - 26*m3^2*mQ3^6*mU3^2 + 
        3*mQ3^8*mU3^2 + 34*m3^6*mU3^4 + 65*m3^4*mQ3^2*mU3^4 - 
        34*m3^2*mQ3^4*mU3^4 + 7*mQ3^6*mU3^4 - 29*m3^4*mU3^6 - 
        26*m3^2*mQ3^2*mU3^6 + 7*mQ3^4*mU3^6 + 9*m3^2*mU3^8 + 3*mQ3^2*mU3^8)*
       Log[mU3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*
       (mQ3^2 - mU3^2)^2))*PolyLog[2, 1 - mQ3^2/mU3^2] + 
   ((160*a3^2*at*k^(3/2)*mt^2*(-msq^2 + mU3^2)*(2*m3^4 - 3*m3^2*msq^2 + 
        3*m3^2*mU3^2 - msq^2*mU3^2 - mU3^4))/((m3^2 - mU3^2)^3*
       (-mQ3^2 + mU3^2)^2) - (80*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*
       (-msq^2 + mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
        msq^2*mU3^2 + mU3^4)*Log[mQ3^2/MR^2])/((mQ3^2 - mU3^2)^3*
       (-m3^2 + mU3^2)^3) + (80*a3^2*at*k^(3/2)*mt^2*(mQ3^2 + mU3^2)*
       (-msq^2 + mU3^2)*(-2*m3^4 + 3*m3^2*msq^2 - 3*m3^2*mU3^2 + 
        msq^2*mU3^2 + mU3^4)*Log[mU3^2/MR^2])/((mQ3^2 - mU3^2)^3*
       (-m3^2 + mU3^2)^3))*PolyLog[2, 1 - msq^2/mU3^2]) + 
 Xt*(((-128*a3*at*k*m3^3*mt^2*Log[m3^2/MR^2])/((m3^2 - mQ3^2)*
       (mQ3^2 - mU3^2)) - (64*a3^2*at*k^(3/2)*mt^2*
       (44*m3^13 - 79*m3^11*mQ3^2 + 122*m3^9*mQ3^4 - 63*m3^7*mQ3^6 - 
        77*m3^11*mU3^2 + 72*m3^9*mQ3^2*mU3^2 - 209*m3^7*mQ3^4*mU3^2 + 
        142*m3^5*mQ3^6*mU3^2 + 22*m3^9*mU3^4 + 101*m3^7*mQ3^2*mU3^4 + 
        36*m3^5*mQ3^4*mU3^4 - 87*m3^3*mQ3^6*mU3^4 + 19*m3^7*mU3^6 - 
        118*m3^5*mQ3^2*mU3^6 + 75*m3^3*mQ3^4*mU3^6)*Log[m3^2/MR^2]^2)/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)))*
    Log[mQ3^2/MR^2] + ((128*a3*at*k*m3^3*mt^2)/((m3^2 - mQ3^2)*
       (mQ3^2 - mU3^2)) + (32*a3^2*at*k^(3/2)*mt^2*
       (18*m3^11 + 195*m3^9*mQ3^2 - 130*m3^7*mQ3^4 + 13*m3^5*mQ3^6 - 
        19*m3^9*mU3^2 - 425*m3^7*mQ3^2*mU3^2 + 263*m3^5*mQ3^4*mU3^2 - 
        11*m3^3*mQ3^6*mU3^2 + 27*m3^7*mU3^4 + 168*m3^5*mQ3^2*mU3^4 - 
        87*m3^3*mQ3^4*mU3^4 - 12*m3*mQ3^6*mU3^4)*Log[m3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)))*
    Log[mQ3^2/MR^2]^2 + (32*a3^2*at*k^(3/2)*mt^2*(20*m3^11*mQ3^2 - 
      202*m3^9*mQ3^4 + 194*m3^7*mQ3^6 - 65*m3^5*mQ3^8 + 8*m3^3*mQ3^10 - 
      3*m3*mQ3^12 + 60*m3^7*mQ3^4*msq^2 - 60*m3^5*mQ3^6*msq^2 - 
      30*m3^7*mQ3^2*msq^4 + 30*m3^5*mQ3^4*msq^4 - 20*m3^11*mU3^2 + 
      145*m3^9*mQ3^2*mU3^2 + 240*m3^7*mQ3^4*mU3^2 - 321*m3^5*mQ3^6*mU3^2 + 
      90*m3^3*mQ3^8*mU3^2 + 10*m3*mQ3^10*mU3^2 - 60*m3^7*mQ3^2*msq^2*mU3^2 - 
      60*m3^5*mQ3^4*msq^2*mU3^2 + 120*m3^3*mQ3^6*msq^2*mU3^2 + 
      30*m3^7*msq^4*mU3^2 + 30*m3^5*mQ3^2*msq^4*mU3^2 - 
      60*m3^3*mQ3^4*msq^4*mU3^2 + 41*m3^9*mU3^4 - 346*m3^7*mQ3^2*mU3^4 + 
      91*m3^5*mQ3^4*mU3^4 + 140*m3^3*mQ3^6*mU3^4 - 70*m3*mQ3^8*mU3^4 + 
      120*m3^5*mQ3^2*msq^2*mU3^4 - 60*m3^3*mQ3^4*msq^2*mU3^4 - 
      60*m3*mQ3^6*msq^2*mU3^4 - 60*m3^5*msq^4*mU3^4 + 
      30*m3^3*mQ3^2*msq^4*mU3^4 + 30*m3*mQ3^4*msq^4*mU3^4 - 24*m3^7*mU3^6 + 
      199*m3^5*mQ3^2*mU3^6 - 174*m3^3*mQ3^4*mU3^6 + 47*m3*mQ3^6*mU3^6 - 
      60*m3^3*mQ3^2*msq^2*mU3^6 + 60*m3*mQ3^4*msq^2*mU3^6 + 
      30*m3^3*msq^4*mU3^6 - 30*m3*mQ3^2*msq^4*mU3^6)*Log[mQ3^2/MR^2]^3)/
    (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2) + 
   ((128*a3*at*k*m3^3*mt^2)/((m3^2 - mU3^2)*(-mQ3^2 + mU3^2)) - 
     (32*a3^2*at*k^(3/2)*mt^2*(18*m3^11 - 19*m3^9*mQ3^2 + 27*m3^7*mQ3^4 + 
        195*m3^9*mU3^2 - 425*m3^7*mQ3^2*mU3^2 + 168*m3^5*mQ3^4*mU3^2 - 
        130*m3^7*mU3^4 + 263*m3^5*mQ3^2*mU3^4 - 87*m3^3*mQ3^4*mU3^4 + 
        13*m3^5*mU3^6 - 11*m3^3*mQ3^2*mU3^6 - 12*m3*mQ3^4*mU3^6)*
       Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*
       (mQ3^2 - mU3^2)) - (32*a3^2*at*k^(3/2)*mt^2*(20*m3^11*mQ3^2 - 
        56*m3^9*mQ3^4 + 12*m3^7*mQ3^6 - 5*m3^5*mQ3^8 + 3*m3^3*mQ3^10 - 
        30*m3^7*mQ3^2*msq^4 + 60*m3^5*mQ3^4*msq^4 - 30*m3^3*mQ3^6*msq^4 - 
        20*m3^11*mU3^2 - 51*m3^9*mQ3^2*mU3^2 + 229*m3^7*mQ3^4*mU3^2 - 
        62*m3^5*mQ3^6*mU3^2 - 5*m3^3*mQ3^8*mU3^2 - 3*m3*mQ3^10*mU3^2 + 
        60*m3^7*mQ3^2*msq^2*mU3^2 - 120*m3^5*mQ3^4*msq^2*mU3^2 + 
        60*m3^3*mQ3^6*msq^2*mU3^2 + 30*m3^7*msq^4*mU3^2 - 
        30*m3^5*mQ3^2*msq^4*mU3^2 - 30*m3^3*mQ3^4*msq^4*mU3^2 + 
        30*m3*mQ3^6*msq^4*mU3^2 + 91*m3^9*mU3^4 - 50*m3^7*mQ3^2*mU3^4 - 
        263*m3^5*mQ3^4*mU3^4 + 104*m3^3*mQ3^6*mU3^4 + 10*m3*mQ3^8*mU3^4 - 
        60*m3^7*msq^2*mU3^4 + 60*m3^5*mQ3^2*msq^2*mU3^4 + 
        60*m3^3*mQ3^4*msq^2*mU3^4 - 60*m3*mQ3^6*msq^2*mU3^4 - 
        30*m3^5*msq^4*mU3^4 + 60*m3^3*mQ3^2*msq^4*mU3^4 - 
        30*m3*mQ3^4*msq^4*mU3^4 - 127*m3^7*mU3^6 + 194*m3^5*mQ3^2*mU3^6 + 
        27*m3^3*mQ3^4*mU3^6 - 38*m3*mQ3^6*mU3^6 + 60*m3^5*msq^2*mU3^6 - 
        120*m3^3*mQ3^2*msq^2*mU3^6 + 60*m3*mQ3^4*msq^2*mU3^6 + 
        40*m3^5*mU3^8 - 65*m3^3*mQ3^2*mU3^8 + 15*m3*mQ3^4*mU3^8)*
       Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*
       (mQ3^2 - mU3^2)^2))*Log[mU3^2/MR^2]^2 + 
   (32*a3^2*at*k^(3/2)*mt^2*(-20*m3^7*mQ3^2 + m3^5*mQ3^4 + 3*m3^3*mQ3^6 + 
      30*m3^3*mQ3^2*msq^4 + 20*m3^7*mU3^2 + 185*m3^5*mQ3^2*mU3^2 - 
      11*m3^3*mQ3^4*mU3^2 - 3*m3*mQ3^6*mU3^2 - 60*m3^3*mQ3^2*msq^2*mU3^2 - 
      30*m3^3*msq^4*mU3^2 - 30*m3*mQ3^2*msq^4*mU3^2 - 202*m3^5*mU3^4 - 
      149*m3^3*mQ3^2*mU3^4 + 10*m3*mQ3^4*mU3^4 + 60*m3^3*msq^2*mU3^4 + 
      60*m3*mQ3^2*msq^2*mU3^4 + 30*m3*msq^4*mU3^4 + 189*m3^3*mU3^6 + 
      32*m3*mQ3^2*mU3^6 - 60*m3*msq^2*mU3^6 - 55*m3*mU3^8)*Log[mU3^2/MR^2]^3)/
    (3*(m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)^2) + 
   Log[msq^2/MR^2]^2*(-24*a3^2*at*k^(3/2)*mt^2*
      ((40*m3*(mQ3^2 - msq^2)^2)/(3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) + 
       (40*m3*(-msq^2 + mU3^2)^2)/(3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)))*
      Log[mQ3^2/MR^2] - 24*a3^2*at*k^(3/2)*mt^2*
      ((40*m3*(mQ3^2 - msq^2)^2)/(3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) + 
       (40*m3*(-msq^2 + mU3^2)^2)/(3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)))*
      Log[mU3^2/MR^2] - (320*a3^2*at*k^(3/2)*m3*mt^2*
       (m3^4*mQ3^2 - 2*m3^4*msq^2 + 2*m3^2*msq^4 - mQ3^2*msq^4 + m3^4*mU3^2 - 
        2*m3^2*mQ3^2*mU3^2 + 2*mQ3^2*msq^2*mU3^2 - msq^4*mU3^2)*
       (-1 + 2*Log[MR^2] - 4*Log[mt] - 4*Log[Csc[beta]] - 4*Log[Sin[beta]]))/
      ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)) + 
   Log[msq^2/MR^2]*((320*a3^2*at*k^(3/2)*(2*m3^5 - 3*m3^3*mQ3^2 + m3*mQ3^4 - 
        12*m3*mQ3^2*msq^2 + 6*m3*msq^4)*mt^2*Log[mQ3^2/MR^2]^2)/
      (3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) + 
     (320*a3^2*at*k^(3/2)*mt^2*(2*m3^5 + 6*m3*msq^4 - 3*m3^3*mU3^2 - 
        12*m3*msq^2*mU3^2 + m3*mU3^4)*Log[mU3^2/MR^2]^2)/
      (3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)) + Log[mQ3^2/MR^2]*
      (a3^2*at*k^(3/2)*mt^2*((320*m3^3*(-2*m3^2 + mQ3^2 + mU3^2))/
          (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) + 
         (640*(m3^5*mQ3^2 + m3^5*mU3^2 - 2*m3^3*mQ3^2*mU3^2))/
          ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*Log[m3^2/MR^2] + 
       a3^2*at*k^(3/2)*mt^2*((-640*m3*msq^2)/((m3^2 - mQ3^2)*
           (m3^2 - mU3^2)) - (640*m3)/(3*(mQ3^2 - mU3^2)) - 
         (320*m3*mQ3^2)/(3*(m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) + 
         (320*(m3^3*mQ3^2 - m3*mQ3^4 + 12*m3*mQ3^2*msq^2 - 6*m3*msq^4))/
          (3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) - 
         (1280*(2*m3*mQ3^2*msq^2 - m3*msq^4)*(Log[MR^2] - 2*Log[mt] - 
            2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/((m3^2 - mQ3^2)^2*
           (mQ3^2 - mU3^2)))) + Log[mU3^2/MR^2]*
      (a3^2*at*k^(3/2)*mt^2*((320*m3^3*(2*m3^2 - mQ3^2 - mU3^2))/
          (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) + 
         (640*(m3^5*mQ3^2 + m3^5*mU3^2 - 2*m3^3*mQ3^2*mU3^2))/
          ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*Log[m3^2/MR^2] - 
       (320*a3^2*at*k^(3/2)*mt^2*(m3^7 - m3^5*mQ3^2 + 12*m3^5*msq^2 - 
          12*m3^3*msq^4 + 6*m3*mQ3^2*msq^4 - m3^5*mU3^2 + m3^3*mQ3^2*mU3^2 - 
          12*m3*mQ3^2*msq^2*mU3^2 + 6*m3*msq^4*mU3^2)*Log[mQ3^2/MR^2])/
        (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + a3^2*at*k^(3/2)*mt^2*
        ((-640*m3*msq^2)/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) + 
         (640*m3)/(3*(mQ3^2 - mU3^2)) - (320*m3*mU3^2)/(3*(m3^2 - mU3^2)*
           (-mQ3^2 + mU3^2)) + (320*(-6*m3*msq^4 + m3^3*mU3^2 + 
            12*m3*msq^2*mU3^2 - m3*mU3^4))/(3*(m3^2 - mU3^2)^2*
           (-mQ3^2 + mU3^2)) - (1280*(-(m3*msq^4) + 2*m3*msq^2*mU3^2)*
           (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
          ((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)))) + 
     (640*a3^2*at*k^(3/2)*mt^2*(m3^5*mQ3^2 + m3^5*mU3^2 - 2*m3^3*mQ3^2*mU3^2)*
       Log[m3^2/MR^2]*(-1 + 2*Log[MR^2] - 4*Log[mt] - 4*Log[Csc[beta]] - 
        4*Log[Sin[beta]]))/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2)) + 
   a3^2*at*k^(3/2)*mt^2*((-64*(14*m3^3 - 3*m3*mQ3^2 - 30*m3*msq^2 - 
        3*m3*mU3^2))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)) + 
     (128*(14*m3^3 - 3*m3*mQ3^2 - 30*m3*msq^2 - 3*m3*mU3^2)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2))) + a3^2*at*k^(3/2)*mt^2*
    Log[m3^2/MR^2]^2*((128*m3^3)/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) + 
     (384*m3^7*(2*m3^4 - 2*m3^2*mQ3^2 + mQ3^4 - 2*m3^2*mU3^2 + mU3^4))/
      ((m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) - 
     (64*(27*m3^11 - 82*m3^9*mQ3^2 + 57*m3^7*mQ3^4 - 82*m3^9*mU3^2 + 
        220*m3^7*mQ3^2*mU3^2 - 142*m3^5*mQ3^4*mU3^2 + 57*m3^7*mU3^4 - 
        142*m3^5*mQ3^2*mU3^4 + 87*m3^3*mQ3^4*mU3^4))/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3) + 
     (512*m3^7*(2*m3^6*(mQ3^2 + mU3^2) + 2*m3^2*(mQ3^2 - mU3^2)^2*
         (mQ3^2 + mU3^2) + m3^4*(-4*mQ3^4 + 2*mQ3^2*mU3^2 - 4*mU3^4) + 
        mQ3^2*mU3^2*(mQ3^4 + mU3^4)))/(3*mQ3^2*(m3^2 - mQ3^2)^3*mU3^2*
       (m3^2 - mU3^2)^3) + (128*(m3^11 - 50*m3^9*mQ3^2 + 41*m3^7*mQ3^4 - 
        50*m3^9*mU3^2 + 196*m3^7*mQ3^2*mU3^2 - 130*m3^5*mQ3^4*mU3^2 + 
        41*m3^7*mU3^4 - 130*m3^5*mQ3^2*mU3^4 + 81*m3^3*mQ3^4*mU3^4)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3)) + a3^2*at*k^(3/2)*mt^2*
    Log[mQ3^2/MR^2]^2*((-32*m3*mQ3^2)/(3*(m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) + 
     (384*m3*mQ3^4*(2*m3^2 - mQ3^2))/((m3^2 - mQ3^2)^3*(mQ3^2 - mU3^2)) + 
     (64*(2*m3^3 - m3*mQ3^2)*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 
        5*mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) - 
     (64*m3*(m3^8 - 91*m3^6*mQ3^2 + 107*m3^4*mQ3^4 - 41*m3^2*mQ3^6 - 
        m3^6*mU3^2 + 83*m3^4*mQ3^2*mU3^2 - 91*m3^2*mQ3^4*mU3^2 + 
        33*mQ3^6*mU3^2))/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)*
       (mQ3^2 - mU3^2)) + (64*(16*m3^7 + 42*m3^5*mQ3^2 - 58*m3^3*mQ3^4 + 
        3*m3*mQ3^6 - 30*m3^3*mQ3^2*msq^2 - 16*m3^5*mU3^2 - 
        50*m3^3*mQ3^2*mU3^2 + 60*m3*mQ3^4*mU3^2 + 30*m3*mQ3^2*msq^2*mU3^2 + 
        3*m3*mQ3^2*mU3^4))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*
       (mQ3^2 - mU3^2)) + 24*((40*m3*(mQ3^2 - msq^2)^2)/
        (3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) + 
       (4*(18*m3^5 + m3^3*mQ3^2 + 3*m3*mQ3^4 - 37*m3^3*mU3^2 - 
          7*m3*mQ3^2*mU3^2 + 22*m3*mU3^4))/(9*(m3^2 - mU3^2)^2*
         (-mQ3^2 + mU3^2)) - (4*(5*m3^5*mQ3^2 - 5*m3^3*mQ3^4 + 3*m3*mQ3^6 - 
          5*m3^5*mU3^2 - 4*m3*mQ3^4*mU3^2 + 5*m3^3*mU3^4 + 4*m3*mQ3^2*mU3^4 - 
          3*m3*mU3^6))/(9*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
       (4*(33*m3^9*mQ3^4 - 75*m3^7*mQ3^6 + 33*m3^5*mQ3^8 + 8*m3^3*mQ3^10 - 
          3*m3*mQ3^12 - 17*m3^9*mQ3^2*mU3^2 - 33*m3^7*mQ3^4*mU3^2 + 
          158*m3^5*mQ3^6*mU3^2 - 106*m3^3*mQ3^8*mU3^2 + 10*m3*mQ3^10*mU3^2 + 
          49*m3^7*mQ3^2*mU3^4 - 78*m3^5*mQ3^4*mU3^4 - 11*m3^3*mQ3^6*mU3^4 + 
          28*m3*mQ3^8*mU3^4 - 5*m3^7*mU3^6 - 22*m3^5*mQ3^2*mU3^6 + 
          43*m3^3*mQ3^4*mU3^6 - 12*m3*mQ3^6*mU3^6 + 5*m3^5*mU3^8 + 
          5*m3^3*mQ3^2*mU3^8 - 10*m3*mQ3^4*mU3^8 - 3*m3^3*mU3^10 + 
          3*m3*mQ3^2*mU3^10))/(9*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*
         (mQ3^2 - mU3^2)^2)) + (64*(18*m3^11*mQ3^2 - 85*m3^9*mQ3^4 + 
        62*m3^7*mQ3^6 - 16*m3^5*mQ3^8 + 8*m3^3*mQ3^10 - 3*m3*mQ3^12 + 
        60*m3^7*mQ3^4*msq^2 - 60*m3^5*mQ3^6*msq^2 - 30*m3^7*mQ3^2*msq^4 + 
        30*m3^5*mQ3^4*msq^4 - 18*m3^11*mU3^2 + 32*m3^9*mQ3^2*mU3^2 + 
        138*m3^7*mQ3^4*mU3^2 - 106*m3^5*mQ3^6*mU3^2 - 8*m3^3*mQ3^8*mU3^2 + 
        10*m3*mQ3^10*mU3^2 - 60*m3^7*mQ3^2*msq^2*mU3^2 - 
        60*m3^5*mQ3^4*msq^2*mU3^2 + 120*m3^3*mQ3^6*msq^2*mU3^2 + 
        30*m3^7*msq^4*mU3^2 + 30*m3^5*mQ3^2*msq^4*mU3^2 - 
        60*m3^3*mQ3^4*msq^4*mU3^2 + 37*m3^9*mU3^4 - 114*m3^7*mQ3^2*mU3^4 - 
        56*m3^5*mQ3^4*mU3^4 + 106*m3^3*mQ3^6*mU3^4 - 21*m3*mQ3^8*mU3^4 + 
        120*m3^5*mQ3^2*msq^2*mU3^4 - 60*m3^3*mQ3^4*msq^2*mU3^4 - 
        60*m3*mQ3^6*msq^2*mU3^4 - 60*m3^5*msq^4*mU3^4 + 
        30*m3^3*mQ3^2*msq^4*mU3^4 + 30*m3*mQ3^4*msq^4*mU3^4 - 22*m3^7*mU3^6 + 
        82*m3^5*mQ3^2*mU3^6 - 42*m3^3*mQ3^4*mU3^6 - 2*m3*mQ3^6*mU3^6 - 
        60*m3^3*mQ3^2*msq^2*mU3^6 + 60*m3*mQ3^4*msq^2*mU3^6 + 
        30*m3^3*msq^4*mU3^6 - 30*m3*mQ3^2*msq^4*mU3^6)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2)) + 
   a3^2*at*k^(3/2)*mt^2*Log[mU3^2/MR^2]^2*
    ((-32*m3*mU3^2)/(3*(m3^2 - mU3^2)*(-mQ3^2 + mU3^2)) + 
     (384*m3*mU3^4*(2*m3^2 - mU3^2))/((m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)) + 
     (64*(2*m3^3 - m3*mU3^2)*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 
        5*mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*
       (-mQ3^2 + mU3^2)) - (64*(16*m3^7 - 16*m3^5*mQ3^2 + 42*m3^5*mU3^2 - 
        50*m3^3*mQ3^2*mU3^2 + 3*m3*mQ3^4*mU3^2 - 30*m3^3*msq^2*mU3^2 + 
        30*m3*mQ3^2*msq^2*mU3^2 - 58*m3^3*mU3^4 + 60*m3*mQ3^2*mU3^4 + 
        3*m3*mU3^6))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) - 
     (64*m3*(m3^8 - m3^6*mQ3^2 - 91*m3^6*mU3^2 + 83*m3^4*mQ3^2*mU3^2 + 
        107*m3^4*mU3^4 - 91*m3^2*mQ3^2*mU3^4 - 41*m3^2*mU3^6 + 
        33*mQ3^2*mU3^6))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^3*
       (-mQ3^2 + mU3^2)) - (32*(-18*m3^7*mQ3^2 + m3^5*mQ3^4 + 3*m3^3*mQ3^6 + 
        30*m3^3*mQ3^2*msq^4 + 18*m3^7*mU3^2 + 69*m3^5*mQ3^2*mU3^2 - 
        11*m3^3*mQ3^4*mU3^2 - 3*m3*mQ3^6*mU3^2 - 60*m3^3*mQ3^2*msq^2*mU3^2 - 
        30*m3^3*msq^4*mU3^2 - 30*m3*mQ3^2*msq^4*mU3^2 - 86*m3^5*mU3^4 - 
        59*m3^3*mQ3^2*mU3^4 + 10*m3*mQ3^4*mU3^4 + 60*m3^3*msq^2*mU3^4 + 
        60*m3*mQ3^2*msq^2*mU3^4 + 30*m3*msq^4*mU3^4 + 99*m3^3*mU3^6 + 
        4*m3*mQ3^2*mU3^6 - 60*m3*msq^2*mU3^6 - 27*m3*mU3^8))/
      (3*(m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)^2) + 
     (64*(-18*m3^7*mQ3^2 + m3^5*mQ3^4 + 3*m3^3*mQ3^6 + 30*m3^3*mQ3^2*msq^4 + 
        18*m3^7*mU3^2 + 68*m3^5*mQ3^2*mU3^2 - 11*m3^3*mQ3^4*mU3^2 - 
        3*m3*mQ3^6*mU3^2 - 60*m3^3*mQ3^2*msq^2*mU3^2 - 30*m3^3*msq^4*mU3^2 - 
        30*m3*mQ3^2*msq^4*mU3^2 - 85*m3^5*mU3^4 - 17*m3^3*mQ3^2*mU3^4 + 
        10*m3*mQ3^4*mU3^4 + 60*m3^3*msq^2*mU3^4 + 60*m3*mQ3^2*msq^2*mU3^4 + 
        30*m3*msq^4*mU3^4 + 57*m3^3*mU3^6 - 17*m3*mQ3^2*mU3^6 - 
        60*m3*msq^2*mU3^6 - 6*m3*mU3^8)*(Log[MR^2] - 2*Log[mt] - 
        2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/(3*(m3^2 - mU3^2)^3*
       (-mQ3^2 + mU3^2)^2)) + a3^2*at*k^(3/2)*mt^2*Log[mQ3^2/MR^2]*
    ((704*m3*mQ3^2)/(3*(m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) + 
     (64*(14*m3^3 - 3*m3*mQ3^2 - 30*m3*msq^2 - 3*m3*mU3^2))/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)) + (384*m3*mQ3^2*(m3^4 - mQ3^2*mU3^2))/
      ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) - 
     (128*m3*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2))/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) - 
     (64*(16*m3^7 + 55*m3^5*mQ3^2 - 61*m3^3*mQ3^4 + 3*m3*mQ3^6 - 
        30*m3^3*mQ3^2*msq^2 - 16*m3^5*mU3^2 - 53*m3^3*mQ3^2*mU3^2 + 
        53*m3*mQ3^4*mU3^2 + 30*m3*mQ3^2*msq^2*mU3^2 + 3*m3*mQ3^2*mU3^4))/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) + 
     (512*m3*(2*m3^6*mQ3^2 - 2*m3^4*mQ3^4 + 2*m3^6*mU3^2 - 
        9*m3^4*mQ3^2*mU3^2 + 6*m3^2*mQ3^4*mU3^2 - 2*m3^4*mU3^4 + 
        6*m3^2*mQ3^2*mU3^4 - 3*mQ3^4*mU3^4))/(3*(m3^2 - mQ3^2)^2*mU3^2*
       (m3^2 - mU3^2)*(mQ3^2 - mU3^2)) - (1024*m3*mQ3^2*Log[MR^2])/
      ((m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) + 
     (512*m3*mQ3^2*(2*Log[at] - Log[k] + 4*Log[Csc[beta]]))/
      ((m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) - 
     (512*m3*mQ3^2*(2*Log[at] - Log[k] - 4*Log[mt] - 4*Log[Sin[beta]]))/
      ((m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) - 
     (1024*m3*mQ3^2*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
        2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) + 
     (128*(16*m3^7 + 42*m3^5*mQ3^2 - 58*m3^3*mQ3^4 + 3*m3*mQ3^6 - 
        30*m3^3*mQ3^2*msq^2 - 16*m3^5*mU3^2 - 50*m3^3*mQ3^2*mU3^2 + 
        60*m3*mQ3^4*mU3^2 + 30*m3*mQ3^2*msq^2*mU3^2 + 3*m3*mQ3^2*mU3^4)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) + 
     (1024*m3*mQ3^2*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
         2*Log[Sin[beta]])^2)/((m3^2 - mQ3^2)*(mQ3^2 - mU3^2))) + 
   a3^2*at*k^(3/2)*mt^2*Log[mU3^2/MR^2]*
    ((704*m3*mU3^2)/(3*(m3^2 - mU3^2)*(-mQ3^2 + mU3^2)) + 
     (64*(14*m3^3 - 3*m3*mQ3^2 - 30*m3*msq^2 - 3*m3*mU3^2))/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)) - (384*m3*mU3^2*(m3^4 - mQ3^2*mU3^2))/
      ((m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) + 
     (128*m3*(-11*m3^4 + 3*m3^2*mQ3^2 + 3*m3^2*mU3^2 + 5*mQ3^2*mU3^2))/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)*(mQ3^2 - mU3^2)) + 
     (512*m3*(-2*m3^6*mQ3^2 + 2*m3^4*mQ3^4 - 2*m3^6*mU3^2 + 
        9*m3^4*mQ3^2*mU3^2 - 6*m3^2*mQ3^4*mU3^2 + 2*m3^4*mU3^4 - 
        6*m3^2*mQ3^2*mU3^4 + 3*mQ3^4*mU3^4))/(3*mQ3^2*(-m3^2 + mQ3^2)*
       (m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)) + 
     (64*(16*m3^7 - 16*m3^5*mQ3^2 + 55*m3^5*mU3^2 - 53*m3^3*mQ3^2*mU3^2 + 
        3*m3*mQ3^4*mU3^2 - 30*m3^3*msq^2*mU3^2 + 30*m3*mQ3^2*msq^2*mU3^2 - 
        61*m3^3*mU3^4 + 53*m3*mQ3^2*mU3^4 + 3*m3*mU3^6))/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) - 
     (1024*m3*mU3^2*Log[MR^2])/((m3^2 - mU3^2)*(-mQ3^2 + mU3^2)) + 
     (512*m3*mU3^2*(2*Log[at] - Log[k] + 4*Log[Csc[beta]]))/
      ((m3^2 - mU3^2)*(-mQ3^2 + mU3^2)) - 
     (512*m3*mU3^2*(2*Log[at] - Log[k] - 4*Log[mt] - 4*Log[Sin[beta]]))/
      ((m3^2 - mU3^2)*(-mQ3^2 + mU3^2)) - 
     (1024*m3*mU3^2*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
        2*Log[Sin[beta]]))/(3*(m3^2 - mU3^2)*(-mQ3^2 + mU3^2)) - 
     (128*(16*m3^7 - 16*m3^5*mQ3^2 + 42*m3^5*mU3^2 - 50*m3^3*mQ3^2*mU3^2 + 
        3*m3*mQ3^4*mU3^2 - 30*m3^3*msq^2*mU3^2 + 30*m3*mQ3^2*msq^2*mU3^2 - 
        58*m3^3*mU3^4 + 60*m3*mQ3^2*mU3^4 + 3*m3*mU3^6)*
       (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
      (3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)) + 
     (1024*m3*mU3^2*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
         2*Log[Sin[beta]])^2)/((m3^2 - mU3^2)*(-mQ3^2 + mU3^2))) - 
   (64*a3*at*k*mt^2*Log[m3^2/MR^2]*(-m3^3 + 2*m3^3*Log[MR^2] - 
      4*m3^3*Log[mt] - 4*m3^3*Log[Csc[beta]] - 4*m3^3*Log[Sin[beta]]))/
    ((m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
   (64*a3*at*k*mt^2*Log[mQ3^2/MR^2]*(2*m3^3 - m3*mQ3^2 - 
      2*m3*mQ3^2*Log[MR^2] + 4*m3*mQ3^2*Log[mt] + 4*m3*mQ3^2*Log[Csc[beta]] + 
      4*m3*mQ3^2*Log[Sin[beta]]))/((m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) - 
   (640*a3^2*at*k^(3/2)*mt^2*Log[msq^2/MR^2]*(-(m3*msq^2) + 
      2*m3*msq^2*Log[MR^2] - 4*m3*msq^2*Log[mt] - 4*m3*msq^2*Log[Csc[beta]] - 
      4*m3*msq^2*Log[Sin[beta]]))/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
   (64*a3^2*at*k^(3/2)*mt^2*Log[m3^2/MR^2]*Log[mQ3^2/MR^2]*
     (16*m3^11*mQ3^2 - 32*m3^9*mQ3^4 + 16*m3^7*mQ3^6 + 4*m3^11*mU3^2 + 
      216*m3^9*mQ3^2*mU3^2 - 223*m3^7*mQ3^4*mU3^2 + 86*m3^5*mQ3^6*mU3^2 - 
      3*m3^3*mQ3^8*mU3^2 - 60*m3^7*mQ3^2*msq^2*mU3^2 + 
      90*m3^5*mQ3^4*msq^2*mU3^2 - 30*m3^3*mQ3^6*msq^2*mU3^2 - 
      130*m3^9*mU3^4 - 106*m3^7*mQ3^2*mU3^4 + 111*m3^5*mQ3^4*mU3^4 - 
      35*m3^3*mQ3^6*mU3^4 + 60*m3^7*msq^2*mU3^4 - 60*m3^5*mQ3^2*msq^2*mU3^4 + 
      123*m3^7*mU3^6 - 68*m3^5*mQ3^2*mU3^6 + 37*m3^3*mQ3^4*mU3^6 - 
      12*m3*mQ3^6*mU3^6 - 30*m3^5*msq^2*mU3^6 + 30*m3^3*mQ3^2*msq^2*mU3^6 - 
      3*m3^5*mU3^8 + 3*m3^3*mQ3^2*mU3^8 + 68*m3^11*mU3^2*Log[MR^2] - 
      234*m3^9*mQ3^2*mU3^2*Log[MR^2] + 142*m3^7*mQ3^4*mU3^2*Log[MR^2] - 
      8*m3^5*mQ3^6*mU3^2*Log[MR^2] - 138*m3^9*mU3^4*Log[MR^2] + 
      458*m3^7*mQ3^2*mU3^4*Log[MR^2] - 242*m3^5*mQ3^4*mU3^4*Log[MR^2] - 
      14*m3^3*mQ3^6*mU3^4*Log[MR^2] + 76*m3^7*mU3^6*Log[MR^2] - 
      258*m3^5*mQ3^2*mU3^6*Log[MR^2] + 150*m3^3*mQ3^4*mU3^6*Log[MR^2] - 
      136*m3^11*mU3^2*Log[mt] + 468*m3^9*mQ3^2*mU3^2*Log[mt] - 
      284*m3^7*mQ3^4*mU3^2*Log[mt] + 16*m3^5*mQ3^6*mU3^2*Log[mt] + 
      276*m3^9*mU3^4*Log[mt] - 916*m3^7*mQ3^2*mU3^4*Log[mt] + 
      484*m3^5*mQ3^4*mU3^4*Log[mt] + 28*m3^3*mQ3^6*mU3^4*Log[mt] - 
      152*m3^7*mU3^6*Log[mt] + 516*m3^5*mQ3^2*mU3^6*Log[mt] - 
      300*m3^3*mQ3^4*mU3^6*Log[mt] - 136*m3^11*mU3^2*Log[Csc[beta]] + 
      468*m3^9*mQ3^2*mU3^2*Log[Csc[beta]] - 284*m3^7*mQ3^4*mU3^2*
       Log[Csc[beta]] + 16*m3^5*mQ3^6*mU3^2*Log[Csc[beta]] + 
      276*m3^9*mU3^4*Log[Csc[beta]] - 916*m3^7*mQ3^2*mU3^4*Log[Csc[beta]] + 
      484*m3^5*mQ3^4*mU3^4*Log[Csc[beta]] + 28*m3^3*mQ3^6*mU3^4*
       Log[Csc[beta]] - 152*m3^7*mU3^6*Log[Csc[beta]] + 
      516*m3^5*mQ3^2*mU3^6*Log[Csc[beta]] - 300*m3^3*mQ3^4*mU3^6*
       Log[Csc[beta]] - 136*m3^11*mU3^2*Log[Sin[beta]] + 
      468*m3^9*mQ3^2*mU3^2*Log[Sin[beta]] - 284*m3^7*mQ3^4*mU3^2*
       Log[Sin[beta]] + 16*m3^5*mQ3^6*mU3^2*Log[Sin[beta]] + 
      276*m3^9*mU3^4*Log[Sin[beta]] - 916*m3^7*mQ3^2*mU3^4*Log[Sin[beta]] + 
      484*m3^5*mQ3^4*mU3^4*Log[Sin[beta]] + 28*m3^3*mQ3^6*mU3^4*
       Log[Sin[beta]] - 152*m3^7*mU3^6*Log[Sin[beta]] + 
      516*m3^5*mQ3^2*mU3^6*Log[Sin[beta]] - 300*m3^3*mQ3^4*mU3^6*
       Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^3*mU3^2*(m3^2 - mU3^2)^2*
     (mQ3^2 - mU3^2)) + a3^2*at*k^(3/2)*mt^2*Log[m3^2/MR^2]*
    ((-704*m3^3)/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
     (384*m3^3*(m3^4 - mQ3^2*mU3^2))/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
     (64*(85*m3^7 - 75*m3^5*mQ3^2 + 3*m3^3*mQ3^4 - 60*m3^5*msq^2 + 
        30*m3^3*mQ3^2*msq^2 - 75*m3^5*mU3^2 + 59*m3^3*mQ3^2*mU3^2 + 
        30*m3^3*msq^2*mU3^2 + 3*m3^3*mU3^4))/(3*(m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^2) - (512*m3^3*(2*m3^6*mQ3^2 - 2*m3^4*mQ3^4 + 
        2*m3^6*mU3^2 - 9*m3^4*mQ3^2*mU3^2 + 6*m3^2*mQ3^4*mU3^2 - 
        2*m3^4*mU3^4 + 6*m3^2*mQ3^2*mU3^4 - 3*mQ3^4*mU3^4))/
      (3*mQ3^2*(m3^2 - mQ3^2)^2*mU3^2*(m3^2 - mU3^2)^2) + 
     (1024*m3^3*Log[MR^2])/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
     (512*m3^3*(2*Log[at] - Log[k] + 4*Log[Csc[beta]]))/
      ((m3^2 - mQ3^2)*(m3^2 - mU3^2)) + 
     (1024*m3^3*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 
        2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
     (128*(24*m3^7 - 24*m3^5*mQ3^2 + m3^3*mQ3^4 - 20*m3^5*msq^2 + 
        10*m3^3*mQ3^2*msq^2 - 24*m3^5*mU3^2 + 22*m3^3*mQ3^2*mU3^2 + 
        10*m3^3*msq^2*mU3^2 + m3^3*mU3^4)*(Log[MR^2] - 2*Log[mt] - 
        2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/((m3^2 - mQ3^2)^2*
       (m3^2 - mU3^2)^2) - 
     (1024*m3^3*(Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]])^
        2)/((m3^2 - mQ3^2)*(m3^2 - mU3^2)) - 
     (2048*m3^3*(-Log[at]/2 + Log[k]/4 + Log[mt] + Log[Sin[beta]]))/
      ((m3^2 - mQ3^2)*(m3^2 - mU3^2))) + Log[mU3^2/MR^2]*
    ((-128*a3*at*k*m3^3*mt^2*Log[m3^2/MR^2])/((m3^2 - mU3^2)*
       (-mQ3^2 + mU3^2)) + (64*a3^2*at*k^(3/2)*mt^2*
       (44*m3^13 - 77*m3^11*mQ3^2 + 22*m3^9*mQ3^4 + 19*m3^7*mQ3^6 - 
        79*m3^11*mU3^2 + 72*m3^9*mQ3^2*mU3^2 + 101*m3^7*mQ3^4*mU3^2 - 
        118*m3^5*mQ3^6*mU3^2 + 122*m3^9*mU3^4 - 209*m3^7*mQ3^2*mU3^4 + 
        36*m3^5*mQ3^4*mU3^4 + 75*m3^3*mQ3^6*mU3^4 - 63*m3^7*mU3^6 + 
        142*m3^5*mQ3^2*mU3^6 - 87*m3^3*mQ3^4*mU3^6)*Log[m3^2/MR^2]^2)/
      (3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^3*(mQ3^2 - mU3^2)) - 
     (512*a3^2*at*k^(3/2)*mt^2*(5*m3^11 - 8*m3^9*mQ3^2 + 5*m3^7*mQ3^4 - 
        8*m3^9*mU3^2 + 16*m3^7*mQ3^2*mU3^2 - 12*m3^5*mQ3^4*mU3^2 + 
        5*m3^7*mU3^4 - 12*m3^5*mQ3^2*mU3^4 + 9*m3^3*mQ3^4*mU3^4)*
       Log[m3^2/MR^2]*Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^3) + (32*a3^2*at*k^(3/2)*mt^2*(20*m3^11*mQ3^2 - 
        91*m3^9*mQ3^4 + 122*m3^7*mQ3^6 - 30*m3^5*mQ3^8 - 8*m3^3*mQ3^10 + 
        3*m3*mQ3^12 + 60*m3^7*mQ3^4*msq^2 - 60*m3^5*mQ3^6*msq^2 - 
        30*m3^7*mQ3^2*msq^4 + 30*m3^5*mQ3^4*msq^4 - 20*m3^11*mU3^2 + 
        51*m3^9*mQ3^2*mU3^2 + 65*m3^7*mQ3^4*mU3^2 - 219*m3^5*mQ3^6*mU3^2 + 
        85*m3^3*mQ3^8*mU3^2 - 10*m3*mQ3^10*mU3^2 - 60*m3^7*mQ3^2*msq^2*
         mU3^2 - 60*m3^5*mQ3^4*msq^2*mU3^2 + 120*m3^3*mQ3^6*msq^2*mU3^2 + 
        30*m3^7*msq^4*mU3^2 + 30*m3^5*mQ3^2*msq^4*mU3^2 - 
        60*m3^3*mQ3^4*msq^4*mU3^2 + 56*m3^9*mU3^4 - 244*m3^7*mQ3^2*mU3^4 + 
        278*m3^5*mQ3^4*mU3^4 - 42*m3^3*mQ3^6*mU3^4 + 120*m3^5*mQ3^2*msq^2*
         mU3^4 - 60*m3^3*mQ3^4*msq^2*mU3^4 - 60*m3*mQ3^6*msq^2*mU3^4 - 
        60*m3^5*msq^4*mU3^4 + 30*m3^3*mQ3^2*msq^4*mU3^4 + 
        30*m3*mQ3^4*msq^4*mU3^4 - 7*m3^7*mU3^6 + 67*m3^5*mQ3^2*mU3^6 - 
        99*m3^3*mQ3^4*mU3^6 + 23*m3*mQ3^6*mU3^6 - 60*m3^3*mQ3^2*msq^2*mU3^6 + 
        60*m3*mQ3^4*msq^2*mU3^6 + 30*m3^3*msq^4*mU3^6 - 
        30*m3*mQ3^2*msq^4*mU3^6)*Log[mQ3^2/MR^2]^2)/(3*(m3^2 - mQ3^2)^3*
       (m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2) - 
     (64*a3*at*k*mt^2*(2*m3^3 - m3*mU3^2 - 2*m3*mU3^2*Log[MR^2] + 
        4*m3*mU3^2*Log[mt] + 4*m3*mU3^2*Log[Csc[beta]] + 
        4*m3*mU3^2*Log[Sin[beta]]))/((m3^2 - mU3^2)*(-mQ3^2 + mU3^2))) + 
   Log[mU3^2/MR^2]*((64*a3^2*at*k^(3/2)*mt^2*Log[mQ3^2/MR^2]*
       (109*m3^7*mQ3^4 - 104*m3^5*mQ3^6 - 2*m3^3*mQ3^8 + 3*m3*mQ3^10 - 
        30*m3^5*mQ3^4*msq^2 - 234*m3^7*mQ3^2*mU3^2 + 123*m3^5*mQ3^4*mU3^2 + 
        60*m3^3*mQ3^6*mU3^2 - 10*m3*mQ3^8*mU3^2 + 60*m3^5*mQ3^2*msq^2*mU3^2 + 
        30*m3*mQ3^6*msq^2*mU3^2 + 109*m3^7*mU3^4 + 138*m3^5*mQ3^2*mU3^4 - 
        159*m3^3*mQ3^4*mU3^4 + 19*m3*mQ3^6*mU3^4 - 30*m3^5*msq^2*mU3^4 - 
        60*m3*mQ3^4*msq^2*mU3^4 - 109*m3^5*mU3^6 + 50*m3^3*mQ3^2*mU3^6 + 
        4*m3*mQ3^4*mU3^6 + 30*m3*mQ3^2*msq^2*mU3^6 + 3*m3^3*mU3^8 - 
        10*m3^5*mQ3^6*Log[MR^2] + 10*m3^3*mQ3^8*Log[MR^2] - 
        6*m3*mQ3^10*Log[MR^2] + 32*m3^7*mQ3^2*mU3^2*Log[MR^2] - 
        28*m3^5*mQ3^4*mU3^2*Log[MR^2] + 10*m3^3*mQ3^6*mU3^2*Log[MR^2] + 
        20*m3*mQ3^8*mU3^2*Log[MR^2] - 58*m3^5*mQ3^2*mU3^4*Log[MR^2] + 
        46*m3^3*mQ3^4*mU3^4*Log[MR^2] - 38*m3*mQ3^6*mU3^4*Log[MR^2] + 
        30*m3^3*mQ3^2*mU3^6*Log[MR^2] - 8*m3*mQ3^4*mU3^6*Log[MR^2] + 
        20*m3^5*mQ3^6*Log[mt] - 20*m3^3*mQ3^8*Log[mt] + 
        12*m3*mQ3^10*Log[mt] - 64*m3^7*mQ3^2*mU3^2*Log[mt] + 
        56*m3^5*mQ3^4*mU3^2*Log[mt] - 20*m3^3*mQ3^6*mU3^2*Log[mt] - 
        40*m3*mQ3^8*mU3^2*Log[mt] + 116*m3^5*mQ3^2*mU3^4*Log[mt] - 
        92*m3^3*mQ3^4*mU3^4*Log[mt] + 76*m3*mQ3^6*mU3^4*Log[mt] - 
        60*m3^3*mQ3^2*mU3^6*Log[mt] + 16*m3*mQ3^4*mU3^6*Log[mt] + 
        20*m3^5*mQ3^6*Log[Csc[beta]] - 20*m3^3*mQ3^8*Log[Csc[beta]] + 
        12*m3*mQ3^10*Log[Csc[beta]] - 64*m3^7*mQ3^2*mU3^2*Log[Csc[beta]] + 
        56*m3^5*mQ3^4*mU3^2*Log[Csc[beta]] - 20*m3^3*mQ3^6*mU3^2*
         Log[Csc[beta]] - 40*m3*mQ3^8*mU3^2*Log[Csc[beta]] + 
        116*m3^5*mQ3^2*mU3^4*Log[Csc[beta]] - 92*m3^3*mQ3^4*mU3^4*
         Log[Csc[beta]] + 76*m3*mQ3^6*mU3^4*Log[Csc[beta]] - 
        60*m3^3*mQ3^2*mU3^6*Log[Csc[beta]] + 16*m3*mQ3^4*mU3^6*
         Log[Csc[beta]] + 20*m3^5*mQ3^6*Log[Sin[beta]] - 
        20*m3^3*mQ3^8*Log[Sin[beta]] + 12*m3*mQ3^10*Log[Sin[beta]] - 
        64*m3^7*mQ3^2*mU3^2*Log[Sin[beta]] + 56*m3^5*mQ3^4*mU3^2*
         Log[Sin[beta]] - 20*m3^3*mQ3^6*mU3^2*Log[Sin[beta]] - 
        40*m3*mQ3^8*mU3^2*Log[Sin[beta]] + 116*m3^5*mQ3^2*mU3^4*
         Log[Sin[beta]] - 92*m3^3*mQ3^4*mU3^4*Log[Sin[beta]] + 
        76*m3*mQ3^6*mU3^4*Log[Sin[beta]] - 60*m3^3*mQ3^2*mU3^6*
         Log[Sin[beta]] + 16*m3*mQ3^4*mU3^6*Log[Sin[beta]]))/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(mQ3^2 - mU3^2)^2) + 
     (64*a3^2*at*k^(3/2)*mt^2*Log[m3^2/MR^2]*(4*m3^11*mQ3^2 - 
        130*m3^9*mQ3^4 + 123*m3^7*mQ3^6 - 3*m3^5*mQ3^8 + 
        60*m3^7*mQ3^4*msq^2 - 30*m3^5*mQ3^6*msq^2 + 16*m3^11*mU3^2 + 
        216*m3^9*mQ3^2*mU3^2 - 106*m3^7*mQ3^4*mU3^2 - 68*m3^5*mQ3^6*mU3^2 + 
        3*m3^3*mQ3^8*mU3^2 - 60*m3^7*mQ3^2*msq^2*mU3^2 - 
        60*m3^5*mQ3^4*msq^2*mU3^2 + 30*m3^3*mQ3^6*msq^2*mU3^2 - 
        32*m3^9*mU3^4 - 223*m3^7*mQ3^2*mU3^4 + 111*m3^5*mQ3^4*mU3^4 + 
        37*m3^3*mQ3^6*mU3^4 + 90*m3^5*mQ3^2*msq^2*mU3^4 + 16*m3^7*mU3^6 + 
        86*m3^5*mQ3^2*mU3^6 - 35*m3^3*mQ3^4*mU3^6 - 12*m3*mQ3^6*mU3^6 - 
        30*m3^3*mQ3^2*msq^2*mU3^6 - 3*m3^3*mQ3^2*mU3^8 + 
        68*m3^11*mQ3^2*Log[MR^2] - 138*m3^9*mQ3^4*Log[MR^2] + 
        76*m3^7*mQ3^6*Log[MR^2] - 234*m3^9*mQ3^2*mU3^2*Log[MR^2] + 
        458*m3^7*mQ3^4*mU3^2*Log[MR^2] - 258*m3^5*mQ3^6*mU3^2*Log[MR^2] + 
        142*m3^7*mQ3^2*mU3^4*Log[MR^2] - 242*m3^5*mQ3^4*mU3^4*Log[MR^2] + 
        150*m3^3*mQ3^6*mU3^4*Log[MR^2] - 8*m3^5*mQ3^2*mU3^6*Log[MR^2] - 
        14*m3^3*mQ3^4*mU3^6*Log[MR^2] - 136*m3^11*mQ3^2*Log[mt] + 
        276*m3^9*mQ3^4*Log[mt] - 152*m3^7*mQ3^6*Log[mt] + 
        468*m3^9*mQ3^2*mU3^2*Log[mt] - 916*m3^7*mQ3^4*mU3^2*Log[mt] + 
        516*m3^5*mQ3^6*mU3^2*Log[mt] - 284*m3^7*mQ3^2*mU3^4*Log[mt] + 
        484*m3^5*mQ3^4*mU3^4*Log[mt] - 300*m3^3*mQ3^6*mU3^4*Log[mt] + 
        16*m3^5*mQ3^2*mU3^6*Log[mt] + 28*m3^3*mQ3^4*mU3^6*Log[mt] - 
        136*m3^11*mQ3^2*Log[Csc[beta]] + 276*m3^9*mQ3^4*Log[Csc[beta]] - 
        152*m3^7*mQ3^6*Log[Csc[beta]] + 468*m3^9*mQ3^2*mU3^2*Log[Csc[beta]] - 
        916*m3^7*mQ3^4*mU3^2*Log[Csc[beta]] + 516*m3^5*mQ3^6*mU3^2*
         Log[Csc[beta]] - 284*m3^7*mQ3^2*mU3^4*Log[Csc[beta]] + 
        484*m3^5*mQ3^4*mU3^4*Log[Csc[beta]] - 300*m3^3*mQ3^6*mU3^4*
         Log[Csc[beta]] + 16*m3^5*mQ3^2*mU3^6*Log[Csc[beta]] + 
        28*m3^3*mQ3^4*mU3^6*Log[Csc[beta]] - 136*m3^11*mQ3^2*Log[Sin[beta]] + 
        276*m3^9*mQ3^4*Log[Sin[beta]] - 152*m3^7*mQ3^6*Log[Sin[beta]] + 
        468*m3^9*mQ3^2*mU3^2*Log[Sin[beta]] - 916*m3^7*mQ3^4*mU3^2*
         Log[Sin[beta]] + 516*m3^5*mQ3^6*mU3^2*Log[Sin[beta]] - 
        284*m3^7*mQ3^2*mU3^4*Log[Sin[beta]] + 484*m3^5*mQ3^4*mU3^4*
         Log[Sin[beta]] - 300*m3^3*mQ3^6*mU3^4*Log[Sin[beta]] + 
        16*m3^5*mQ3^2*mU3^6*Log[Sin[beta]] + 28*m3^3*mQ3^4*mU3^6*
         Log[Sin[beta]]))/(3*mQ3^2*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*
       (mQ3^2 - mU3^2))) + 
   ((256*a3*at*k*m3^3*mt^2)/((m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) + 
     (512*a3^2*at*k^(3/2)*mt^2*(11*m3^11 - 10*m3^9*mQ3^2 + 7*m3^7*mQ3^4 - 
        18*m3^9*mU3^2 + 8*m3^7*mQ3^2*mU3^2 - 6*m3^5*mQ3^4*mU3^2 + 
        11*m3^7*mU3^4 - 6*m3^5*mQ3^2*mU3^4 + 3*m3^3*mQ3^4*mU3^4)*
       Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^3*(m3^2 - mU3^2)^2*
       (mQ3^2 - mU3^2)) + a3^2*at*k^(3/2)*mt^2*
      ((128*(m3^7 - 50*m3^5*mQ3^2 + 17*m3^3*mQ3^4))/(3*(m3^2 - mQ3^2)^3*
         (mQ3^2 - mU3^2)) - (64*(18*m3^5 + m3^3*mQ3^2 + 3*m3*mQ3^4 - 
          37*m3^3*mU3^2 - 7*m3*mQ3^2*mU3^2 + 22*m3*mU3^4))/
        (3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)))*Log[mQ3^2/MR^2] + 
     (1280*a3^2*at*k^(3/2)*m3^3*mt^2*Log[msq^2/MR^2])/
      (3*(m3^2 - mQ3^2)*(mQ3^2 - mU3^2)) - 
     (64*a3^2*at*k^(3/2)*mt^2*(20*m3^9 - 37*m3^7*mQ3^2 + 19*m3^5*mQ3^4 - 
        5*m3^3*mQ3^6 + 3*m3*mQ3^8 - 73*m3^7*mU3^2 + 135*m3^5*mQ3^2*mU3^2 - 
        23*m3^3*mQ3^4*mU3^2 - 7*m3*mQ3^6*mU3^2 + 24*m3^5*mU3^4 - 
        78*m3^3*mQ3^2*mU3^4 + 22*m3*mQ3^4*mU3^4)*Log[mU3^2/MR^2])/
      (3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)) + 
     a3^2*at*k^(3/2)*mt^2*((256*m3^3*(-11*m3^4 + 3*m3^2*mQ3^2 + 
          3*m3^2*mU3^2 + 5*mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)*
         (mQ3^2 - mU3^2)) + (64*(18*m3^5 + m3^3*mQ3^2 + 3*m3*mQ3^4 - 
          37*m3^3*mU3^2 - 7*m3*mQ3^2*mU3^2 + 22*m3*mU3^4))/
        (3*(m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)) - 
       (128*(18*m3^5 + m3^3*mQ3^2 + 3*m3*mQ3^4 - 37*m3^3*mU3^2 - 
          7*m3*mQ3^2*mU3^2 + 22*m3*mU3^4)*(Log[MR^2] - 2*Log[mt] - 
          2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/(3*(m3^2 - mU3^2)^2*
         (-mQ3^2 + mU3^2))))*PolyLog[2, 1 - m3^2/mQ3^2] + 
   ((-640*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*(m3^3*mQ3^2 - 2*m3^3*msq^2 + 
        m3*mQ3^2*msq^2 + m3^3*mU3^2 - 2*m3*mQ3^2*mU3^2 + m3*msq^2*mU3^2)*
       Log[mQ3^2/MR^2])/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) - 
     (640*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*(m3^3*mQ3^2 - 2*m3^3*msq^2 + 
        m3*mQ3^2*msq^2 + m3^3*mU3^2 - 2*m3*mQ3^2*mU3^2 + m3*msq^2*mU3^2)*
       Log[mU3^2/MR^2])/((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) - 
     (640*a3^2*at*k^(3/2)*(m3^2 - msq^2)*mt^2*(m3^3*mQ3^2 - 2*m3^3*msq^2 + 
        m3*mQ3^2*msq^2 + m3^3*mU3^2 - 2*m3*mQ3^2*mU3^2 + m3*msq^2*mU3^2)*
       (-1 + 2*Log[MR^2] - 4*Log[mt] - 4*Log[Csc[beta]] - 4*Log[Sin[beta]]))/
      ((m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2))*PolyLog[2, 1 - m3^2/msq^2] + 
   ((-640*a3^2*at*k^(3/2)*m3*(mQ3^2 - msq^2)^2*mt^2*Log[mQ3^2/MR^2])/
      ((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) - 
     (640*a3^2*at*k^(3/2)*m3*(mQ3^2 - msq^2)^2*mt^2*Log[mU3^2/MR^2])/
      ((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)) - 
     (640*a3^2*at*k^(3/2)*m3*(mQ3^2 - msq^2)^2*mt^2*(-1 + 2*Log[MR^2] - 
        4*Log[mt] - 4*Log[Csc[beta]] - 4*Log[Sin[beta]]))/
      ((m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2)))*PolyLog[2, 1 - msq^2/mQ3^2] + 
   ((256*a3*at*k*m3^3*mt^2)/((m3^2 - mU3^2)*(-mQ3^2 + mU3^2)) - 
     (512*a3^2*at*k^(3/2)*mt^2*(11*m3^11 - 18*m3^9*mQ3^2 + 11*m3^7*mQ3^4 - 
        10*m3^9*mU3^2 + 8*m3^7*mQ3^2*mU3^2 - 6*m3^5*mQ3^4*mU3^2 + 
        7*m3^7*mU3^4 - 6*m3^5*mQ3^2*mU3^4 + 3*m3^3*mQ3^4*mU3^4)*
       Log[m3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^3*
       (mQ3^2 - mU3^2)) - (64*a3^2*at*k^(3/2)*mt^2*(20*m3^9 - 73*m3^7*mQ3^2 + 
        24*m3^5*mQ3^4 - 37*m3^7*mU3^2 + 135*m3^5*mQ3^2*mU3^2 - 
        78*m3^3*mQ3^4*mU3^2 + 19*m3^5*mU3^4 - 23*m3^3*mQ3^2*mU3^4 + 
        22*m3*mQ3^4*mU3^4 - 5*m3^3*mU3^6 - 7*m3*mQ3^2*mU3^6 + 3*m3*mU3^8)*
       Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2*
       (mQ3^2 - mU3^2)) + (1280*a3^2*at*k^(3/2)*m3^3*mt^2*Log[msq^2/MR^2])/
      (3*(m3^2 - mU3^2)*(-mQ3^2 + mU3^2)) + a3^2*at*k^(3/2)*mt^2*
      ((-64*(18*m3^5 - 37*m3^3*mQ3^2 + 22*m3*mQ3^4 + m3^3*mU3^2 - 
          7*m3*mQ3^2*mU3^2 + 3*m3*mU3^4))/(3*(m3^2 - mQ3^2)^2*
         (mQ3^2 - mU3^2)) + (128*(m3^7 - 50*m3^5*mU3^2 + 17*m3^3*mU3^4))/
        (3*(m3^2 - mU3^2)^3*(-mQ3^2 + mU3^2)))*Log[mU3^2/MR^2] + 
     a3^2*at*k^(3/2)*mt^2*((256*m3^3*(-11*m3^4 + 3*m3^2*mQ3^2 + 
          3*m3^2*mU3^2 + 5*mQ3^2*mU3^2))/(3*(m3^2 - mQ3^2)*(m3^2 - mU3^2)^2*
         (-mQ3^2 + mU3^2)) + (64*(18*m3^5 - 37*m3^3*mQ3^2 + 22*m3*mQ3^4 + 
          m3^3*mU3^2 - 7*m3*mQ3^2*mU3^2 + 3*m3*mU3^4))/(3*(m3^2 - mQ3^2)^2*
         (mQ3^2 - mU3^2)) - (128*(18*m3^5 - 37*m3^3*mQ3^2 + 22*m3*mQ3^4 + 
          m3^3*mU3^2 - 7*m3*mQ3^2*mU3^2 + 3*m3*mU3^4)*
         (Log[MR^2] - 2*Log[mt] - 2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/
        (3*(m3^2 - mQ3^2)^2*(mQ3^2 - mU3^2))))*PolyLog[2, 1 - m3^2/mU3^2] + 
   ((32*a3^2*at*k^(3/2)*mt^2*(10*m3^5*mQ3^2 - 10*m3^3*mQ3^4 + 6*m3*mQ3^6 - 
        10*m3^5*mU3^2 - 8*m3*mQ3^4*mU3^2 + 10*m3^3*mU3^4 + 8*m3*mQ3^2*mU3^4 - 
        6*m3*mU3^6)*Log[mQ3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
     (32*a3^2*at*k^(3/2)*mt^2*(10*m3^5*mQ3^2 - 10*m3^3*mQ3^4 + 6*m3*mQ3^6 - 
        10*m3^5*mU3^2 - 8*m3*mQ3^4*mU3^2 + 10*m3^3*mU3^4 + 8*m3*mQ3^2*mU3^4 - 
        6*m3*mU3^6)*Log[mU3^2/MR^2])/(3*(m3^2 - mQ3^2)^2*(m3^2 - mU3^2)^2) + 
     a3^2*at*k^(3/2)*mt^2*((-32*(10*m3^5*mQ3^2 - 10*m3^3*mQ3^4 + 6*m3*mQ3^6 - 
          10*m3^5*mU3^2 - 8*m3*mQ3^4*mU3^2 + 10*m3^3*mU3^4 + 
          8*m3*mQ3^2*mU3^4 - 6*m3*mU3^6))/(3*(m3^2 - mQ3^2)^2*
         (m3^2 - mU3^2)^2) + (64*(10*m3^5*mQ3^2 - 10*m3^3*mQ3^4 + 
          6*m3*mQ3^6 - 10*m3^5*mU3^2 - 8*m3*mQ3^4*mU3^2 + 10*m3^3*mU3^4 + 
          8*m3*mQ3^2*mU3^4 - 6*m3*mU3^6)*(Log[MR^2] - 2*Log[mt] - 
          2*Log[Csc[beta]] - 2*Log[Sin[beta]]))/(3*(m3^2 - mQ3^2)^2*
         (m3^2 - mU3^2)^2)))*PolyLog[2, 1 - mQ3^2/mU3^2] + 
   ((-640*a3^2*at*k^(3/2)*m3*mt^2*(-msq^2 + mU3^2)^2*Log[mQ3^2/MR^2])/
      ((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)) - 
     (640*a3^2*at*k^(3/2)*m3*mt^2*(-msq^2 + mU3^2)^2*Log[mU3^2/MR^2])/
      ((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)) - 
     (640*a3^2*at*k^(3/2)*m3*mt^2*(-msq^2 + mU3^2)^2*(-1 + 2*Log[MR^2] - 
        4*Log[mt] - 4*Log[Csc[beta]] - 4*Log[Sin[beta]]))/
      ((m3^2 - mU3^2)^2*(-mQ3^2 + mU3^2)))*PolyLog[2, 1 - msq^2/mU3^2])
);
